# aclStft
## 支持的产品型号

Atlas A2训练系列产品。

## 接口原型

每个算子分为两段式接口，必须先调用“aclStftGetWorkspaceSize”接口获取计算所需workspace大小以及包含了算子计算流程的执行器，再调用“aclStft”接口执行计算。

* `aclnnStatus aclStftGetWorkspaceSize(const aclTensor *self, const aclTensor *window, aclTensor *out, int64_t nFft, int64_t hopLength, int64_t winLength, bool normalized, bool onesided, bool returnComplex, uint64_t *workspaceSize, aclOpExecutor **executor)`
* `aclnnStatus aclStft(void* workspace, uint64_t workspaceSize, aclOpExecutor* executor, const aclrtStream stream)`

## 功能描述

-   **算子功能：**计算输入在滑动窗口内的傅里叶变换。
-   **计算公式：**

    $$
    X[w,m]=\sum_{k=0}^{winLength-1}window[k]self[m*hopLength+k]exp(-j*\frac{2{\pi}wk}{nFft})
    $$

## aclStftGetWorkspaceSize

-   **参数说明：**
    -   self（aclTensor\*，计算输入）：必选参数，待计算的输入，要求是一个1D/2D的Tensor，shape为\(L\)/\(B, L\)，数据类型支持FLOAT32，数据格式要求为ND。
    -   window（aclTensor\*，计算输入）：可选参数，要求是一个1D的Tensor，shape为\(winLength\)，数据类型支持FLOAT32，数据格式要求为ND。
    -   nFft（int64\_t，计算输入）：必选参数，Host侧的int，FFT的点数。
    -   hopLength（int64\_t，计算输入）：可选参数，Host侧的int，滑动窗口的间隔。
    -   winLength（int64\_t，计算输入）：可选参数，Host侧的int，window的大小。
    -   normalized（bool，计算输入）：可选参数，Host侧的bool，是否对傅里叶变换结果进行标准化。
    -   onesided（bool，计算输入）：可选参数，Host侧的bool，是否返回全部的结果或者一半结果。
    -   returnComplex（bool，计算输入）：可选参数，Host侧的bool，确认返回值是complex tensor或者是实部、虚部分开的tensor。
    -   workspaceSize（uint64\_t\*，出参）：Device侧的整型，返回需要在Device侧申请的workspace大小。
    -   executor（aclOpExecutor\*\*，出参）：Device侧的aclOpExecutor，返回op执行器，包含了算子计算流程。
    -   out（aclTensor\*，计算输出）：self在window内的傅里叶变换结果，要求是一个2D/3D/4D的Tensor。
        - 如果returnComplex=True，out是shape为\[N, T\]或者\[B, N, T\]的复数tensor。
        - 如果returnComplex=False，out是shape为\[N, T, 2\]或者\[B, N, T, 2\]的tensor。
        
        其中，N=nFft\(onesided=False\)或者(nFft // 2 + 1)(onesided=True)，T是滑动窗口的个数，T = L // hopLength + 1。当前仅支持returnComplex=False，数据类型支持FLOAT32，数据格式要求为ND。

-   **返回值：**

    返回aclnnStatus状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

## aclStft

-   **参数说明：**
    -   workspace（void\*，入参）：在Device侧申请的workspace内存起址。
    -   workspaceSize（uint64\_t，入参）：在Device侧申请的workspace大小，由第一段接口aclStftGetWorkspaceSize获取。
    -   executor（aclOpExecutor\*，入参）：op执行器，包含了算子计算流程。
    -   stream（aclrtStream，入参）：指定执行任务的AscendCL stream流。

## 约束与限制

-   输入self与PyTorch接口的不同：PyTorch接口的输入self为原始输入；aclStftGetWorkspaceSize的入参self是原始输入经过前端PyTorch补pad后得到的结果。
-   nFft <= sizeof\(self\)。
-   hopLength的缺省值：None，此时 hopLength = floor\(nFft / 4\)，滑动窗口的大小为nFft。
-   winLength的缺省值：None，此时 winLength=nFft。
-   normalized的缺省值：False；当normalized=True时，
    
    $$
     STFT(w,m)=\frac{1}{\sqrt{N}}X[w,m]
    $$

-   onesided的缺省值：True（input和window为实数）；False（其他）。
-   returnComplex的缺省值：False，在PyTorch 2.0版本后缺省值：True，且输入为实数时，此参数为必选。

## 调用示例

```
#include <iostream>
#include <vector>
#include "acl/acl.h"
#include "aclnnop/acl_stft.h"
#define CHECK_RET(cond, return_expr) \
  do {                               \
    if (!(cond)) {                   \
      return_expr;                   \
    }                                \
  } while (0)
#define LOG_PRINT(message, ...)     \
  do {                              \
    printf(message, ##__VA_ARGS__); \
  } while (0)
int64_t GetShapeSize(const std::vector<int64_t>& shape) {
  int64_t shapeSize = 1;
  for (auto i : shape) {
    shapeSize *= i;
  }
  return shapeSize;
}
int Init(int32_t deviceId, aclrtContext* context, aclrtStream* stream) {
  // 固定写法，acl初始化
  auto ret = aclrtSetDevice(deviceId);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSetDevice failed. ERROR: %d\n", ret); return ret);
  ret = aclrtCreateContext(context, deviceId);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtCreateContext failed. ERROR: %d\n", ret); return ret);
  ret = aclrtSetCurrentContext(*context);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSetCurrentContext failed. ERROR: %d\n", ret); return ret);
  ret = aclrtCreateStream(stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtCreateStream failed. ERROR: %d\n", ret); return ret);
  ret = aclInit(nullptr);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclInit failed. ERROR: %d\n", ret); return ret);
  return 0;
}
template <typename T>
int CreateAclTensor(const std::vector<T>& hostData, const std::vector<int64_t>& shape, void** deviceAddr,
                    aclDataType dataType, aclTensor** tensor) {
  auto size = GetShapeSize(shape) * sizeof(T);
  // 调用aclrtMalloc申请device侧内存
  auto ret = aclrtMalloc(deviceAddr, size, ACL_MEM_MALLOC_HUGE_FIRST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtMalloc failed. ERROR: %d\n", ret); return ret);
  // 调用aclrtMemcpy将host侧数据拷贝到device侧内存上
  ret = aclrtMemcpy(*deviceAddr, size, hostData.data(), size, ACL_MEMCPY_HOST_TO_DEVICE);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtMemcpy failed. ERROR: %d\n", ret); return ret);
  // 计算连续tensor的strides
  std::vector<int64_t> strides(shape.size(), 1);
  for (int64_t i = shape.size() - 2; i >= 0; i--) {
    strides[i] = shape[i + 1] * strides[i + 1];
  }
  // 调用aclCreateTensor接口创建aclTensor
  *tensor = aclCreateTensor(shape.data(), shape.size(), dataType, strides.data(), 0, aclFormat::ACL_FORMAT_ND,
                            shape.data(), shape.size(), *deviceAddr);
  return 0;
}
int main() {
  // 1. （固定写法）device/context/stream初始化，参考acl对外接口列表
  // 根据自己的实际device填写deviceId
  int32_t deviceId = 0;
  aclrtContext context;
  aclrtStream stream;
  auto ret = Init(deviceId, &context, &stream);
  // check根据自己的需要处理
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("Init acl failed. ERROR: %d\n", ret); return ret);
  // 2. 构造输入与输出，需要根据API的接口自定义构造
  std::vector<int64_t> selfShape = {5};
  std::vector<int64_t> windowShape = {4};
  std::vector<int64_t> outShape = {3, 1, 2};
  void* selfDeviceAddr = nullptr;
  void* windowDeviceAddr = nullptr;
  void* outDeviceAddr = nullptr;
  aclTensor* self = nullptr;
  aclTensor* window = nullptr;
  aclTensor* out = nullptr;
  std::vector<float> selfHostData = {1, 6, 8, 5, 7};
  std::vector<float> windowHostData = {1, 1, 1, 1};
  std::vector<float> outHostData = {0, 0, 0, 0, 0, 0};
  // 创建self aclTensor
  ret = CreateAclTensor(selfHostData, selfShape, &selfDeviceAddr, aclDataType::ACL_FLOAT, &self);
  // 创建window aclTensor
  ret = CreateAclTensor(windowHostData, windowShape, &windowDeviceAddr, aclDataType::ACL_FLOAT, &self);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  // 创建out aclTensor
  ret = CreateAclTensor(outHostData, outShape, &outDeviceAddr, aclDataType::ACL_FLOAT, &out);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  int n_fft = 4;
  int hop_length = 2;
  int win_length = 4;
  bool normalized = false;
  bool onesided = true;
  bool returnComplex = false;
  // 3. 调用CANN算子库API，需要修改为具体的Api名称
  uint64_t workspaceSize = 0;
  aclOpExecutor* executor;
  // 调用aclStft第一段接口
  ret = aclStftGetWorkspaceSize(self, window, out, n_fft, hop_length, win_length, normalized, onesided, returnComplex, &workspaceSize, &executor);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclStftGetWorkspaceSize failed. ERROR: %d\n", ret); return ret);
  // 根据第一段接口计算出的workspaceSize申请device内存
  void* workspaceAddr = nullptr;
  if (workspaceSize > 0) {
    ret = aclrtMalloc(&workspaceAddr, workspaceSize, ACL_MEM_MALLOC_HUGE_FIRST);
    CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("allocate workspace failed. ERROR: %d\n", ret); return ret);
  }
  // 调用aclStft第二段接口
  ret = aclStft(workspaceAddr, workspaceSize, executor, stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclStft failed. ERROR: %d\n", ret); return ret);
  // 4. （固定写法）同步等待任务执行结束
  ret = aclrtSynchronizeStream(stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSynchronizeStream failed. ERROR: %d\n", ret); return ret);
  // 5. 获取输出的值，将device侧内存上的结果拷贝至host侧，需要根据具体API的接口定义修改
  auto size = GetShapeSize(outShape);
  std::vector<float> resultData(size, 0);
  ret = aclrtMemcpy(resultData.data(), resultData.size() * sizeof(resultData[0]), outDeviceAddr,
                    size * sizeof(resultData[0]), ACL_MEMCPY_DEVICE_TO_HOST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("copy result from device to host failed. ERROR: %d\n", ret); return ret);
  for (int64_t i = 0; i < size; i++) {
    LOG_PRINT("result[%ld] is: %f\n", i, resultData[i]);
  }
  // 6. 释放aclTensor，需要根据具体API的接口定义修改
  aclDestroyTensor(self);
  aclDestroyTensor(window);
  aclDestroyTensor(out);
  // 7. 释放device资源，需要根据具体API的接口定义修改
  aclrtFree(selfDeviceAddr);
  aclrtFree(windowDeviceAddr);
  aclrtFree(outDeviceAddr);
  if (workspaceSize > 0) {
    aclrtFree(workspaceAddr);
  }
  aclrtDestroyStream(stream);
  aclrtDestroyContext(context);
  aclrtResetDevice(deviceId);
  aclFinalize();
  return 0;
}
```

