# aclnnAddLayerNormGrad

## 支持的产品型号
Atlas A2训练系列产品。

## 接口原型
每个算子分为两段式接口，必须先调用“aclnnAddLayerNormGradGetWorkspaceSize”接口获取计算所需workspace大小以及包含了算子计算流程的执行器，再调用“aclnnAddLayerNormGrad”接口执行计算。
*  `aclnnStatus aclnnAddLayerNormGradGetWorkspaceSize(const aclTensor *dy, const aclTensor *x1, const aclTensor *x2, const aclTensor *rstd, const aclTensor *mean, const aclTensor *gamma, const aclTensor *dsumOptional, const aclTensor *dxOut, const aclTensor *dgammaOut, const aclTensor *dbetaOut, uint64_t *workspaceSize, aclOpExecutor **executor)`
*  `aclnnStatus aclnnAddLayerNormGrad(void *workspace, uint64_t workspaceSize, aclOpExecutor *executor, const aclrtStream stream)`

## 功能描述
-   **算子功能**：Add与LayerNorm融合算子的反向计算

-   **计算公式**：

    - 正向公式：（D为reduce轴大小）

    $$
    \operatorname{LayerNorm}(x)=\frac{x_i−\operatorname{E}(x)}{\sqrt{\operatorname{Var}(x)+ \epsilon}}*\gamma + \beta
    $$

    $$
    其中\operatorname{E}(x_i)=\frac{1}{D}\sum_{1}^{D}{x_i}
    $$

    $$
    \operatorname{Var}(x_i)=\frac{1}{D}\sum_{1}^{D}{(x_i-\operatorname{E}(x))^2}
    $$

    - 反向公式

    $$
    \frac{{\rm d}L}{{\rm d}x_i} = \sum_{j}{\frac{{\rm d}L}{{\rm d}y_j} * \gamma_j * \frac{{\rm d}\hat{x_j}}{{\rm d}x_i}}
    $$

    $$
    其中\hat{x_j}=\frac {x_i-\operatorname{E}(x)}{\sqrt{\operatorname{Var}(x)+\epsilon}}
    $$

    $$
    \frac{{\rm d}\hat{x_j}}{{\rm d}x_i}=(\delta_{ij} - \frac{{\rm d}\operatorname{E}(x)}{{\rm d}x_i}) * \frac{1}{\sqrt{\operatorname{Var}(x_i)}}-\frac{1}{\operatorname{Var}(x_i)}(x_j-\operatorname{E}(x))\frac{\rm d \operatorname{Var}(x_i)}{\rm dx}
    $$

    $$
    其中\frac{{\rm d}\operatorname{E}(x)}{{\rm d}x_i}=\frac{1}{D}
    $$

    $$
    \frac{\rm d \operatorname{Var}(x_i)}{\rm dx}=\frac{1}{D}\frac{1}{\sqrt{\operatorname{Var}(x_i)}}(x_i-\operatorname{E}(x))
    $$



## aclnnAddLayerNormGradGetWorkspaceSize

- **参数说明：**
  * dy（aclTensor \*，计算输入）：主要的grad输入。Device侧的aclTensor，数据类型支持FLOAT32，FLOAT16，BFLOAT16，shape支持2维~8维，数据格式支持ND，不支持非连续输入。

  * x1（aclTensor \*，计算输入）：为正向融合算子的输入x1。Device侧的aclTensor，数据类型支持FLOAT32，FLOAT16，BFLOAT16，shape支持2维~8维，数据格式支持ND，不支持非连续输入。

  * x2（aclTensor \*，计算输入）：为正向融合算子的输入x2。Device侧的aclTensor，数据类型支持FLOAT32，FLOAT16，BFLOAT16，shape支持2维~8维，数据格式支持ND，不支持非连续输入。

  * rstd（aclTensor \*，计算输入）：表示正向输入x1、x2之和的rstd，以下计算中inputVariace 是针对尾轴。输入数据类型支持FLOAT32，shape支持2维~8维，数据格式支持ND，不支持非连续输入。计算逻辑：
  $$
  rstd = np.power((inputVariace + EPSLON), (-0.5))
  $$

  * mean（aclTensor \*，计算输入）：表示正向输入x1、x2之和的mean，inputMean 是针对尾轴。数据类型支持FLOAT32，shape支持2维~8维，数据格式支持ND，不支持非连续输入。

  * gamma（aclTensor \*，计算输入）：数据类型支持FLOAT32，FLOAT16，BFLOAT16，shape支持1维，数据格式支持ND，不支持非连续输入。

  * dsumOptional（aclTensor \*，计算输入）：数据类型支持FLOAT32，FLOAT16，BFLOAT16，shape支持2维~8维，数据格式支持ND，不支持非连续输入。

  * dx（aclTensor \*，计算输出）：数据类型支持FLOAT32，FLOAT16，BFLOAT16，shape支持2维~8维，数据格式支持ND。

  * dgamma（aclTensor \*，计算输出）：数据类型支持FLOAT32，shape支持1维，数据格式支持ND。

  * dbeta（aclTensor \*，计算输出）：数据类型支持FLOAT32，shape支持1维，数据格式支持ND。

  * workspaceSize（uint64_t \*，出参）：返回用户需要在Device侧申请的workspace大小。

  * executor（aclOpExecutor \*\*，出参）：返回op执行器，包含了算子计算流程。

- **返回值：**

  aclnnStatus：返回状态码。（参见[aclnn返回码](common/aclnn返回码.md)）

    **说明：**
    第一段接口完成入参校验，出现以下场景时报错：

    - 161001 (ACLNN_ERR_PARAM_NULLPTR)：如果传入参数是必选输入，输出或者必选属性，且是空指针，则返回161001。


## aclnnAddLayerNormGrad
- **参数说明：**
  * workspace（void \*，入参）：在Device侧申请的workspace内存起址。
  * workspaceSize（uint64_t，入参）：在Device侧申请的workspace大小，由第一段接口aclnnAddLayerNormGradGetWorkspaceSize获取。
  * executor（aclOpExecutor \*，入参）：op执行器，包含了算子计算流程。
  * stream（aclrtStream，入参）：指定执行任务的AscendCL stream流。

- **返回值：**

  aclnnStatus：返回状态码。（具体参见[aclnn返回码](common/aclnn返回码.md)）

## 约束与限制
- **功能维度**
  * 数据类型支持
    * dy、x1、x2、gamma、dsumOptional支持：FLOAT32、FLOAT16、BFLOAT16。
    * rstd、mean支持：FLOAT32。
  * 数据格式支持：ND。
- **未支持类型说明**
  * DOUBLE：指令不支持DOUBLE。
  * 是否支持空tensor：不支持空进空出。
  * 是否非连续tensor：不支持输入非连续，不支持数据非连续。
- **边界值场景说明**
  * 当输入是inf时，输出为inf。
  * 当输入是nan时，输出为nan。

## 调用示例

示例代码如下，仅供参考：
```C++
#include <iostream>
#include <vector>
#include "acl/acl.h"
#include "aclnnop/aclnn_add_layer_norm_grad.h"

#define CHECK_RET(cond, return_expr)\
do {                                \
  if (!(cond)) {                    \
    return_expr;                    \
  }                                 \
} while (0)

#define LOG_PRINT(message, ...)   \
    do {                          \
  printf(message, ##__VA_ARGS__); \
} while (0)

int64_t GetShapeSize(const std::vector<int64_t> &shape) {
  int64_t shapeSize = 1;
  for (auto i : shape) {
    shapeSize *= i;
  }
  return shapeSize;
}

int Init(int32_t deviceId, aclrtContext *context, aclrtStream *stream) {
  // 固定写法，AscendCL初始化
  auto ret = aclInit(nullptr);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclInit failed. ERROR: %d\n", ret); return ret);
  ret = aclrtSetDevice(deviceId);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSetDevice failed. ERROR: %d\n", ret); return ret);
  ret = aclrtCreateContext(context, deviceId);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtCreateContext failed. ERROR: %d\n", ret); return ret);
  ret = aclrtSetCurrentContext(*context);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSetCurrentContext failed. ERROR: %d\n", ret); return ret);
  ret = aclrtCreateStream(stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtCreateStream failed. ERROR: %d\n", ret); return ret);
  return 0;
}

template <typename T>
int CreateAclTensor(const std::vector<T> &hostData, const std::vector<int64_t> &shape, void **deviceAddr,
                    aclDataType dataType, aclTensor **tensor) {
  auto size = GetShapeSize(shape) * sizeof(T);
  // 调用aclrtMalloc申请device侧内存
  auto ret = aclrtMalloc(deviceAddr, size, ACL_MEM_MALLOC_HUGE_FIRST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtMalloc failed. ERROR: %d\n", ret); return ret);
  // 调用aclrtMemcpy将host侧数据拷贝到device侧内存上
  ret = aclrtMemcpy(*deviceAddr, size, hostData.data(), size, ACL_MEMCPY_HOST_TO_DEVICE);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtMemcpy failed. ERROR: %d\n", ret); return ret);

  // 计算连续tensor的strides
  std::vector<int64_t> strides(shape.size(), 1);
  for (int64_t i = shape.size() - 2; i >= 0; i--) {
    strides[i] = shape[i + 1] * strides[i + 1];
  }

  // 调用aclCreateTensor接口创建aclTensor
  *tensor = aclCreateTensor(shape.data(), shape.size(), dataType, strides.data(), 0, aclFormat::ACL_FORMAT_ND,
                            shape.data(), shape.size(), *deviceAddr);
  return 0;
}

int main() {
  // 1. （固定写法）device/context/stream初始化，参考AscendCL对外接口列表
  // 根据自己的实际device填写deviceId
  int32_t deviceId = 0;
  aclrtContext context;
  aclrtStream stream;
  auto ret = Init(deviceId, &context, &stream);
  // check根据自己的需要处理
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("Init acl failed. ERROR: %d\n", ret); return ret);
  // 2. 构造输入与输出，需要根据API的接口自定义构造
  std::vector<int64_t> dyShape = {3, 1, 4};
  std::vector<int64_t> dx_1Shape = {3, 1, 4};
  std::vector<int64_t> dx_2Shape = {3, 1, 4};
  std::vector<int64_t> rstdShape = {3, 1, 1};
  std::vector<int64_t> meanShape = {3, 1, 1};
  std::vector<int64_t> gammaShape = {4};
  std::vector<int64_t> dsumOptionalShape = {3, 1, 4};
  std::vector<int64_t> outputpdxShape = {3, 1, 4};
  std::vector<int64_t> outputpdgammaShape = {4};
  std::vector<int64_t> outputpdbetaShape = {4};
  void *dyDeviceAddr = nullptr;
  void *dx_1DeviceAddr = nullptr;
  void *dx_2DeviceAddr = nullptr;
  void *rstdDeviceAddr = nullptr;
  void *meanDeviceAddr = nullptr;
  void *gammaDeviceAddr = nullptr;
  void *dsumOptionalDeviceAddr = nullptr;
  void *outputpdxDeviceAddr = nullptr;
  void *outputpdgammaDeviceAddr = nullptr;
  void *outputpdbetaDeviceAddr = nullptr;
  aclTensor *dy = nullptr;
  aclTensor *dx_1 = nullptr;
  aclTensor *dx_2 = nullptr;
  aclTensor *rstd = nullptr;
  aclTensor *mean = nullptr;
  aclTensor *gamma = nullptr;
  aclTensor *dsumOptional = nullptr;
  aclTensor *outputpdx = nullptr;
  aclTensor *outputpdgamma = nullptr;
  aclTensor *outputpdbeta = nullptr;
  std::vector<float> dyHostData = {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12};
  std::vector<int32_t> dx_1HostData = {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12};
  std::vector<int32_t> dx_2HostData = {2, 2, 2, 4, 4, 4, 6, 6, 6, 8, 8, 8};
  std::vector<int32_t> rstdHostData = {0, 1, 2};
  std::vector<int32_t> meanHostData = {0, 1, 2};
  std::vector<int32_t> gammaHostData = {0, 1, 2, 3};
  std::vector<int32_t> dsumOptionalHostData = {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12};
  std::vector<int32_t> outputpdxHostData = {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12};
  std::vector<int32_t> outputpdgammaHostData = {0, 1, 2, 3};
  std::vector<int32_t> outputpdbetaHostData = {0, 1, 2, 3};

  // 创建self aclTensor
  ret = CreateAclTensor(dyHostData, dyShape, &dyDeviceAddr, aclDataType::ACL_FLOAT, &dy);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  ret = CreateAclTensor(dx_1HostData, dx_1Shape, &dx_1DeviceAddr, aclDataType::ACL_FLOAT, &dx_1);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  ret = CreateAclTensor(dx_2HostData, dx_2Shape, &dx_2DeviceAddr, aclDataType::ACL_FLOAT, &dx_2);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  ret = CreateAclTensor(rstdHostData, rstdShape, &rstdDeviceAddr, aclDataType::ACL_FLOAT, &rstd);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  ret = CreateAclTensor(meanHostData, meanShape, &meanDeviceAddr, aclDataType::ACL_FLOAT, &mean);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  ret = CreateAclTensor(gammaHostData, gammaShape, &gammaDeviceAddr, aclDataType::ACL_FLOAT, &gamma);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  ret = CreateAclTensor(dsumOptionalHostData, dsumOptionalShape, &dsumOptionalDeviceAddr, aclDataType::ACL_FLOAT,
                        &dsumOptional);
  CHECK_RET(ret == ACL_SUCCESS, return ret);

  ret = CreateAclTensor(outputpdxHostData, outputpdxShape, &outputpdxDeviceAddr, aclDataType::ACL_FLOAT, &outputpdx);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  ret = CreateAclTensor(outputpdgammaHostData, outputpdgammaShape, &outputpdgammaDeviceAddr, aclDataType::ACL_FLOAT,
                        &outputpdgamma);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  ret = CreateAclTensor(outputpdbetaHostData, outputpdbetaShape, &outputpdbetaDeviceAddr, aclDataType::ACL_FLOAT,
                        &outputpdbeta);
  CHECK_RET(ret == ACL_SUCCESS, return ret);

  uint64_t workspaceSize = 0;
  aclOpExecutor *executor;

  // aclnnAddLayerNormGrad接口调用示例
  // 3. 调用CANN算子库API，需要修改为具体的Api名称
  // 调用aclnnAddLayerNormGrad第一段接口
  LOG_PRINT("\nUse aclnnAddLayerNormGrad Port.");
  ret = aclnnAddLayerNormGradGetWorkspaceSize(dy, dx_1, dx_2, rstd, mean, gamma, dsumOptional, outputpdx,
                                              outputpdgamma, outputpdbeta,
                                              &workspaceSize, &executor);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclnnAddLayerNormGradGetWorkspaceSize failed. ERROR: %d\n", ret);
            return ret);
  // 根据第一段接口计算出的workspaceSize申请device内存
  void *workspaceAddr = nullptr;
  if (workspaceSize > 0) {
    ret = aclrtMalloc(&workspaceAddr, workspaceSize, ACL_MEM_MALLOC_HUGE_FIRST);
    CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("allocate workspace failed. ERROR: %d\n", ret); return ret;);
  }
  // 调用aclnnAddLayerNormGrad第二段接口
  ret = aclnnAddLayerNormGrad(workspaceAddr, workspaceSize, executor, stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclnnAddLayerNormGrad failed. ERROR: %d\n", ret); return ret);

  // 4. （固定写法）同步等待任务执行结束
  ret = aclrtSynchronizeStream(stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSynchronizeStream failed. ERROR: %d\n", ret); return ret);

  // 5. 获取输出的值，将device侧内存上的结果拷贝至host侧，需要根据具体API的接口定义修改
  auto outputpdxsize = GetShapeSize(outputpdxShape);
  std::vector<float> resultDataPdx(outputpdxsize, 0);
  ret = aclrtMemcpy(resultDataPdx.data(), resultDataPdx.size() * sizeof(resultDataPdx[0]), outputpdxDeviceAddr,
                    outputpdxsize * sizeof(resultDataPdx[0]), ACL_MEMCPY_DEVICE_TO_HOST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("copy result from device to host failed. ERROR: %d\n", ret); return ret);
  LOG_PRINT("== pdx output");
  for (int64_t i = 0; i < outputpdxsize; i++) {
    LOG_PRINT("result[%ld] is: %f\n", i, resultDataPdx[i]);
  }
  auto outputpdgammasize = GetShapeSize(outputpdgammaShape);
  std::vector<float> resultDataPdGamma(outputpdgammasize, 0);
  ret =
      aclrtMemcpy(resultDataPdGamma.data(), resultDataPdGamma.size() * sizeof(resultDataPdGamma[0]),
                  outputpdgammaDeviceAddr, outputpdgammasize * sizeof(resultDataPdGamma[0]), ACL_MEMCPY_DEVICE_TO_HOST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("copy result from device to host failed. ERROR: %d\n", ret); return ret);
  LOG_PRINT("== pdgamma output");
  for (int64_t i = 0; i < outputpdgammasize; i++) {
    LOG_PRINT("result[%ld] is: %f\n", i, resultDataPdGamma[i]);
  }
  auto outputpdbetasize = GetShapeSize(outputpdbetaShape);
  std::vector<float> resultDataPdBeta(outputpdbetasize, 0);
  ret = aclrtMemcpy(resultDataPdBeta.data(), resultDataPdBeta.size() * sizeof(resultDataPdBeta[0]),
                    outputpdbetaDeviceAddr, outputpdbetasize * sizeof(resultDataPdBeta[0]), ACL_MEMCPY_DEVICE_TO_HOST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("copy result from device to host failed. ERROR: %d\n", ret); return ret);
  LOG_PRINT("== pdbeta output");
  for (int64_t i = 0; i < outputpdbetasize; i++) {
    LOG_PRINT("result[%ld] is: %f\n", i, resultDataPdBeta[i]);
  }

  // 6. 释放aclTensor和aclScalar，需要根据具体API的接口定义修改
  aclDestroyTensor(dy);
  aclDestroyTensor(dx_1);
  aclDestroyTensor(dx_2);
  aclDestroyTensor(rstd);
  aclDestroyTensor(mean);
  aclDestroyTensor(gamma);
  aclDestroyTensor(dsumOptional);
  aclDestroyTensor(outputpdx);
  aclDestroyTensor(outputpdgamma);
  aclDestroyTensor(outputpdbeta);

  // 7. 释放device资源，需要根据具体API的接口定义修改
  aclrtFree(dyDeviceAddr);
  aclrtFree(dx_1DeviceAddr);
  aclrtFree(dx_2DeviceAddr);
  aclrtFree(rstdDeviceAddr);
  aclrtFree(meanDeviceAddr);
  aclrtFree(gammaDeviceAddr);
  aclrtFree(dsumOptionalDeviceAddr);
  aclrtFree(outputpdxDeviceAddr);
  aclrtFree(outputpdgammaDeviceAddr);
  aclrtFree(outputpdbetaDeviceAddr);
  if (workspaceSize > 0) {
    aclrtFree(workspaceAddr);
  }
  aclrtDestroyStream(stream);
  aclrtDestroyContext(context);
  aclrtResetDevice(deviceId);
  aclFinalize();
  return 0;
}
```
