# aclnnAddbmm&aclnnInplaceAddbmm

## 支持的产品型号
- Atlas 训练产品。
- Atlas A2训练系列产品。

## 接口原型
- aclnnAddbmm和aclnnInplaceAddbmm实现相同的功能，使用区别如下，请根据自身实际场景选择合适的算子。
  - aclnnAddbmm：需新建一个输出张量对象存储计算结果。
  - aclnnInplaceAddbmm：无需新建输出张量对象，直接在输入张量的内存中存储计算结果。
- 每个算子分为[两段式接口](common/两段式接口.md)，必须先调用“aclnnAddbmmGetWorkspaceSize”或者”aclnnInplaceAddbmmGetWorkspaceSize“接口获取计算所需workspace大小以及包含了算子计算流程的执行器，再调用“aclnnAddbmm”或者”aclnnInplaceAddbmm“接口执行计算。

  * `aclnnStatus aclnnAddbmmGetWorkspaceSize(const aclTensor *self, const aclTensor *batch1, const aclTensor *batch2, const aclScalar *beta, const aclScalar *alpha, aclTensor *out, int8_t cubeMathType, uint64_t *workspaceSize, aclOpExecutor **executor)`
  * `aclnnStatus aclnnAddbmm(void *workspace, uint64_t workspaceSize, aclOpExecutor *executor, aclrtStream stream)`
  * `aclnnstatus aclnnInplaceAddbmmGetWorkspaceSize(aclTensor *selfRef, const aclTensor *batch1, const aclTensor *batch2, const aclScalar *beta, const aclScalar *alpha, int8_t cubeMathType, uint64_t *workspaceSize, aclOpExecutor **executor)`
  * `aclnnstatus aclnnInplaceAddbmm(void *workspace, uint64_t workspaceSize, aclOpExecutor *executor, aclrtStream stream)`

## 功能描述

- 算子功能：首先进行batch1、batch2的矩阵乘计算，然后将该结果按照第一维（batch维度）批处理相加，将三维向量压缩为二维向量（shape大小为后两维的shape），然后该结果与α作乘积计算，再与β和self的乘积求和得到结果。

- 计算公式：

  $$
  out = βself+α(\sum_{i=0}^{b-1} batch1_{i}@batch2_{i})
  $$

  注意：如果β为0，则self会被忽略，不参与计算。

- 示例：

  self的shape是[3,5],batch1的shape是[10,3,4],batch2的shape是[10,4,5]，计算输出out的shape是[3,5]。

## aclnnAddbmmGetWorkspaceSize

- **参数说明：**
  
  * self(aclTensor \*，计算输入): 公式中的输入`self`，Device侧的aclTensor，数据类型支持FLOAT、FLOAT16、BFLOAT16（仅Atlas A2训练系列产品支持），且数据类型与batch1@batch2的数据类型需满足数据类型推导规则（参见[互推导关系](common/互推导关系.md)），shape需要与batch1@batch2的后两维的shape满足[broadcast关系](common/broadcast关系.md)（注意：只能是self通过broadcast变成和batch1@batch2的后两维一样的shape，不可以batch1@batch2的后两维通过broadcast变成和self一样，举例：self:[3,5]，batch1@batch2:[1,1,1]，会失败）。支持[非连续的Tensor](common/非连续的Tensor.md)，支持空Tensor，[数据格式](common/数据格式.md)支持ND。
  * batch1(aclTensor \*，计算输入): 公式中的输入`batch1`，Device侧的aclTensor，数据类型支持FLOAT、FLOAT16、BFLOAT16（仅Atlas A2训练系列产品支持），且数据类型与self、batch2的数据类型需满足数据类型推导规则（参见[互推导关系](common/互推导关系.md)），shape需要与batch2满足bmm输入约束条件）。支持[非连续的Tensor](common/非连续的Tensor.md)，支持[空Tensor](#)，[数据格式](common/数据格式.md)支持ND。
  * batch2(aclTensor \*，计算输入): 公式中的输入`batch2`，Device侧的aclTensor，数据类型支持FLOAT、FLOAT16、BFLOAT16（仅Atlas A2训练系列产品支持），且数据类型与self、batch1的数据类型需满足数据类型推导规则（参见[互推导关系](common/互推导关系.md)），shape需要与batch1满足bmm输入约束条件）。支持[非连续的Tensor](common/非连续的Tensor.md)，支持[空Tensor](#)，[数据格式](common/数据格式.md)支持ND。
  * beta(aclScalar \*，计算输入):公式中的`β`，Host侧的aclScalar，数据类型支持FLOAT、FLOAT16、DOUBLE、INT32、INT64、INT16、INT8、UINT8、BOOL，且数据类型需要可转换成self与batch1@batch2推导后的数据类型。
  * alpha(aclScalar \*，计算输入):公式中的`α`，Host侧的aclScalar，数据类型支持FLOAT、FLOAT16、DOUBLE、INT32、INT64、INT16、INT8、UINT8、BOOL，且数据类型需要可转换成self与batch1@batch2推导后的数据类型。
  * out(aclTensor \*，计算输出): 公式中的输出`out`，Device侧的aclTensor，数据类型支持FLOAT、FLOAT16、BFLOAT16（仅Atlas A2训练系列产品支持），且数据类型与self保持一致，shape要求与batch1@batch2的后两维shape保持一致。支持[非连续的Tensor](common/非连续的Tensor.md)，支持空Tensor，[数据格式](common/数据格式.md)支持ND。
  * cubeMathType(int8_t \*，计算输入): Host侧的整型，判断Cube单元应该使用那种计算逻辑进行运算，数据类型支持INT8，具体的枚举值如下：
    * 0:KEEP_DTYE，保持输入的数据类型进行计算。当输入是FLOAT，Atlas 训练产品Cube计算单元暂不支持，取0时会报错。
    * 1:ALLOW_FP32_DOWN_PRECISON，允许将输入数据降精度计算。当输入是FLOAT，Atlas 训练产品转换为FLOAT16计算，Atlas A2训练系列产品转换为HFLOAT32计算。
    * 2:USE_FP16，允许转换为数据类型FLOAT16进行计算。当输入数据类型是FLOAT，转换为FLOAT16计算。
    * 3:USE_HF32，允许转换为数据类型HFLOAT32计算。当输入是FLOAT，Atlas 训练产品Cube计算单元暂不支持，取3时会报错，Atlas A2训练系列产品转换为HFLOAT32计算。
  * workspaceSize(uint64_t \*，出参): 返回用户需要在Device侧申请的workspace大小。
  * executor(aclOpExecutor \*\*，出参): 返回op执行器，包含了算子计算流程。
  
- **返回值：**

  aclnnStatus: 返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

  ```
  161001(ACLNN_ERR_PARAM_NULLPTR): 1. 传入的self、batch1或batch2是空指针。
  161002(ACLNN_ERR_PARAM_INVALID): 1. self、batch1、batch2或out的数据类型不在支持的范围内。
                                   2. self与out的数据类型不一致。
                                   3. self不能与batch1@batch2做broadcast操作。
                                   4. self与batch1的dtype不同且self为FLOAT16，batch1为FLOAT32。
                                   5. batch1和batch2的第一维度不相等。
                                   6. batch1和batch2的维度不是三维。
                                   7. batch1的最后一维和batch2的倒数第二维不相等。
  ```

## aclnnAddbmm

- **参数说明：**
  
  * workspace(void \*，入参): 在Device侧申请的workspace内存起址。
  * workspaceSize(uint64_t，入参): 在Device侧申请的workspace大小，由第一段接口aclnnAddbmmGetWorkspaceSize获取。
  * executor(aclOpExecutor \*，入参):指定执行任务的 AscendCL Stream流。
  * stream(aclrtStream，入参): op执行器，包含了算子计算流程。
  
- **返回值：**

  aclnnStatus: 返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

## aclnnInplaceAddbmmGetWorkspaceSize

- **参数说明：**
  
  * selfRef(aclTensor \*，计算输入|计算输出): 输入输出tensor，即公式中的输入`self`与`out`，Device侧的aclTensor，数据类型支持FLOAT、FLOAT16、BFLOAT16（仅Atlas A2训练系列产品支持），且数据类型与batch1@batch2的数据类型需满足数据类型推导规则（参见[互推导关系](common/互推导关系.md)），shape需要与batch1@batch2的后两维的shape满足[broadcast关系](common/broadcast关系.md)（注意：只能是self通过broadcast变成和batch1@batch2的后两维一样的shape，不可以batch1@batch2的后两维通过broadcast变成和selfRef一样，举例：selfRef:[3,5]，batch1@batch2:[1,1,1]，会失败）。支持[非连续的Tensor](common/非连续的Tensor.md)，支持[空Tensor](#)，[数据格式](common/数据格式.md)支持ND。
  * batch1(aclTensor \*，计算输入): 公式中的输入`batch1`，Device侧的aclTensor，数据类型支持FLOAT、FLOAT16、BFLOAT16（仅Atlas A2训练系列产品支持），且数据类型与selfRef、batch2的数据类型需满足数据类型推导规则（参见[互推导关系](common/互推导关系.md)），shape需要与batch2满足bmm输入约束条件）。支持[非连续的Tensor](common/非连续的Tensor.md)，支持[空Tensor](#)，[数据格式](common/数据格式.md)支持ND。
  * batch2(aclTensor \*，计算输入): 公式中的输入`batch2`，Device侧的aclTensor，数据类型支持FLOAT、FLOAT16、BFLOAT16（仅Atlas A2训练系列产品支持），且数据类型与selfRef、batch1的数据类型需满足数据类型推导规则（参见[互推导关系](common/互推导关系.md)），shape需要与batch1满足bmm输入约束条件）。支持[非连续的Tensor](common/非连续的Tensor.md)，支持[空Tensor](#)，[数据格式](common/数据格式.md)支持ND。
  * beta(aclScalar \*，计算输入):公式中的`β`，Host侧的aclScalar，数据类型支持FLOAT、FLOAT16、DOUBLE、INT32、INT64、INT16、INT8、UINT8、BOOL，且数据类型需要可转换成selfRef与batch1@batch2推导后的数据类型。
  * alpha(aclScalar \*，计算输入):公式中的`α`，Host侧的aclScalar，数据类型支持FLOAT、FLOAT16、DOUBLE、INT32、INT64、INT16、INT8、UINT8、BOOL，且数据类型需要可转换成selfRef与batch1@batch2推导后的数据类型。
  * cubeMathType(int8_t \*，计算输入): Host侧的整型，判断Cube单元应该使用那种计算逻辑进行运算，数据类型支持INT8，具体的枚举值如下：
    * 0:KEEP_DTYE，保持输入的数据类型进行计算。当输入是FLOAT，Atlas 训练产品Cube计算单元暂不支持，取0时会报错。
    * 1:ALLOW_FP32_DOWN_PRECISON，允许将输入数据降精度计算。当输入是FLOAT，Atlas 训练产品转换为FLOAT16计算，Atlas A2训练系列产品转换为HFLOAT32计算。
    * 2:USE_FP16，使用数据类型FLOAT16进行计算。当输入数据类型是FLOAT，转换为FLOAT16计算。
    * 3:USE_HF32，使用数据类型HFLOAT32计算。当输入是FLOAT，Atlas 训练产品Cube计算单元暂不支持，取3时会报错，Atlas A2训练系列产品转换为HFLOAT32计算。
  * workspaceSize**(uint64_t \*，出参): 返回用户需要在Device侧申请的workspace大小。
  * executor(aclOpExecutor \*\*，出参): 返回op执行器，包含了算子计算流程。
  
- **返回值：**

  aclnnStatus: 返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

  ```
  161001(ACLNN_ERR_PARAM_NULLPTR): 1. 传入的selfRef、batch1或batch2是空指针。
  161002(ACLNN_ERR_PARAM_INVALID): 1. selfRef、batch1或batch2的数据类型不在支持的范围内。
                                   2. selfRef与out的数据类型不一致。
                                   3. selfRef不能与batch1@batch2做broadcast操作。
                                   4. selfRef与batch1的dtype不同且selfRef为FLOAT16，batch1为FLOAT32。
                                   5. batch1和batch2的第一维度不相等。
                                   6. batch1和batch2的维度不是三维。
                                   7. batch1的最后一维和batch2的倒数第二维不相等。
  ```

## aclnnInplaceAddbmm

- **参数说明：**
  
  * workspace(void \*，入参): 在Device侧申请的workspace内存起址。
  * workspaceSize(uint64_t)，入参: 在Device侧申请的workspace大小，由第一段接口aclnnInplaceAddbmmGetWorkspaceSize获取。
  * executor(aclOpExecutor \*，入参): 指定执行任务的 AscendCL Stream流。
  * stream(aclrtStream，入参): op执行器，包含了算子计算流程。
  
- **返回值：**

  aclnnStatus: 返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

## 约束与限制
对于Atlas 训练产品，Cube单元不支持FLOAT32计算。当输入为FLOAT32，可通过设置cubeMathType=1（ALLOW_FP32_DOWN_PRECISON）来允许接口内部cast到FLOAT16进行计算。

## 调用示例
示例代码如下，仅供参考，具体编译和执行过程请参考[编译与运行样例](common/编译与运行样例.md)。
```c++
#include <iostream>
#include <vector>
#include "acl/acl.h"
#include "aclnnop/aclnn_addbmm.h"

#define CHECK_RET(cond, return_expr) \
  do {                               \
    if (!(cond)) {                   \
      return_expr;                   \
    }                                \
  } while (0)

#define LOG_PRINT(message, ...)     \
  do {                              \
    printf(message, ##__VA_ARGS__); \
  } while (0)

int64_t GetShapeSize(const std::vector<int64_t>& shape) {
  int64_t shapeSize = 1;
  for (auto i : shape) {
    shapeSize *= i;
  }
  return shapeSize;
}

int Init(int32_t deviceId, aclrtContext* context, aclrtStream* stream) {
  // 固定写法，AscendCL初始化
  auto ret = aclInit(nullptr);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclInit failed. ERROR: %d\n", ret); return ret);
  ret = aclrtSetDevice(deviceId);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSetDevice failed. ERROR: %d\n", ret); return ret);
  ret = aclrtCreateContext(context, deviceId);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtCreateContext failed. ERROR: %d\n", ret); return ret);
  ret = aclrtSetCurrentContext(*context);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSetCurrentContext failed. ERROR: %d\n", ret); return ret);
  ret = aclrtCreateStream(stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtCreateStream failed. ERROR: %d\n", ret); return ret);
  return 0;
}

template <typename T>
int CreateAclTensor(const std::vector<T>& hostData, const std::vector<int64_t>& shape, void** deviceAddr,
                    aclDataType dataType, aclTensor** tensor) {
  auto size = GetShapeSize(shape) * sizeof(T);
  // 调用aclrtMalloc申请device侧内存
  auto ret = aclrtMalloc(deviceAddr, size, ACL_MEM_MALLOC_HUGE_FIRST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtMalloc failed. ERROR: %d\n", ret); return ret);
  // 调用aclrtMemcpy将host侧数据拷贝到device侧内存上
  ret = aclrtMemcpy(*deviceAddr, size, hostData.data(), size, ACL_MEMCPY_HOST_TO_DEVICE);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtMemcpy failed. ERROR: %d\n", ret); return ret);

  // 计算连续tensor的strides
  std::vector<int64_t> strides(shape.size(), 1);
  for (int64_t i = shape.size() - 2; i >= 0; i--) {
    strides[i] = shape[i + 1] * strides[i + 1];
  }

  // 调用aclCreateTensor接口创建aclTensor
  *tensor = aclCreateTensor(shape.data(), shape.size(), dataType, strides.data(), 0, aclFormat::ACL_FORMAT_ND,
                            shape.data(), shape.size(), *deviceAddr);
  return 0;
}

int main() {
  // 1. （固定写法）device/context/stream初始化，参考AscendCL对外接口列表
  // 根据自己的实际device填写deviceId
  int32_t deviceId = 0;
  aclrtContext context;
  aclrtStream stream;
  auto ret = Init(deviceId, &context, &stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("Init acl failed. ERROR: %d\n", ret); return ret);

  // 2. 构造输入与输出，需要根据API的接口自定义构造
  std::vector<int64_t> selfShape = {1, 2, 4};
  std::vector<int64_t> batch1Shape = {1, 2, 3};
  std::vector<int64_t> batch2Shape = {1, 3, 4};
  std::vector<int64_t> outShape = {1, 2, 4};
  void* selfDeviceAddr = nullptr;
  void* batch1DeviceAddr = nullptr;
  void* batch2DeviceAddr = nullptr;
  void* outDeviceAddr = nullptr;
  aclTensor* self = nullptr;
  aclTensor* batch1 = nullptr;
  aclTensor* batch2 = nullptr;
  aclScalar* alpha = nullptr;
  aclScalar* beta = nullptr;
  aclTensor* out = nullptr;
  std::vector<float> selfHostData = {0, 1, 2, 3, 4, 5, 6, 7};
  std::vector<float> batch1HostData = {1, 1, 1, 2, 2, 2};
  std::vector<float> batch2HostData = {1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4};
  std::vector<float> outHostData(8, 0);
  int8_t cubeMathType = 1;
  float alphaValue = 1.2f;
  float betaValue = 1.0f;
  // 创建self aclTensor
  ret = CreateAclTensor(selfHostData, selfShape, &selfDeviceAddr, aclDataType::ACL_FLOAT, &self);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  // 创建batch1 aclTensor
  ret = CreateAclTensor(batch1HostData, batch1Shape, &batch1DeviceAddr, aclDataType::ACL_FLOAT, &batch1);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  // 创建batch2 aclTensor
  ret = CreateAclTensor(batch2HostData, batch2Shape, &batch2DeviceAddr, aclDataType::ACL_FLOAT, &batch2);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  // 创建alpha aclScalar
  alpha = aclCreateScalar(&alphaValue,aclDataType::ACL_FLOAT);
  CHECK_RET(alpha != nullptr, return ret);
  // 创建beta aclScalar
  beta = aclCreateScalar(&betaValue,aclDataType::ACL_FLOAT);
  CHECK_RET(beta != nullptr, return ret);
  // 创建out aclTensor
  ret = CreateAclTensor(outHostData, outShape, &outDeviceAddr, aclDataType::ACL_FLOAT, &out);
  CHECK_RET(ret == ACL_SUCCESS, return ret);

  // 3. 调用CANN算子库API，需要修改为具体的Api名称
  uint64_t workspaceSize = 0;
  aclOpExecutor* executor;
  // 调用aclnnAddbmm第一段接口
  ret = aclnnAddbmmGetWorkspaceSize(self, batch1, batch2, alpha, beta, out, cubeMathType, &workspaceSize, &executor);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclnnAddbmmGetWorkspaceSize failed. ERROR: %d\n", ret); return ret);
  // 根据第一段接口计算出的workspaceSize申请device内存
  void* workspaceAddr = nullptr;
  if (workspaceSize > 0) {
    ret = aclrtMalloc(&workspaceAddr, workspaceSize, ACL_MEM_MALLOC_HUGE_FIRST);
    CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("allocate workspace failed. ERROR: %d\n", ret); return ret);
  }
  // 调用aclnnAddbmm第二段接口
  ret = aclnnAddbmm(workspaceAddr, workspaceSize, executor, stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclnnAddbmm failed. ERROR: %d\n", ret); return ret);

  // 4. （固定写法）同步等待任务执行结束
  ret = aclrtSynchronizeStream(stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSynchronizeStream failed. ERROR: %d\n", ret); return ret);

  // 5. 获取输出的值，将device侧内存上的结果拷贝至host侧，需要根据具体API的接口定义修改
  auto size = GetShapeSize(outShape);
  std::vector<float> resultData(size, 0);
  ret = aclrtMemcpy(resultData.data(), resultData.size() * sizeof(resultData[0]), outDeviceAddr,
                    size * sizeof(resultData[0]), ACL_MEMCPY_DEVICE_TO_HOST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("copy result from device to host failed. ERROR: %d\n", ret); return ret);
  for (int64_t i = 0; i < size; i++) {
    LOG_PRINT("result[%ld] is: %f\n", i, resultData[i]);
  }

  // aclnnInplaceAddbmm
  // step3 调用CANN算子库API
  LOG_PRINT("\ntest aclnnInplaceAddbmm\n");
  // 调用aclnnInplaceAddbmm第一段接口
  ret = aclnnInplaceAddbmmGetWorkspaceSize(self, batch1, batch2, alpha, beta, cubeMathType, &workspaceSize, &executor);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclnnInplaceAddbmmGetWorkspaceSize failed. ERROR: %d\n", ret); return ret);
  // 根据第一段接口计算出的workspaceSize申请device内存
  if (workspaceSize > 0) {
    ret = aclrtMalloc(&workspaceAddr, workspaceSize, ACL_MEM_MALLOC_HUGE_FIRST);
    CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("allocate workspace failed. ERROR: %d\n", ret); return ret);
  }
  // 调用aclnnInplaceAddbmm第二段接口
  ret = aclnnInplaceAddbmm(workspaceAddr, workspaceSize, executor, stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclnnInplaceAddbmm failed. ERROR: %d\n", ret); return ret);

  // step4（固定写法）同步等待任务执行结束
  ret = aclrtSynchronizeStream(stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSynchronizeStream failed. ERROR: %d\n", ret); return ret);

  // step5 获取输出的值，将device侧内存上的结果拷贝至host侧
  ret = aclrtMemcpy(resultData.data(), resultData.size() * sizeof(resultData[0]), selfDeviceAddr,
                    size * sizeof(resultData[0]), ACL_MEMCPY_DEVICE_TO_HOST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("copy result from device to host failed. ERROR: %d\n", ret); return ret);
  for (int64_t i = 0; i < size; i++) {
    LOG_PRINT("result[%ld] is: %f\n", i, resultData[i]);
  }

  // 6. 释放aclTensor和aclScalar，需要根据具体API的接口定义修改
  aclDestroyTensor(self);
  aclDestroyTensor(batch1);
  aclDestroyTensor(batch2);
  aclDestroyScalar(alpha);
  aclDestroyScalar(beta);
  aclDestroyTensor(out);

  // 7.释放device资源，需要根据具体API的接口定义修改
  aclrtFree(selfDeviceAddr);
  aclrtFree(batch1DeviceAddr);
  aclrtFree(batch2DeviceAddr);
  aclrtFree(outDeviceAddr);
  if (workspaceSize > 0) {
    aclrtFree(workspaceAddr);
  }
  aclrtFree(stream);
  aclrtFree(context);
  aclrtFree(deviceId);
  aclFinalize();

  return 0;
}
```
