# aclnnAddmv

## 支持的产品型号
- Atlas 推理系列产品（Ascend 310P处理器）。
- Atlas 训练产品。
- Atlas A2训练系列产品。

## 接口原型
每个算子分为[两段式接口](common/两段式接口.md)，必须先调用“aclnnAddmvGetWorkspaceSize”接口获取计算所需workspace大小以及包含了算子计算流程的执行器，再调用“aclnnAddmv”接口执行计算。

* `aclnnStatus aclnnAddmvGetWorkspaceSize(const aclTensor *self, const aclTensor *mat, const aclTensor *vec, const aclScalar *alpha, const aclScalar *beta, aclTensor *out, int8_t cubeMathType, uint64_t *workspaceSize, aclOpExecutor **executor)`
* `aclnnStatus aclnnAddmv(void *workspace, uint64_t workspaceSize, aclOpExecutor *executor, aclrtStream stream)`

## 功能描述

- 算子功能：完成矩阵乘向量，然后和向量相加。
- 计算公式：

$$
out = β  self + α  (mat @ vec)
$$

## aclnnAddmvGetWorkspaceSize

- **参数说明：**

  * self(aclTensor*, 计算输入)：Device侧的aclTensor，需要和后续乘法结果相加的1维向量，数据类型支持FLOAT16、FLOAT、INT32、INT64、INT16、INT8、UINT8、DOUBLE、BOOL，且数据类型需要与$mat@vec$构成互相推导关系（参见[互推导关系](common/互推导关系.md)），shape需要与 $mat1@vec$ 满足[broadcast关系](common/broadcast关系.md)。支持[非连续的Tensor](common/非连续的Tensor.md)，[数据格式](common/数据格式.md)支持ND。
  * mat(aclTensor*, 计算输入)：Device侧的aclTensor，和vec进行乘法运算的2维矩阵，数据类型支持FLOAT16、FLOAT、INT32、INT64、INT16、INT8、UINT8、DOUBLE、BOOL，数据类型需要与self构成互相推导关系（参见[互推导关系](common/互推导关系.md)），shape需要与 vec 满足乘法关系。支持[非连续的Tensor](common/非连续的Tensor.md)，[数据格式](common/数据格式.md)支持ND。
  * vec(aclTensor*, 计算输入)：Device侧的aclTensor，和mat进行乘法运算的1维向量，数据类型支持FLOAT16、FLOAT、INT32、INT64、INT16、INT8、UINT8、DOUBLE、BOOL，数据类型需要与self构成互相推导关系（参见[互推导关系](common/互推导关系.md)），shape需要与 mat 满足乘法关系。支持[非连续的Tensor](common/非连续的Tensor.md)，[数据格式](common/数据格式.md)支持ND。
  * alpha(aclScalar*, 计算输入)：Host侧的aclScalar，mat和vec乘积的系数，默认为1。
  * beta(aclScalar*, 计算输入)：Host侧的aclScalar，self的系数，默认为1。
  * out(aclTensor*, 计算输出)：Device侧的aclTensor，指定的1维输出向量；数据类型支持FLOAT16、FLOAT、INT32、INT64、INT16、INT8、UINT8、DOUBLE、BOOL，且数据类型需要与self构成互相推导关系（参见[互推导关系](common/互推导关系.md)），shape与mat和vec的乘积相同，[数据格式](common/数据格式.md)支持ND。
  * cubeMathType(int8_t \*，计算输入): Host侧的整型，判断Cube单元应该使用那种计算逻辑进行运算，数据类型支持INT8，具体的枚举值如下：
    * 0:KEEP_DTYE，保持输入的数据类型进行计算。当输入是FLOAT，Atlas 训练产品和Atlas 推理系列产品（Ascend 310P处理器）暂不支持，取0时会报错。
    * 1:ALLOW_FP32_DOWN_PRECISON，允许将输入数据降精度计算。当输入是FLOAT，Atlas 训练产品和Atlas 推理系列产品（Ascend 310P处理器）允许转换为FLOAT16计算，Atlas A2训练系列产品允许转换为HFLOAT32计算。
    * 2:USE_FP16，允许转换为数据类型FLOAT16进行计算。当输入数据类型是FLOAT，转换为FLOAT16计算。
    * 3:USE_HF32，允许转换为数据类型HFLOAT32计算。当输入是FLOAT，Atlas 训练产品和Atlas 推理系列产品（Ascend 310P处理器）暂不支持，取3时会报错，Atlas A2训练系列产品转换为HFLOAT32计算。
  * workspaceSize：返回用户需要在Device侧申请的workspace大小。
  * executor：返回op执行器，包含了算子计算流程。

- **返回值：**

  aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

  ```
  第一段接口完成入参校验，出现如下场景时报错：
  ACLNN_ERR_PARAM_NULLPTR: 1. 传入的self、mat、vec、alpha、beta或out是空指针。
  ACLNN_ERR_PARAM_INVALID: 1. self和mat、vec的数据类型和数据格式不在支持的范围之内。
                           2. self和mat、vec无法做数据类型推导。
                           3. 推导出的数据类型/数据格式无法转换为指定输出out的类型/格式。
                           4. mat、vec的shape不满足乘法运算条件或者self和乘法运算结果不满足加法运算条件。
  ```

## aclnnAddmv

- **参数说明：**

  * workspace：在Device侧申请的workspace内存起址。
  * workspaceSize：在Device侧申请的workspace大小，由第一段接口aclnnAddmvGetWorkspaceSize获取。
  * executor：op执行器，包含了算子计算流程。
  * stream：指定执行任务的 AscendCL Stream流。

- **返回值：**

  aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

## 约束与限制
对于Atlas 训练产品，Cube单元不支持FLOAT32计算。当输入为FLOAT32，可通过设置cubeMathType=1（ALLOW_FP32_DOWN_PRECISON）来允许接口内部cast到FLOAT16进行计算。

## 调用示例
示例代码如下，仅供参考，具体编译和执行过程请参考[编译与运行样例](common/编译与运行样例.md)。
```c++
#include <iostream>
#include <vector>
#include "acl/acl.h"
#include "aclnnop/aclnn_addmv.h"

#define CHECK_RET(cond, return_expr) \
  do {                               \
    if (!(cond)) {                   \
      return_expr;                   \
    }                                \
  } while (0)

#define LOG_PRINT(message, ...)     \
  do {                              \
    printf(message, ##__VA_ARGS__); \
  } while (0)

int64_t GetShapeSize(const std::vector<int64_t>& shape) {
  int64_t shapeSize = 1;
  for (auto i : shape) {
    shapeSize *= i;
  }
  return shapeSize;
}

int Init(int32_t deviceId, aclrtContext* context, aclrtStream* stream) {
  // 固定写法，AscendCL初始化
  auto ret = aclInit(nullptr);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclInit failed. ERROR: %d\n", ret); return ret);
  ret = aclrtSetDevice(deviceId);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSetDevice failed. ERROR: %d\n", ret); return ret);
  ret = aclrtCreateContext(context, deviceId);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtCreateContext failed. ERROR: %d\n", ret); return ret);
  ret = aclrtSetCurrentContext(*context);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSetCurrentContext failed. ERROR: %d\n", ret); return ret);
  ret = aclrtCreateStream(stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtCreateStream failed. ERROR: %d\n", ret); return ret);
  return 0;
}

template <typename T>
int CreateAclTensor(const std::vector<T>& hostData, const std::vector<int64_t>& shape, void** deviceAddr,
                    aclDataType dataType, aclTensor** tensor) {
  auto size = GetShapeSize(shape) * sizeof(T);
  // 调用aclrtMalloc申请device侧内存
  auto ret = aclrtMalloc(deviceAddr, size, ACL_MEM_MALLOC_HUGE_FIRST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtMalloc failed. ERROR: %d\n", ret); return ret);
  // 调用aclrtMemcpy将host侧数据拷贝到device侧内存上
  ret = aclrtMemcpy(*deviceAddr, size, hostData.data(), size, ACL_MEMCPY_HOST_TO_DEVICE);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtMemcpy failed. ERROR: %d\n", ret); return ret);

  // 计算连续tensor的strides
  std::vector<int64_t> strides(shape.size(), 1);
  for (int64_t i = shape.size() - 2; i >= 0; i--) {
    strides[i] = shape[i + 1] * strides[i + 1];
  }

  // 调用aclCreateTensor接口创建aclTensor
  *tensor = aclCreateTensor(shape.data(), shape.size(), dataType, strides.data(), 0, aclFormat::ACL_FORMAT_ND,
                            shape.data(), shape.size(), *deviceAddr);
  return 0;
}

int main() {
  // 1. （固定写法）device/context/stream初始化，参考AscendCL对外接口列表
  // 根据自己的实际device填写deviceId
  int32_t deviceId = 0;
  aclrtContext context;
  aclrtStream stream;
  auto ret = Init(deviceId, &context, &stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("Init acl failed. ERROR: %d\n", ret); return ret);

  // 2. 构造输入与输出，需要根据API的接口自定义构造
  std::vector<int64_t> selfShape = {2};
  std::vector<int64_t> matShape = {2, 2};
  std::vector<int64_t> vecShape = {2};
  std::vector<int64_t> outShape = {2};
  void* selfDeviceAddr = nullptr;
  void* matDeviceAddr = nullptr;
  void* vecDeviceAddr = nullptr;
  void* outDeviceAddr = nullptr;
  aclTensor* self = nullptr;
  aclTensor* mat = nullptr;
  aclTensor* vec = nullptr;
  aclScalar* alpha = nullptr;
  aclScalar* beta = nullptr;
  aclTensor* out = nullptr;
  std::vector<float> selfHostData = {1, 1};
  std::vector<float> matHostData = {1, 1, 1, 1};
  std::vector<float> vecHostData = {1, 1};
  std::vector<float> outHostData(0, 0);
  int8_t cubeMathType = 1;
  float alphaValue = 1.0f;
  float betaValue = 1.0f;
  // 创建self aclTensor
  ret = CreateAclTensor(selfHostData, selfShape, &selfDeviceAddr, aclDataType::ACL_FLOAT, &self);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  // 创建mat aclTensor
  ret = CreateAclTensor(matHostData, matShape, &matDeviceAddr, aclDataType::ACL_FLOAT, &mat);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  // 创建vec aclTensor
  ret = CreateAclTensor(vecHostData, vecShape, &vecDeviceAddr, aclDataType::ACL_FLOAT, &vec);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  // 创建alpha aclScalar
  alpha = aclCreateScalar(&alphaValue,aclDataType::ACL_FLOAT);
  CHECK_RET(alpha != nullptr, return ret);
  // 创建beta aclScalar
  beta = aclCreateScalar(&betaValue,aclDataType::ACL_FLOAT);
  CHECK_RET(beta != nullptr, return ret);
  // 创建out aclTensor
  ret = CreateAclTensor(outHostData, outShape, &outDeviceAddr, aclDataType::ACL_FLOAT, &out);
  CHECK_RET(ret == ACL_SUCCESS, return ret);

  // 3. 调用CANN算子库API，需要修改为具体的Api名称
  uint64_t workspaceSize = 0;
  aclOpExecutor* executor;
  // 调用aclnnAddmv第一段接口
  ret = aclnnAddmvGetWorkspaceSize(self, mat, vec, alpha, beta, out, cubeMathType, &workspaceSize, &executor);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclnnAddmvGetWorkspaceSize failed. ERROR: %d\n", ret); return ret);

  // 根据第一段接口计算出的workspaceSize申请device内存
  void* workspaceAddr = nullptr;
  if (workspaceSize > 0) {
    ret = aclrtMalloc(&workspaceAddr, workspaceSize, ACL_MEM_MALLOC_HUGE_FIRST);
    CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("allocate workspace failed. ERROR: %d\n", ret); return ret);
  }
  // 调用aclnnAddmv第二段接口
  ret = aclnnAddmv(workspaceAddr, workspaceSize, executor, stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclnnAddmv failed. ERROR: %d\n", ret); return ret);

  // 4. （固定写法）同步等待任务执行结束
  ret = aclrtSynchronizeStream(stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSynchronizeStream failed. ERROR: %d\n", ret); return ret);

  // 5. 获取输出的值，将device侧内存上的结果拷贝至host侧，需要根据具体API的接口定义修改
  auto size = GetShapeSize(outShape);
  std::vector<float> resultData(size, 0);
  ret = aclrtMemcpy(resultData.data(), resultData.size() * sizeof(resultData[0]), outDeviceAddr,
                    size * sizeof(resultData[0]), ACL_MEMCPY_DEVICE_TO_HOST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("copy result from device to host failed. ERROR: %d\n", ret); return ret);
  for (int64_t i = 0; i < size; i++) {
    LOG_PRINT("result[%ld] is: %f\n", i, resultData[i]);
  }

  // 6. 释放aclTensor和aclScalar，需要根据具体API的接口定义修改
  aclDestroyTensor(self);
  aclDestroyTensor(mat);
  aclDestroyTensor(vec);
  aclDestroyScalar(alpha);
  aclDestroyScalar(beta);
  aclDestroyTensor(out);

  // 7.释放device资源，需要根据具体API的接口定义修改
  aclrtFree(selfDeviceAddr);
  aclrtFree(matDeviceAddr);
  aclrtFree(vecDeviceAddr);
  aclrtFree(outDeviceAddr);
  if (workspaceSize > 0) {
    aclrtFree(workspaceAddr);
  }
  aclrtFree(stream);
  aclrtFree(context);
  aclrtFree(deviceId);
  aclFinalize();

  return 0;
}
```
