# aclnnAllGatherMatmul

## 支持的产品型号
Atlas A2训练系列产品。

## 接口原型

每个算子分为两段式接口，必须先调用“aclnnAllGatherMatmulGetWorkspaceSize”接口获取计算所需workspace大小以及包含了算子计算流程的执行器，再调用“aclnnAllGatherMatmul”接口执行计算。

*  `aclnnStatus aclnnAllGatherMatmulGetWorkspaceSize(const aclTensor *x1, const aclTensor *x2, const aclTensor *bias, const char *group, int64_t gatherIndex, int64_t commTurn,int64_t stream_mode, const aclTensor *output, const aclTensor *gatherOut, uint64_t *workspaceSize, aclOpExecutor **executor)`
*  `aclnnStatus aclnnAllGatherMatmul(void *workspace, uint64_t workspaceSize, aclOpExecutor *executor, const aclrtStream stream)`

## 功能描述

-   **算子功能**：完成all\_gather通信+matmul融合。
-   **计算公式**：

   $$
   output=x1*x2
   $$
   $$
   gatherOut=x1
   $$

## aclnnAllGatherMatmulGetWorkspaceSize

-   **参数说明：**
    -   x1（aclTensor\*，计算输入）：Device侧的两维aclTensor，mm左矩阵。数据类型支持：FLOAT16、BFLOAT16。数据格式支持：ND。**当前版本仅支持两维输入**。
    -   x2（aclTensor\*，计算输入）：Device侧的两维aclTensor，mm右矩阵。数据类型支持：FLOAT16、BFLOAT16。数据格式支持：ND。**当前版本仅支持两维输入**。
    -   bias（aclTensor\*，计算输入）：Device侧的aclTensor。数据类型支持：FLOAT16、BFLOAT16。数据格式支持：ND。**当前版本暂不支持bias输入**。
    -   group（char\*，计算输入）：Host侧标识列组的字符串。数据类型支持：string。
    -   gatherIndex（int64\_t，计算输入）：Host侧的整型，标识gather目标，0：左矩阵，1：右矩阵。数据类型支持：int64\_t。**当前版本仅支持输入0。**
    -   comm\_turn（int64\_t，计算输入）：Host侧的整型，通信数据切分数，即总数据量/单次通信量。数据类型支持：int64\_t。**当前版本仅支持输入0。**
    -   stream\_mode（int64\_t，计算输入）：Host侧的整型，acl流模式的枚举，当前只支持枚举值1，类型支持：int64\_t。
    -   output（aclTensor\*，计算输出）：Device侧的aclTensor，mm计算+all\_gather通信的结果。数据类型支持：同x1输入。
    -   gatherOut（aclTensor\*，计算输出）：Device侧的aclTensor，仅输出all\_gather通信后的结果。数据类型支持：同x1输入。
    -   workspaceSize（uint64\_t\*，出参）：Device侧的整型，返回需要在npu device侧申请的workspace大小。
    -   executor（aclOpExecutor\*\*，出参）：Device侧的aclOpExecutor，返回op执行器，包含了算子计算流程。

-   **返回值：**

    返回aclnnStatus状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

## aclnnAllGatherMatmul

-   **参数说明：**
    -   workspace（void\*，入参）：在Device侧申请的workspace内存起址。
    -   workspaceSize（uint64\_t，入参）：在Device侧申请的workspace大小，由第一段接口aclnnAllGatherMatmulGetWorkspaceSize获取。
    -   executor（aclOpExecutor\*，入参）：op执行器，包含了算子计算流程。
    -   stream（aclrtStream，入参）：指定执行任务的AscendCL stream流。

-   **返回值：**

    返回aclnnStatus状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

## 约束与限制

无。

## 调用示例

示例代码如下，仅供参考：

```
#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include <chrono>
#include <vector>
#include <string>
#include <cmath>
#include <cstdint>
#include <memory>
#include <thread>
#include "hccl/hccl.h"
#include <hccl/hccl_types.h>
#include <mutex>
#include <unistd.h>
#include <vector>
#include <sys/prctl.h>
#include <string.h>
#include <iostream>
#include <fstream>
#include <cstdio>
#include <string>
#include <vector>
#include <iomanip>
#include <cassert>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <algorithm>
#include <iostream>
#include "acl/acl.h"
#include <climits>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#define epsilon 1e-4
#ifndef TEST_ALL_GATHER
#define TEST_ALL_GATHER 1
#endif
#ifndef REDUCE_SCATTER_TRANS_B
#define REDUCE_SCATTER_TRANS_B 0
#endif
#if TEST_ALL_GATHER == 1
#include "aclnnop/aclnn_all_gather_matmul.h"
#else
#include "aclnn_matmul_reduce_scatter_local.h"
#endif
using namespace std;
//using namespace hccl;
std::mutex mt1;
int g_thread_status[8] = {0};
thread_local int g_dev_id;
thread_local int g_loop_cnt;
long g_time_w_ns[8][100];
long g_time_l_ns[8][100];
long g_time_i_ns[8][100];
long g_time_h_ns[8][100];
int64_t shapeM { 1 };
int64_t shapeK { 640 };
int64_t shapeN { 5120 };
int loop_cnt { 100 };
int communicationDataSize { 10 };
char *g_golden_file;
void set_device_id(int id) {
  g_dev_id = id;
}
void set_loop_cnt(int loop)
{
  g_loop_cnt = loop;
}
int get_loop_cnt(void)
{
  return g_loop_cnt;
}
int get_device_id(void)
{
  return g_dev_id;
}
void set_workspace_time(long ns)
{
  g_time_w_ns[get_device_id()][get_loop_cnt()] = ns;
}
long get_workspace_time(int i)
{
  return g_time_w_ns[get_device_id()][i];
}
void set_launch_time(long ns)
{
  g_time_l_ns[get_device_id()][get_loop_cnt()] = ns;
}
long get_launch_time(int i)
{
  return g_time_l_ns[get_device_id()][i];
}
void set_ilaunch_time(long ns)
{
  g_time_i_ns[get_device_id()][get_loop_cnt()] = ns;
}
long get_ilaunch_time(int i)
{
  return g_time_i_ns[get_device_id()][i];
}
void set_hccl_time(int loop, long ns)
{
  g_time_h_ns[get_device_id()][loop] = ns;
}
long get_hccl_time(int i)
{
  return g_time_h_ns[get_device_id()][i];
}
void show_time(int loop)
{
  int device = get_device_id();
  long wtt = 0, ltt = 0, iit = 0, htt = 0;
  for (int i = 0; i < loop; i++) {
    long wt = get_workspace_time(i);
    long lt = get_launch_time(i);
    long it = get_ilaunch_time(i);
    long ht = get_hccl_time(i);
    printf("### Device %d Index %d GetWorkspaceSize %ld ns, MixCore Kernel Launch %ld ns Aicpu Kernel Launch %ld ns, Hccl %ld ns.\n",
           device, i, wt, lt, it, ht);
    if (i != 0) {
      wtt += wt;
      ltt += lt;
      iit += it;
      htt += ht;
    }
  }
  printf("### Device %d GetWorkspaceSize avr %ld ns, MixCore Kernel Launch avr %ld ns Aicpu Kernel Launch avr %ld ns Hccl avr %ld.\n",
         device, wtt / (loop - 1), ltt / (loop - 1), iit / (loop - 1), htt / (loop - 1));
}
//=====================================================Start Test function======================================
void syncAllThread(int device_id) {
    //// sleep(2);
    printf("qmy syncAllThread start, device_id = %d\n", device_id);
    mt1.lock();
    g_thread_status[device_id] = 1;
    mt1.unlock();
    while(true) {
        mt1.lock();
        bool all_finished = true;
        for(int i=0;i<8;++i) {
            all_finished = all_finished && (g_thread_status[i] == 1);
        }
        mt1.unlock();
        if(all_finished == true) {
            break;
        }
    }
    mt1.lock();
    for(int i=0;i<8;++i) {
        g_thread_status[i] = 0;
    }
    mt1.unlock();
    struct timeval start;
    gettimeofday(&start,NULL);
    printf("qmy syncAllThread end, device_id = %d, tv_sec:%ld, tv_usec:%ld\n", device_id, start.tv_sec, start.tv_usec);
}
#define CLOCK_MONOTONIC  1
#define NSEC_PER_USEC 1000
#define MSEC_PER_SEC 1000
#define USEC_PER_SEC 1000000U
#define NSEC_PER_SEC 1000000000U
static inline uint64_t GetCurCpuTimestamp(void)
{
    struct timespec timestamp;
    (void)clock_gettime(CLOCK_MONOTONIC, ×tamp);
    return (uint64_t)((timestamp.tv_sec * NSEC_PER_SEC) + (timestamp.tv_nsec));
}
//=====================================================End Test function======================================
//=====================================================Start Profling function======================================
extern "C" void NnopbaseReportLaunchInfo(const uint64_t beginTime, const char *const opType);
extern "C" uint64_t MsprofSysCycleTime();
//=====================================================End Profling function======================================
//=====================================================Start Base function======================================
#define ALLOW_FP32_DOWN_PRECISION 1
#define N 128
typedef signed char s8;
typedef signed short s16;
typedef signed int s32;
typedef signed long long s64;
typedef unsigned char u8;
typedef unsigned short u16;
typedef unsigned int u32;
typedef unsigned long long u64;
#define ACL_CHECK(ret) do { \
    auto retcode = ret;    \
    if(retcode != ACL_SUCCESS) \
    {\
        printf("[ERROR] acl interface return err %s:%d, retcode: %d \n", __FILE__, __LINE__, retcode);\
        /*return retcode;*/\
    }\
} while(0)
#define CHECK_RET(cond, return_expr) \
    do {                             \
        if (!(cond)) {               \
            return_expr;             \
        }                            \
    } while (0)
struct DataSize {
    u64 min_bytes;
    u64 max_bytes;
    u64 step_bytes = 0;
    double step_factor;
    u64 count;
    u64 data_size;
    u64 type_size;
    int op_flag;
    void Print() {
        printf("acl min_bytes : %lld, max_bytes: %lld, step_bytes: %lld, step_factor: %f, count: %lld, data_size: %lld,  type_size: %lld\n",
            min_bytes, max_bytes, step_bytes, step_factor, count, data_size, type_size);
    }
};
struct Resource {
    aclrtStream rtStream;
    aclrtStream rtStream_aicpu;
    aclrtStream rtStream_record;
    aclrtEvent startEvent;
    aclrtEvent endEvent;
    aclrtContext context;
};
struct Args {
    int ndev;
    int rankId;
    uint32_t logicDeviceId;
    uint32_t rootRank;
    HcclComm hcclComm;
    Resource resources;
    DataSize dataPara;
};
aclTensor *CreateTensor(const int64_t *dims, uint64_t dimNum, aclDataType dataType, aclFormat format, void *dataAddr)
{
    std::vector<int64_t> strides(dimNum, 1);
    for (int64_t i = dimNum - 2; i >= 0; i--) {
        strides[i] = dims[i + 1] * strides[i + 1];
    }
    auto ret = aclCreateTensor(dims, dimNum,
                               dataType,
                               strides.data(), 0, format,
                               dims, dimNum,
                               dataAddr);
    return ret;
}
aclTensorList *CreateTensorList(u64 addr_list[], const int tensor_cnt, std::vector<int64_t> &shape,
                                aclDataType dataType, aclFormat format)
{
    aclTensor *tensors_value[tensor_cnt];
    for (auto i = 0; i < tensor_cnt; i++) {
        tensors_value[i] = aclCreateTensor(shape.data(), shape.size(), dataType, nullptr, 0,
                                           format, shape.data(), shape.size(), (void *)(addr_list[i]));
    }
    auto tensor_value_list = aclCreateTensorList(tensors_value, tensor_cnt);
    return tensor_value_list;
}
int64_t GetShapeSize(const std::vector<int64_t> &shape)
{
    int64_t shape_size = 1;
    for (auto i : shape) {
        shape_size *= i;
    }
    return shape_size;
}
std::string GetShapeString(const std::vector<int64_t> &shape)
{
    std::string res = "gm";
    for (auto i : shape) {
        res += "_" + to_string(i);
    }
    return res;
}
int Init(int32_t deviceId, aclrtContext *context, aclrtStream *stream)
{
    auto ret = aclInit(nullptr);
    CHECK_RET(ret == ACL_SUCCESS, std::cout << "aclInit failed. ERROR" << ret << std::endl;return -1);
    ret = aclrtSetDevice(deviceId);
    CHECK_RET(ret == ACL_SUCCESS, std::cout << "aclrtSetDevice failed. ERROR" << ret << std::endl;return -1);
    ret = aclrtCreateContext(context, deviceId);
    CHECK_RET(ret == ACL_SUCCESS, std::cout << "aclrtCreateContext failed. ERROR" << ret << std::endl;return -1);
    ret = aclrtSetCurrentContext(*context);
    CHECK_RET(ret == ACL_SUCCESS, std::cout << "aclrtSetCurrentContext failed. ERROR" << ret << std::endl;return -1);
    ret = aclrtCreateStream(stream);
    CHECK_RET(ret == ACL_SUCCESS, std::cout << "aclrtCreateStream failed. ERROR" << ret << std::endl;return -1);
    return 0;
}
typedef struct rtHostInputInfo {
    uint16_t addrOffset;
    uint16_t dataOffset;
} rtHostInputInfo_t;
typedef struct tagRtAicpuArgsEx {
    void *args; // args host mem addr
    rtHostInputInfo_t *hostInputInfoPtr; // nullptr means no host mem input
    void *kernelOffsetInfoPtr; // KernelOffsetInfo, it is different for CCE Kernel and fwk kernel
    uint32_t argsSize;
    uint16_t hostInputInfoNum; // hostInputInfo num
    uint16_t kernelOffsetInfoNum; // KernelOffsetInfo num
    uint16_t soNameAddrOffset; // just for CCE Kernel, default value is 0xffff for FWK kernel
    uint16_t kernelNameAddrOffset; // just for CCE Kernel, default value is 0xffff for FWK kernel
    bool isNoNeedH2DCopy; // is no need host to device copy: 0 means need H2D copy,
                               // other means doesn't need H2D copy.
    uint8_t reserved[3];
} rtAicpuArgsEx_t;
//===================================start aicpu======================================
const int MAX_RANK_NUM = 8;
typedef struct tagHcclCombinOpParam {
    u32 rank_id;
    u32 rank_num;
    u64 bufferSize;
    u64 windows[8];
    u64 no_ipc_notify_ids[16];
    u64 ipc_notify_ids[32];
    s64 stream_ids[8];
    u64 sq_ids[8];
} HcclCombinOpParam;
typedef enum {
    KFC_PREPARE_FIRST_TASK = 1,     // 只根据参数生成首任务，参数还是有所有任务信息
    KFC_PREPARE_ALL_TASK = 2,       // 直接生成所有任务
    KFC_PREPARE_END,                // 不生成
} KfcPrepareTaskType;
typedef struct tagGatherMsgBase {
    uint32_t msgCnt; // 一共几轮消息
    uint32_t sendOff; // 发送数据地址偏移，count * dataTypeSize
    uint32_t recvOff; // 接收数据地址偏移, count * dataTypeSize
    uint32_t stride;  // 跳写间隔
    uint32_t gatherOff; // 如果没有gather输出则代表gather地址在workspace偏移（内部使用）
    uint8_t funID;     // funtion ID
    uint8_t reuseMode; // 消息重用模式(内部使用)
    uint8_t gatherOut;  // hccl 数据类型
    uint8_t groupNum;   // groupNum
} GatherMsgBase;
typedef struct tagAicpuRpcServerModeLaunchOpAPI {
    //HcclOperatorType comm_type;   // 不同通信类型需要的资源可能不同，下面都是以最大需要数量定义。可以直接把全量资源下发，也可以按需只下发部分
    HcclReduceOp op_type;
    uint32_t rank_id;
    uint32_t rank_num;
    void *context;
    uint64_t windows[MAX_RANK_NUM];
    uint64_t no_ipc_notify_ids[MAX_RANK_NUM * 2];
    uint64_t ipc_notify_ids[MAX_RANK_NUM * 4];
    uint64_t stream_ids[MAX_RANK_NUM];
    uint64_t sq_ids[MAX_RANK_NUM];
    uint64_t workSpaceAddr;
    uint64_t genTaskNum;
    uint64_t send_buffer;
    uint64_t recv_buffer;
    GatherMsgBase gather_base;
    char kernel_name[32];
    char so_name[32];
    char hcomName[128];
} AicpuRpcServerModeLaunchOpAPI;
// msg between aicore and aicpu
typedef struct tagHcclAicpuOpParam {
    //HcclOperatorType comm_type; // 32b
    HcclReduceOp op_type;   // 32b
    uint64_t send_buffer;   //
    uint64_t recv_buffer;
    uint64_t count;
    // offset 32B
    HcclDataType hccl_data_type;
    uint32_t rank_dim;
    uint32_t rank_id;
    uint32_t blockDim; // 启动的BlockDim
    uint32_t blockIdx;
    uint32_t sendCnt; // 发送计数
    uint32_t rcvCnt;
    uint32_t funID; // 函数ID
    uint32_t valid; // 检查函数的有效性
    uint32_t orgM;
    uint32_t orgN;
    uint32_t width;
    uint32_t high;
    uint32_t isLast;
    uint32_t res;
    uint32_t res2;
    uint64_t res64[4];
} HcclAicpuOpParam;
// commResource: return filled param
//extern "C" HcclResult HcclCreatComResource(void* commResource, HcclOperatorType OpType, HcclComm comm);
/**
 * @ingroup rt_kernel
 * @brief op name
 */
typedef struct rtKernelLaunchNames {
    const char *soName;      // defined for so name
    const char *kernelName;  // defined for kernel type name
    const char *opName;      // defined for operator name
} rtKernelLaunchNames_t;
/*
 * @param [in] blockDim      block dimentions
 * @param [in] args          argments address for kernel function
 * @param [in] argsSize      argments size
 * @param [in] smDesc        shared memory description
 * @param [in] stm        associated stream
 * @return RT_ERROR_NONE for ok
 * @return RT_ERROR_INVALID_VALUE for error input
 */
extern "C" int rtAicpuKernelLaunch(const rtKernelLaunchNames_t *launchNames,
    uint32_t blockDim, const void *args, uint32_t argsSize, void *smDesc, aclrtStream stm);
extern "C" int rtAicpuKernelLaunchExWithArgs(const uint32_t kernelType, const char * opName,
                                                const uint32_t blockDim, const rtAicpuArgsEx_t *argsInfo,
                                                void * const smDesc, aclrtStream stm,
                                                const uint32_t flags);
//===================================end aicpu======================================
//=====================================================Start Matmul function======================================
int PrepareDeviceMemAndDataMatmulRcsTp(int devid, const std::vector<int64_t> &selfShape, void **selfDeviceAddr,
                                    const std::vector<int64_t> &otherShape, void **otherDeviceAddr,
                                  const std::vector<int64_t> &outShape, void **outDeviceAddr,
                                  const std::vector<int64_t> &gatherShape, void **gatherDeviceAddr)
{
    auto selfSize = GetShapeSize(selfShape) * sizeof(int16_t);
    auto ret = aclrtMalloc(selfDeviceAddr, selfSize, ACL_MEM_MALLOC_HUGE_FIRST);
    CHECK_RET(ret == ACL_SUCCESS, printf("[ERROR] PrepareDeviceMemAndDataMatmulRcsTp  aclrtMalloc failed. ret = %d \n", ret);return -1);
    auto otherSize = GetShapeSize(otherShape) * sizeof(int16_t);
    ret = aclrtMalloc(otherDeviceAddr, otherSize, ACL_MEM_MALLOC_HUGE_FIRST);
    auto outSize = GetShapeSize(outShape) * sizeof(int16_t);
    ret = aclrtMalloc(outDeviceAddr, outSize, ACL_MEM_MALLOC_HUGE_FIRST);
    CHECK_RET(ret == ACL_SUCCESS, printf("[ERROR] 842 aclrtMalloc failed. ret = %d \n", ret);return -1);
    printf("device %d matmul out addr %p size %lu\n", devid, *outDeviceAddr, outSize);
    auto gatherSize = GetShapeSize(gatherShape) * sizeof(int16_t);
    ret = aclrtMalloc(gatherDeviceAddr, gatherSize, ACL_MEM_MALLOC_HUGE_FIRST);
    return 0;
}
int PrepareTestParamsMatmulRcsTp(const std::vector<int64_t> &selfShape, void *selfDeviceAddr,
                         const std::vector<int64_t> &otherShape, void *otherDeviceAddr,
                         const std::vector<int64_t> &outShape, void *outDeviceAddr,
                         const std::vector<int64_t> &aicpuWinShape, void *aicpuWinDeviceAddr,
                         aclDataType dataType,
                         aclTensor **self, aclTensor **other, aclTensor **out, aclTensor **aicpuWin)
{
    *self = CreateTensor(selfShape.data(), selfShape.size(), dataType, aclFormat::ACL_FORMAT_ND, selfDeviceAddr);
    *other = CreateTensor(otherShape.data(), otherShape.size(), dataType, aclFormat::ACL_FORMAT_ND, otherDeviceAddr);
    *out = CreateTensor(outShape.data(), outShape.size(), dataType, aclFormat::ACL_FORMAT_ND, outDeviceAddr);
    *aicpuWin = CreateTensor(outShape.data(), outShape.size(), dataType, aclFormat::ACL_FORMAT_ND, outDeviceAddr);
    return 0;
}
thread_local uint64_t g_workspace_size = 0;
thread_local void *g_workspace_ptr = NULL;
int Aclnn_MatmulRcsTp(aclrtStream stream, aclrtStream stream_aicpu, const std::vector<int64_t> &selfShape, void *&selfDeviceAddr,
                    const std::vector<int64_t> &otherShape, void *&otherDeviceAddr,
                    const std::vector<int64_t> &outShape, void *&outDeviceAddr,
                    const std::vector<int64_t> &gatherShape, void *&gatherDeviceAddr,
                    AicpuRpcServerModeLaunchOpAPI& apiParam, aclTensorList *wins, void* workspaceAddr)
{
    // CALL aclnn API
    aclTensor *self = CreateTensor(selfShape.data(), selfShape.size(),
            aclDataType::ACL_FLOAT16, aclFormat::ACL_FORMAT_ND, selfDeviceAddr);
    aclTensor *other = CreateTensor(otherShape.data(), otherShape.size(),
            aclDataType::ACL_FLOAT16, aclFormat::ACL_FORMAT_ND, otherDeviceAddr);
    aclTensor *out = CreateTensor(outShape.data(), outShape.size(),
            aclDataType::ACL_FLOAT16, aclFormat::ACL_FORMAT_ND, outDeviceAddr);
    aclTensor *gatherOut = CreateTensor(gatherShape.data(), gatherShape.size(),
            aclDataType::ACL_FLOAT16, aclFormat::ACL_FORMAT_ND, gatherDeviceAddr);
    printf("### rank %d Output tensor address %p\n", get_device_id(), outDeviceAddr);
    ACL_CHECK(aclrtMemset(outDeviceAddr, 32, get_device_id() + 0x80, 32));
    uint64_t workspaceSize = 0;
    aclOpExecutor *e = NULL;
    workspaceAddr = NULL;
    auto ret = aclnnAllGatherMatmulGetWorkspaceSize(self, other, NULL, apiParam.hcomName, 0, 0, 1, out, gatherOut, &workspaceSize, &e);
    printf("rank %d aclnnAllGatherMatmulGetWorkspaceSize %p, ret %d\n", (int)apiParam.rank_id, (void *)workspaceSize, (int)ret);
    if (workspaceSize > 0) {
    if (g_workspace_size < workspaceSize) {
          ACL_CHECK(aclrtMalloc(&g_workspace_ptr, workspaceSize, ACL_MEM_MALLOC_HUGE_FIRST));
      g_workspace_size = workspaceSize;
    }
        workspaceAddr = g_workspace_ptr;
    }
    ret = aclnnAllGatherMatmul(workspaceAddr, workspaceSize, e, stream);
    printf("rank %d aclnnAllGatherMatmul %p ret %d\n", (int)apiParam.rank_id, workspaceAddr, (int)ret);
    aclDestroyTensor(self);
    aclDestroyTensor(other);
    aclDestroyTensor(out);
    return 0;
}
typedef int32_t rtError_t;
typedef void *rtNotify_t;
typedef void *rtStream_t;
// init hccl args end
extern "C" rtError_t rtNotifyWait(rtNotify_t notify, rtStream_t stm);
extern "C" rtError_t rtNotifyRecord(rtNotify_t notify, rtStream_t stm);
extern "C" rtError_t rtNotifyCreate(int32_t deviceId, rtNotify_t *notify);
extern "C" HcclResult HcclCreateComResource(char* commName, uint32_t streamMode, void** commContext);
extern "C" HcclResult HcclGetAicpuOpStreamNotify(char* commName, rtStream_t* Opstream, void** aicpuNotify);
int callMatmulAndHccl(HcclComm hcclComm, void* sendBuff, void* recvBuff, void* checkBuff, void* hostBuff, void* recvBuffTemp, DataSize& dataPara,
    aclrtStream rtStream, aclrtStream rtStream_aicpu, aclrtStream rtStream_record, aclrtEvent startEvent, aclrtEvent endEvent, int checkErr, int ndev, int device_id, int rootRank, void* workspaceAddr)//主函数
{
    set_device_id(device_id);
    int loop_cnt = 1; //ctrl_loop
    int init_hccl = 1;
    printf("rank %d input rtStream_aicpu %p\n", device_id, rtStream_aicpu);
    void *hcom_context;
    uint32_t rank_id = device_id;
    uint32_t rank_num = ndev;
    aclrtStream aicpu_stm = rtStream_aicpu;
    void *aicpu_notify;
    struct timespec tp1, tp2;
    long cost;
    char hcom_name[128];
    for (int l = 0; l < init_hccl; l++) {
      auto hccl_name_ret = HcclGetCommName(hcclComm, hcom_name);
      CHECK_RET(hccl_name_ret == HCCL_SUCCESS,
      printf("[ERROR] HcclGetCommName failed. hccl_ret = %d \n", hccl_name_ret ); return -1);
      printf("rank %d hcom name is %s\n", device_id, hcom_name);
      auto hccl_ret = HcclGetRankId(hcclComm, &rank_id);
      CHECK_RET(hccl_ret == HCCL_SUCCESS,
      printf("[ERROR] HcclGetRankId failed. hccl_ret = %d \n", hccl_ret ); return -1);
      hccl_ret = HcclGetRankSize(hcclComm, &rank_num);
      CHECK_RET(hccl_ret == HCCL_SUCCESS,
      printf("[ERROR] HcclGetRankSize failed. hccl_ret = %d \n", hccl_ret ); return -1);
      printf("rank id %d rank size %d\n", rank_id, rank_num);
    }
    AicpuRpcServerModeLaunchOpAPI apiParam = {
        //.comm_type = comm_type,
        .op_type = HcclReduceOp::HCCL_REDUCE_SUM,
        .rank_id = rank_id,
        .rank_num = rank_num,
    };
    strcpy(apiParam.hcomName, hcom_name);
    std::vector<int64_t> selfShape = {shapeM, shapeK};
    void *selfDeviceAddr = nullptr;
#if (REDUCE_SCATTER_TRANS_B == 0)
    std::vector<int64_t> otherShape = {shapeK, shapeN};
#else
    std::vector<int64_t> otherShape = {shapeN, shapeK};
#endif
    void *otherDeviceAddr = nullptr;
    void *gatherDeviceAddr = nullptr;
#if (TEST_ALL_GATHER == 1)
    std::vector<int64_t> gatherOutShape = {shapeM * 8, shapeK};
    std::vector<int64_t> matmulOutShape = {shapeM * 8, shapeN};
#else
    std::vector<int64_t> gatherOutShape = {shapeM, shapeK};
    std::vector<int64_t> matmulOutShape = {shapeM, shapeN};
#endif
    void *matmulOutDeviceAddr = nullptr;
    {
         auto ret = PrepareDeviceMemAndDataMatmulRcsTp(device_id, selfShape, &selfDeviceAddr,
                                                     otherShape, &otherDeviceAddr,
                                                  matmulOutShape, &matmulOutDeviceAddr,
                        gatherOutShape, &gatherDeviceAddr);
         CHECK_RET(ret == 0, return ret);
    }
    for (int i=0;i<loop_cnt;i++) {
    set_loop_cnt(i);
        printf("=====================Startloop: %d =======================\n", i);
        {
            printf("start run Aclnn_MatmulRcsTp  TIME %ld\n", GetCurCpuTimestamp());
            auto ret = Aclnn_MatmulRcsTp(rtStream, aicpu_stm, selfShape, selfDeviceAddr,
                            otherShape, otherDeviceAddr,
                            matmulOutShape, matmulOutDeviceAddr,
                gatherOutShape, gatherDeviceAddr,
                            apiParam, nullptr, workspaceAddr);
            CHECK_RET(ret == ACL_SUCCESS, printf("[ERROR] Aclnn_MatmulRcsTp failed. ret = %d \n", ret);return -1);
        }
    }
    ACL_CHECK(aclrtSynchronizeStreamWithTimeout(rtStream, 5000));
    auto matmul_shape_cnt = GetShapeSize(matmulOutShape);
    auto gather_shape_cnt = GetShapeSize(gatherOutShape);
    show_time(loop_cnt);
    printf("### Device %d HcclCreatComResource time %ld ns.\n", device_id, cost);
    return 0;
}
int launchOneThread(Args& args)
{
    std::string thread_name = "test" + std::to_string(args.logicDeviceId);
    prctl(PR_SET_NAME, thread_name.c_str());
    cpu_set_t cpu_set;
    CPU_ZERO(&cpu_set);
    int numa_node = args.logicDeviceId;
    int cpu_start = 0;
    int cpu_num = 24;
    for (int i = cpu_start; i < cpu_start + cpu_num; i++) {
        CPU_SET(i, &cpu_set);
    }
    int ret = sched_setaffinity(gettid(), sizeof(cpu_set), &cpu_set);
    if (ret) {
        printf("failed to set cpu affinity, errno = %d\n", errno);
        return -1;
    }
    printf("0 rankId: %d, stream: %p, context : %p\n", args.logicDeviceId, args.resources.rtStream, args.resources.context);
    DataSize dataPara = args.dataPara;
    ret = 0;
    int checkErr = 0;
    ACL_CHECK(aclrtSetDevice(args.logicDeviceId));
    ACL_CHECK(aclrtSetCurrentContext(args.resources.context));
    void* sendBuff = nullptr;
    void* recvBuff = nullptr;
    void* checkBuff = nullptr;
    void* hostBuff = nullptr;
    void* recvBuffTemp = nullptr;
    void* workspaceAddr = nullptr;
    uint64_t workspaceSize = 16*1024*1024 + 36*1024*1024 + 500 * 1024 * 1024;
    auto ret1012 = aclrtMalloc(&workspaceAddr, workspaceSize, ACL_MEM_MALLOC_HUGE_FIRST);
    CHECK_RET(ret1012 == ACL_SUCCESS, printf("[ERROR] tesetet aclrtMalloc failed. ret = %d \n", ret1012); return -1;);
    printf("workspaceAddr:%p\r\n", workspaceAddr);
    for (int i = 0; i< 1; i++){
        ret = callMatmulAndHccl(args.hcclComm, sendBuff, recvBuff, checkBuff, hostBuff, recvBuffTemp, dataPara,
                args.resources.rtStream, args.resources.rtStream_aicpu, args.resources.rtStream_record,
                args.resources.startEvent, args.resources.endEvent,
                                    checkErr, args.ndev, args.logicDeviceId, args.rootRank, workspaceAddr);
        if (ret != 0)
        {
            printf("TestCall execute failed, Detailed logs are stored in path: /root/ascend/log/");
            return ret;
        }
    }
    ACL_CHECK(aclrtSynchronizeStreamWithTimeout(args.resources.rtStream, 10000));
    ACL_CHECK(aclrtDestroyStreamForce(args.resources.rtStream));
    ACL_CHECK(aclrtSynchronizeStreamWithTimeout(args.resources.rtStream_aicpu, 10000));
    ACL_CHECK(aclrtDestroyStreamForce(args.resources.rtStream_aicpu));
    ACL_CHECK(aclrtSynchronizeStreamWithTimeout(args.resources.rtStream_record, 10));
    ACL_CHECK(aclrtDestroyStreamForce(args.resources.rtStream_record));
    auto aclnnRet = aclrtFree(workspaceAddr);
    CHECK_RET(aclnnRet == ACL_SUCCESS,printf("[ERROR] aclnnFree failed. aclnnRet = %d \n", aclnnRet ); return -1);
    //销毁集合通信域
    HcclCommDestroy(args.hcclComm);
    //重置设备
    ACL_CHECK(aclrtResetDevice(args.logicDeviceId));
    return 0;
}
extern "C" int rtStreamCreateWithFlags(void **stm, int32_t priority, uint32_t flags);
int launchMultiThread(uint32_t ndev, int32_t* devices, HcclComm* comms, Resource* resources)
{
    DataSize dataPara;
    dataPara.data_size = 10485760;
    dataPara.type_size = sizeof(uint16_t);
    dataPara.count = (dataPara.data_size + sizeof(uint16_t) - 1)/sizeof(uint16_t);// data->count;
    dataPara.step_factor = 1;//data->step_factor;
    dataPara.step_bytes = 1;//data->step_bytes;
    dataPara.max_bytes = 10485760;//data->max_bytes;
    dataPara.min_bytes = 10485760;//data->min_bytes;
    dataPara.op_flag = 0;//op_flag;
    Args args[ndev];
    aclrtStream rtStream[ndev*3];
    aclrtContext context[ndev];
    for (uint32_t rankId = 0; rankId < ndev; rankId++) {
            ACL_CHECK(aclrtSetDevice(rankId));
            ACL_CHECK(rtStreamCreateWithFlags(&rtStream[rankId], 0, 0x600));
            ACL_CHECK(rtStreamCreateWithFlags(&rtStream[ndev + rankId], 0, 0x600));
            ACL_CHECK(rtStreamCreateWithFlags(&rtStream[ndev*2 + rankId], 0, 0x600));
            ACL_CHECK(aclrtGetCurrentContext(&context[rankId]));
    }
    std::vector<std::unique_ptr<std::thread>> threads(ndev);
    for (uint32_t rankId = 0; rankId < ndev; rankId++) {
        args[rankId].rankId = rankId;
        args[rankId].rootRank = devices[0];
        args[rankId].ndev = ndev;
        args[rankId].logicDeviceId = devices[rankId];
        args[rankId].hcclComm = comms[rankId];
        args[rankId].resources = resources[rankId];
        args[rankId].dataPara = dataPara;
        args[rankId].resources.rtStream = rtStream[rankId];
        args[rankId].resources.rtStream_aicpu = rtStream[ndev + rankId];
        args[rankId].resources.rtStream_record = rtStream[ndev*2 + rankId];
        args[rankId].resources.context = context[rankId];
        // printf("2 rankId: %d, stream: %p\n", rankId,  args[rankId].resources.rtStream);
        threads[rankId].reset(new (std::nothrow)
            std::thread(&launchOneThread, std::ref(args[rankId])));
    }
    for (uint32_t rankId = 0; rankId < ndev; rankId++) {
        threads[rankId]->join();
    }
    return 0;
}
int main(int argc, char** argv) {
    shapeM = 16384;
    shapeK = 3416;
    shapeN = 5120;
    int ret = 0;
    uint32_t ndev = 8;
    int32_t devices[8] = {0, 1, 2, 3, 4, 5, 6, 7};
    HcclComm comms[N];
    Resource resources[N];
    ACL_CHECK(aclInit(NULL));
    for (int i = 0; i < 1; i++) {
        //初始化集合通信域
        for (int i = 0; i < ndev; i++) {
            ACL_CHECK(aclrtSetDevice(devices[i]));
        }
        ret = HcclCommInitAll(ndev, devices, comms);
        if (ret != 0) {
            printf("This is an error in init_hcclComm.\n");
            return -1;
        }
        // //启动测试
        ret = launchMultiThread( ndev, devices, comms, resources);
        if (ret != 0) {
            printf("This is an error in launchMultiThread.\n");
            return -1;
        }
    }
    ACL_CHECK(aclFinalize());
    return ret;
}
```

