# aclnnAvgPool2d

## 支持的产品型号
- Atlas 推理系列产品（Ascend 310P处理器）。
- Atlas 训练产品。
- Atlas A2训练系列产品。

## 接口原型
每个算子分为[两段式接口](common/两段式接口.md)，必须先调用“aclnnAvgPool2dGetWorkspaceSize”接口获取计算所需workspace大小以及包含了算子计算流程的执行器，再调用“aclnnAvgPool2d”接口执行计算。

- `aclnnStatus aclnnAvgPool2dGetWorkspaceSize(const aclTensor *self, const aclIntArray *kernelSize, const aclIntArray *stride, const aclIntArray *padding, const bool ceilMode, const bool countIncludePad, const uint64_t divisorOverride, int8_t cubeMathType, aclTensor *out, uint64_t *workspaceSize, aclOpExecutor **executor)`

- `aclnnStatus aclnnAvgPool2d(void *workspace, uint64_t workspaceSize, aclOpExecutor *executor, const aclrtStream stream)`

## 功能描述

- 算子功能：对输入Tensor进行窗口为$kH * kW$、步长为$sH * sW$的二维平均池化操作，其中$k$为kernelSize，表示池化窗口的大小，$s$为stride，表示池化操作的步长。
- 计算公式：
  输入input（$N,C,H,W$）、输出out（$N,C,H_{out},W_{out}$）和池化步长（$stride$）、池化窗口大小（$kH,kW$）的关系是

$$
H_{out}=\lfloor \frac{H_{in}+2*padding[0]-kH}{stride[0]}+1 \rfloor
$$

$$
W_{out}=\lfloor \frac{W_{in}+2*padding[1]-kW}{stride[1]}+1 \rfloor
$$

$$
out(N_i,C_i,h,w)=\frac{1}{kH*kW}\sum_{m=0}^{kH-1}\sum_{n=0}^{kW-1}input(N_i,C_i,stride[0]*h+m,stride[1]*w+n)
$$

## aclnnAvgPool2dGetWorkspaceSize

- **参数说明**：

  - self（const aclTensor\*，计算输入）：公式中的$input$，Device侧的tensor，数据类型支持FLOAT16和FLOAT。支持非连续的Tensor。支持[数据格式](common/数据格式.md)为NCHW和NCL。

  - kernelSize（const aclIntArray\*，计算输入）:Host侧的aclIntArray，长度为1（$kH=kW$）或2（$kH, kW$），表示池化窗口大小。数据类型支持INT32和INT64。数值必须大于0。

  - stride（const aclIntArray\*，计算输入）: 公式中的$stride$，Host侧的aclIntArray，长度为1（$sH=sW$）或2（$sH, sW$），表示池化操作的步长。数据类型支持INT32和INT64。数值必须大于0。

  - padding（const aclIntArray\*，计算输入）: 公式中的$padding$，Host侧的aclIntArray，长度为1（$padH=padW$）或2（$padH, padW$），表示在输入的H、W方向上padding补0的层数。数据类型支持INT32和INT64。数值不能小于0。

  - ceilMode（const bool，计算输入）: 表示推导的输出out的shape是否向上取整。数据类型支持BOOL。

  - countIncludePad（const bool，计算输入）: 计算平均池化时是否包括padding填充的0。数据类型支持BOOL。

  - divisorOverride（const int64_t，计算输入）: 表示取平均的除数。数据类型支持INT64。

  - cubeMathType（int8_t，计算输入）: host侧的整型，判断Cube单元应该使用哪种计算逻辑进行运算，数据类型支持INT8，支持的枚举值如下：
    * 0:KEEP_DTYE，保持输入的数据类型进行计算。当输入是FLOAT，Atlas 训练产品和Atlas 推理系列产品（Ascend 310P处理器）暂不支持，取0时会报错。
    * 1:ALLOW_FP32_DOWN_PRECISON，允许将输入数据降精度计算。当输入是FLOAT，Atlas 训练产品和Atlas 推理系列产品（Ascend 310P处理器）允许转换为FLOAT16计算，Atlas A2训练系列产品允许转换为HFLOAT32计算。
    * 2:USE_FP16，允许转换为数据类型FLOAT16进行计算。当输入数据类型是FLOAT，转换为FLOAT16计算。
    * 3:USE_HF32，允许转换为数据类型HFLOAT32计算。当输入是FLOAT，Atlas 训练产品和Atlas 推理系列产品（Ascend 310P处理器）暂不支持，取3时会报错，Atlas A2训练系列产品转换为HFLOAT32计算。

  - out（aclTensor\*，计算输出）: 输出的tensor，公式中的$out$。数据类型支持FLOAT16和FLOAT。支持非连续的Tensor。支持[数据格式](common/数据格式.md)为NCHW和CHW。数据类型、[数据格式](common/数据格式.md)需要与self一致。

  - workSpaceSize（uint64_t\*，出参）：返回用户需要在Device侧申请的workspace大小。

  - executor（aclOpExecutor\*\*，出参）：返回op执行器，包含了算子计算流程。

- **返回值**：

  aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

```text
第一段接口完成入参校检，出现以下场景时报错：
161001 (ACLNN_ERR_PARAM_NULLPTR)：1. 传入的self、kernelSize、stride、padding或out是空指针。
161002 (ACLNN_ERR_PARAM_INVALID)：1. 传入的self或out的数据类型/数据格式不在支持的范围之内。
                                  2. 传入的self和out的数据类型/数据格式不一致。
                                  3. 传入的self、kernelSize、stride或out存在某维度的值小于0。
```

## aclnnAvgPool2d

- **参数说明**：

  - workspace：在Device侧申请的workspace内存起址。
  - workspaceSize：在Device侧申请的workspace大小，由第一段接口aclnnAddGetWorkspaceSize获取。
  - executor：op执行器，包含了算子计算流程。
  - stream：指定执行任务的 AscendCL Stream流。

- **返回值：**

  aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

## 约束与限制
- 对于Atlas 训练产品，Cube单元不支持FLOAT32计算。当输入为FLOAT32，可通过设置cubeMathType=1（ALLOW_FP32_DOWN_PRECISON）来允许接口内部cast到FLOAT16进行计算。

## 调用示例
示例代码如下，仅供参考，具体编译和执行过程请参考[编译与运行样例](common/编译与运行样例.md)。
```c++
#include <iostream>
#include <vector>
#include "acl/acl.h"
#include "aclnnop/aclnn_avgpool2d.h"

#define CHECK_RET(cond, return_expr) \
  do {                               \
    if (!(cond)) {                   \
      return_expr;                   \
    }                                \
  } while (0)

#define LOG_PRINT(message, ...)     \
  do {                              \
    printf(message, ##__VA_ARGS__); \
  } while (0)

int64_t GetShapeSize(const std::vector<int64_t>& shape) {
  int64_t shapeSize = 1;
  for (auto i : shape) {
    shapeSize *= i;
  }
  return shapeSize;
}

int Init(int32_t deviceId, aclrtContext* context, aclrtStream* stream) {
  // 固定写法，AscendCL初始化
  auto ret = aclInit(nullptr);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclInit failed. ERROR: %d\n", ret); return ret);
  ret = aclrtSetDevice(deviceId);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSetDevice failed. ERROR: %d\n", ret); return ret);
  ret = aclrtCreateContext(context, deviceId);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtCreateContext failed. ERROR: %d\n", ret); return ret);
  ret = aclrtSetCurrentContext(*context);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSetCurrentContext failed. ERROR: %d\n", ret); return ret);
  ret = aclrtCreateStream(stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtCreateStream failed. ERROR: %d\n", ret); return ret);
  return 0;
}

template <typename T>
int CreateAclTensor(const std::vector<T>& hostData, const std::vector<int64_t>& shape, void** deviceAddr,
                    aclDataType dataType, aclTensor** tensor) {
  auto size = GetShapeSize(shape) * sizeof(T);
  // 调用aclrtMalloc申请device侧内存
  auto ret = aclrtMalloc(deviceAddr, size, ACL_MEM_MALLOC_HUGE_FIRST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtMalloc failed. ERROR: %d\n", ret); return ret);
  // 调用aclrtMemcpy将host侧数据拷贝到device侧内存上
  ret = aclrtMemcpy(*deviceAddr, size, hostData.data(), size, ACL_MEMCPY_HOST_TO_DEVICE);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtMemcpy failed. ERROR: %d\n", ret); return ret);

  // 计算连续tensor的strides
  std::vector<int64_t> strides(shape.size(), 1);
  for (int64_t i = shape.size() - 2; i >= 0; i--) {
    strides[i] = shape[i + 1] * strides[i + 1];
  }

  // 调用aclCreateTensor接口创建aclTensor
  *tensor = aclCreateTensor(shape.data(), shape.size(), dataType, strides.data(), 0, aclFormat::ACL_FORMAT_ND, shape.data(), shape.size(), *deviceAddr);
  return 0;
}

int main() {
  // 1. （固定写法）device/context/stream初始化，参考AscendCL对外接口列表
  // 根据自己的实际device填写deviceId
  int32_t deviceId = 0;
  aclrtContext context;
  aclrtStream stream;
  auto ret = Init(deviceId, &context, &stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("Init acl failed. ERROR: %d\n", ret); return ret);

  // 2. 构造输入与输出，需要根据API的接口自定义构造
  int64_t divisorOverride = 0;
  bool countIncludePad = true;
  bool ceilMode = false;
  int8_t cubeMathType = 1;

  std::vector<int64_t> selfShape = {1, 16, 4, 4};
  std::vector<int64_t> outShape = {1, 16, 1, 1};

  std::vector<int64_t> kernelDims = {4, 4};
  std::vector<int64_t> strideDims = {1, 1};
  std::vector<int64_t> paddingDims = {0, 0};

  void* selfDeviceAddr = nullptr;
  void* outDeviceAddr = nullptr;

  aclTensor* self = nullptr;
  aclTensor* out = nullptr;

  std::vector<float> selfHostData(256, 2);
  std::vector<float> outHostData(16, 0);

  // 创建self aclTensor
  ret = CreateAclTensor(selfHostData, selfShape, &selfDeviceAddr, aclDataType::ACL_FLOAT, &self);
  CHECK_RET(ret == ACL_SUCCESS, return ret);

  // 创建kernel aclIntArray
  aclIntArray *kernelSize = aclCreateIntArray(kernelDims.data(), 2);

  // 创建stride aclIntArray
  aclIntArray *strides = aclCreateIntArray(strideDims.data(), 2);

  // 创建paddings aclIntArray
  aclIntArray *paddings = aclCreateIntArray(paddingDims.data(), 2);

  // 创建out aclTensor
  ret = CreateAclTensor(outHostData, outShape, &outDeviceAddr, aclDataType::ACL_FLOAT, &out);
  CHECK_RET(ret == ACL_SUCCESS, return ret);

  // 3. 调用CANN算子库API，需要修改为具体的Api名称
  uint64_t workspaceSize = 0;
  aclOpExecutor* executor;
  // 调用aclnnAvgPool2d第一段接口
  ret = aclnnAvgPool2dGetWorkspaceSize(self,
                                       kernelSize,
                                       strides,
                                       paddings,
                                       ceilMode,
                                       countIncludePad,
                                       divisorOverride,
                                       cubeMathType,
                                       out,
                                       &workspaceSize,
                                       &executor);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclnnAvgPool2dGetWorkspaceSize failed. ERROR: %d\n", ret); return ret);
  // 根据第一段接口计算出的workspaceSize申请device内存
  void* workspaceAddr = nullptr;
  if (workspaceSize > 0) {
    ret = aclrtMalloc(&workspaceAddr, workspaceSize, ACL_MEM_MALLOC_HUGE_FIRST);
    CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("allocate workspace failed. ERROR: %d\n", ret); return ret);
  }
  // 调用aclnnAvgPool2d第二段接口
  ret = aclnnAvgPool2d(workspaceAddr, workspaceSize, executor, stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclnnAvgPool2d failed. ERROR: %d\n", ret); return ret);

  // 4. （固定写法）同步等待任务执行结束
  ret = aclrtSynchronizeStream(stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSynchronizeStream failed. ERROR: %d\n", ret); return ret);

  // 5. 获取输出的值，将device侧内存上的结果拷贝至host侧，需要根据具体API的接口定义修改
  auto size = GetShapeSize(outShape);
  std::vector<float> outData(size, 0);
  ret = aclrtMemcpy(outData.data(), outData.size() * sizeof(outData[0]), outDeviceAddr,
                    size * sizeof(outData[0]), ACL_MEMCPY_DEVICE_TO_HOST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("copy result from device to host failed. ERROR: %d\n", ret); return ret);
  for (int64_t i = 0; i < size; i++) {
    LOG_PRINT("out result[%ld] is: %f\n", i, outData[i]);
  }

  // 6. 释放aclTensor和aclScalar，需要根据具体API的接口定义修改
  aclDestroyTensor(self);
  aclDestroyTensor(out);
  aclDestroyIntArray(kernelSize);
  aclDestroyIntArray(strides);
  aclDestroyIntArray(paddings);
  
  // 7. 释放device资源，需要根据具体API的接口定义修改
  aclrtFree(selfDeviceAddr);
  aclrtFree(outDeviceAddr);
  if (workspaceSize > 0) {
    aclrtFree(workspaceAddr);
  }
  aclrtDestroyStream(stream);
  aclrtDestroyContext(context);
  aclrtResetDevice(deviceId);
  aclFinalize();

  return 0;
}
```
