# aclnnConvDepthwise2d

## 支持的产品型号
- Atlas 训练产品。
- Atlas A2训练系列产品。

## 接口原型
每个算子分为[两段式接口](common/两段式接口.md)，必须先调用“aclnnConvDepthwise2dGetWorkspaceSize”接口获取计算所需workspace大小以及包含了算子计算流程的执行器，再调用“aclnnConvDepthwise2d”接口执行计算。

* `aclnnStatus aclnnConvDepthwise2dGetWorkspaceSize(const aclTensor *self, const aclTensor *weight, const aclIntArray *kernelSize, const aclTensor *bias, const aclIntArray *stride, const aclIntArray *padding, const aclIntArray *dilation, aclTensor *out, int8_t cubeMathType, uint64_t *workspaceSize, aclOpExecutor **executor)`
* `aclnnStatus aclnnConvDepthwise2d(void *const workspace, uint64_t workspaceSize, aclOpExecutor *executor, aclStream stream)`

## 功能描述

- 算子功能：2D深度卷积。在每一深度上卷积，不求和。

- 计算公式：

​        我们假定输入（self）的shape是 $(N, C_{\text{in}}, H, W)$ ,输出（out）的shape是 $(N, N*C_{\text{out}}, H_{\text{out}}, W_{\text{out}})$，那么每个卷积核的输出将被表示为：

$$
\text{out}(N_i, C_{\text{out}_j}) = \text{bias}(C_{\text{out}_j}) + \text{weight}(C_{\text{out}_j}, C_{\text{in}_j}) \star \text{self}(N_i, C_{\text{in}_j})
$$

​       其中，$\star$表示互相关的计算，$N$代表batch size，$C$代表通道数，$W$和$H$分别代表宽和高。


## aclnnConvDepthwise2dGetWorkspaceSize

- **参数说明：**

  * self(const aclTensor *, 计算输入)：公式中的`self`，数据类型仅支持FLOAT，FLOAT16，支持[非连续的Tensor](common/非连续的Tensor.md)， [数据格式](common/数据格式.md)为NCL，NCHW（NHWC）。
  * weight(const aclTensor *, 计算输入)：公式中的`weight`，数据类型与self一致，支持[非连续的Tensor](common/非连续的Tensor.md)， [数据格式](common/数据格式.md)为NCHW。weight第一维的数值应等于self通道数的整数倍，H, W两维的数值需小于self的H, W两维的数值。
  * kernelSize(const aclIntArray *, 计算输入)：卷积核尺寸，(int, int)型元组。
  * bias(const aclTensor *, 计算输入)：公式中的`bias`，数据类型与self, weight一致，仅支持一维且数值与weight第一维相等，支持[非连续的Tensor](common/非连续的Tensor.md)，[数据格式](common/数据格式.md)为NCHW。
  * stride(const aclIntArray *, 计算输入)：卷积扫描步长，数组长度需等于1或者self的维度减2。
  * padding(const aclIntArray *, 计算输入)：对self的填充，数组长度需等于1或者self的维度减2。
  * dilation(const aclIntArray *, 计算输入)：卷积核中元素的间隔，数组长度需等于1或者self的维度减2。
  * out(aclTensor *, 计算输出)：公式中的`out`，数据类型与self, weight一致，[数据格式](common/数据格式.md)与self一致。out通道数应等于weight第一维的数值。
  * cubeMathType(int8_t *, 计算输入)：用于判断Cube单元应该使用哪种计算逻辑进行运算，INT8类型的枚举值，枚举如下：
    * 0:KEEP_DTYE，保持输入的数据类型进行计算。当输入是FLOAT，Atlas 训练产品Cube计算单元暂不支持，取0时会报错。
    * 1:ALLOW_FP32_DOWN_PRECISON，允许将输入数据降精度计算。当输入是FLOAT，Atlas 训练产品转换为FLOAT16计算，Atlas A2训练系列产品转换为HFLOAT32计算。
    * 2:USE_FP16，允许转换为数据类型FLOAT16进行计算。当输入数据类型是FLOAT，转换为FLOAT16计算。
    * 3:USE_HF32，允许转换为数据类型HFLOAT32计算。当输入是FLOAT，Atlas 训练产品Cube计算单元暂不支持，取3时会报错，Atlas A2训练系列产品转换为HFLOAT32计算。
  * workspaceSize(uint64_t *, 出参)：返回用户需要在Device侧申请的workspace大小。
  * executor(aclOpExecutor **, 出参)：返回op执行器，包含了算子计算流程。

- **返回值：**

  aclnnStatus： 返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

  ```c++
  161001 ACLNN_ERR_PARAM_NULLPTR：1. 传入的指针类型入参是空指针。
  161002 ACLNN_ERR_PARAM_INVALID：1. self，weight，bias，out数据类型和数据格式不在支持的范围之内。
                                  2. stride, padding, dilation输入shape不对。
                                  3. self, weight, bias, out数据类型不一致。
                                  4. weight和self通道数不满足要求。
                                  5. out的shape不满足infershape结果。
                                  6. self, weight或bias为空tensor。
  ```

## aclnnConvDepthwise2d

- **参数说明：**

  * workspace（void *, 入参）：在Device侧申请的workspace内存起址。
  * workspace_size（uint64_t，入参）：在Device侧申请的workspace大小，由第一段接口aclnnConvDepthwise2dGetWorkspaceSize获取。
  * executor（aclOpExecutor *，入参）：op执行器，包含了算子计算流程。
  * stream（aclrtStream，入参）：指定执行任务的 AscendCL Stream流。

- **返回值：**

  aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

## 约束与限制
无

## 调用示例

示例代码如下，仅供参考，具体编译和执行过程请参考[编译与运行样例](common/编译与运行样例.md)。
```c++
#include <iostream>
#include <vector>
#include "acl/acl.h"
#include "aclnnop/aclnn_convolution.h"

#define CHECK_RET(cond, return_expr) \
 do {                                \
    if (!(cond)) {                   \
        return_expr;                 \
    }                                \
 } while (0)

#define LOG_PRINT(message, ...)      \
 do {                                \
    printf(message, ##__VA_ARGS__);  \
 } while (0)

int64_t GetShapeSize(const std::vector<int64_t>& shape) {
    int64_t shape_size = 1;
    for (auto i: shape) {
        shape_size *= i;
    }
    return shape_size;
}

int Init(int32_t deviceId, aclrtContext* context, aclrtStream* stream) {
  // 固定写法，AscendCL初始化
  auto ret = aclInit(nullptr);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclInit failed. ERROR: %d\n", ret); return ret);
  ret = aclrtSetDevice(deviceId);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSetDevice failed. ERROR: %d\n", ret); return ret);
  ret = aclrtCreateContext(context, deviceId);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtCreateContext failed. ERROR: %d\n", ret); return ret);
  ret = aclrtSetCurrentContext(*context);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSetCurrentContext failed. ERROR: %d\n", ret); return ret);
  ret = aclrtCreateStream(stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtCreateStream failed. ERROR: %d\n", ret); return ret);
  return 0;
}
template <typename T>
int CreateAclTensor(const std::vector<T>& hostData, const std::vector<int64_t>& shape, void** deviceAddr,
                    aclDataType dataType, aclTensor** tensor) {
  auto size = GetShapeSize(shape) * sizeof(T);
  // 调用aclrtMalloc申请device侧内存
  auto ret = aclrtMalloc(deviceAddr, size, ACL_MEM_MALLOC_HUGE_FIRST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtMalloc failed. ERROR: %d\n", ret); return ret);

  // 调用aclrtMemcpy将host侧数据拷贝到device侧内存上
  ret = aclrtMemcpy(*deviceAddr, size, hostData.data(), size, ACL_MEMCPY_HOST_TO_DEVICE);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtMemcpy failed. ERROR: %d\n", ret); return ret);

  // 计算连续tensor的strides
  std::vector<int64_t> strides(shape.size(), 1);
  for (int64_t i = shape.size() - 2; i >= 0; i--) {
    strides[i] = shape[i + 1] * strides[i + 1];
  }

  // 调用aclCreateTensor接口创建aclTensor
  *tensor = aclCreateTensor(shape.data(), shape.size(), dataType, strides.data(), 0, aclFormat::ACL_FORMAT_NCHW,
                            shape.data(), shape.size(), *deviceAddr);
  return 0;
}

int main() {
  // 1. （固定写法）device/context/stream初始化，参考AscendCL对外接口列表
  // 根据自己的实际device填写deviceId
  int32_t deviceId = 0;
  aclrtContext context;
  aclrtStream stream;
  auto ret = Init(deviceId, &context, &stream);
  // check根据自己的需要处理
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("Init acl failed. ERROR: %d\n", ret); return ret);

  // 2. 构造输入与输出，需要根据API的接口自定义构造
  std::vector<int64_t> shapeSelf = {2, 2, 2, 2};
  std::vector<int64_t> shapeWeight = {2, 1, 1, 1};
  std::vector<int64_t> shapeBias = {2};
  std::vector<int64_t> shapeResult = {2, 2, 4, 4};

  void* deviceDataSelf = nullptr;
  void* deviceDataWeight = nullptr;
  void* deviceDataBias = nullptr;
  void* deviceDataResult = nullptr;

  aclTensor* self = nullptr;
  aclTensor* weight = nullptr;
  aclTensor* bias = nullptr;
  aclTensor* result = nullptr;
  aclIntArray* kernelSize = nullptr;
  aclIntArray* stride = nullptr;
  aclIntArray* padding = nullptr;
  aclIntArray* dilation = nullptr;

  std::vector<float> selfData(GetShapeSize(shapeSelf) * 2, 1);
  std::vector<float> weightData(GetShapeSize(shapeWeight) * 2, 1);
  std::vector<float> biasData(GetShapeSize(shapeBias) * 2, 1);
  std::vector<float> outData(GetShapeSize(shapeResult) * 2, 1);
  std::vector<int64_t> kernelSizeData = {1, 1};
  std::vector<int64_t> strideData = {1, 1};
  std::vector<int64_t> paddingData = {1, 1};
  std::vector<int64_t> dilationData = {1, 1};

  // 创建self aclTensor
  ret = CreateAclTensor(selfData, shapeSelf, &deviceDataSelf, aclDataType::ACL_FLOAT, &self);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  // 创建weight aclTensor
  ret = CreateAclTensor(weightData, shapeWeight, &deviceDataWeight, aclDataType::ACL_FLOAT, &weight);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  // 创建bias aclTensor
  ret = CreateAclTensor(biasData, shapeBias, &deviceDataBias, aclDataType::ACL_FLOAT, &bias);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  // 创建out aclTensor
  ret = CreateAclTensor(outData, shapeResult, &deviceDataResult, aclDataType::ACL_FLOAT, &result);
  CHECK_RET(ret == ACL_SUCCESS, return ret);

  kernelSize = aclCreateIntArray(kernelSizeData.data(), 2);
  stride = aclCreateIntArray(strideData.data(), 2);
  padding = aclCreateIntArray(paddingData.data(), 2);
  dilation = aclCreateIntArray(dilationData.data(), 2);

  // 3. 调用CANN算子库API，需要修改为具体的API
  uint64_t workspaceSize = 0;
  aclOpExecutor* executor;
  // 调用aclnnConvDepthwise2d第一段接口
  ret = aclnnConvDepthwise2dGetWorkspaceSize(self, weight, kernelSize, bias, stride, padding, dilation, result, 1, &workspaceSize, &executor);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclnnConvDepthwise2dGetWorkspaceSize failed. ERROR: %d\n", ret); return ret);
  // 根据第一段接口计算出的workspaceSize申请device内存
  void* workspaceAddr = nullptr;
  if (workspaceSize > 0) {
    ret = aclrtMalloc(&workspaceAddr, workspaceSize, ACL_MEM_MALLOC_HUGE_FIRST);
    CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("allocate workspace failed. ERROR: %d\n", ret); return ret);
  }
  // 调用aclnnConvDepthwise2d第二段接口
  ret = aclnnConvDepthwise2d(workspaceAddr, workspaceSize, executor, stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclnnConvDepthwise2d failed. ERROR: %d\n", ret); return ret);

  // 4. （固定写法）同步等待任务执行结束
  ret = aclrtSynchronizeStream(stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSynchronizeStream failed. ERROR: %d\n", ret); return ret);

  // 5. 获取输出的值，将device侧内存上的结果拷贝至host侧，需要根据具体API的接口定义修改
  auto size = GetShapeSize(shapeResult);
  std::vector<float> resultData(size, 0);
  ret = aclrtMemcpy(resultData.data(), resultData.size() * sizeof(resultData[0]), deviceDataResult,
                    size * sizeof(float), ACL_MEMCPY_DEVICE_TO_HOST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("copy result from device to host failed. ERROR: %d\n", ret); return ret);
  for (int64_t i = 0; i < size; i++) {
    LOG_PRINT("result[%ld] is: %f\n", i, resultData[i]);
  }

  // 6. 释放aclTensor，需要根据具体API的接口定义修改
  aclDestroyTensor(self);
  aclDestroyTensor(weight);
  aclDestroyTensor(bias);
  aclDestroyTensor(result);
  aclDestroyIntArray(kernelSize);
  aclDestroyIntArray(stride);
  aclDestroyIntArray(padding);
  aclDestroyIntArray(dilation);

  // 7.释放device资源，需要根据具体API的接口定义修改
  aclrtFree(deviceDataSelf);
  aclrtFree(deviceDataWeight);
  aclrtFree(deviceDataBias);
  aclrtFree(deviceDataResult);
  if (workspaceSize > 0) {
    aclrtFree(workspaceAddr);
  }
  aclrtDestroyStream(stream);
  aclrtDestroyContext(context);
  aclrtResetDevice(deviceId);
  aclFinalize();
  return 0;
}
```
