# aclnnConvolutionBackward

## 支持的产品型号
- Atlas 推理系列产品（Ascend 310P处理器）。
- Atlas 训练产品。
- Atlas A2训练系列产品。

## 接口原型
每个算子分为[两段式接口](common/两段式接口.md)，必须先调用“aclnnConvolutionBackwardGetWorkspaceSize”接口获取计算所需workspace大小以及包含了算子计算流程的执行器，再调用“aclnnConvolutionBackward”接口执行计算。

- `aclnnStatus aclnnConvolutionBackwardGetWorkspaceSize(const aclTensor *grad_output, const aclTensor *input, const aclTensor *weight, const int[]? bias_sizes, const int[] stride, const int[] padding, const int[] dilation, const bool transposed, const int[] output_padding, const int groups, const bool[3] output_mask, aclTensor[3] outputs, uint64_t &workspace_size, aclOpExecutor **executor)`

- `aclnnStatus aclnnConvolutionBackward(void *workspace, uint64_t workspace_size, const aclStream stream, aclOpExecutor **executor)`

## 功能描述

- 算子功能：卷积的反向传播。根据输出掩码设置计算输入、权重和偏差的梯度。此函数支持1D、2D或3D卷积。
- 计算公式：

  卷积反向传播需要计算对卷积正向的输入张量 $x$、卷积核权重张量 $w$ 和偏置 $b$ 的梯度。

  对于 $x$ 的梯度 $\frac{\partial L}{\partial x}$：

  $$
  \frac{\partial L}{\partial x_{n, c_{in}, i, j}} = \sum_{c_{out}=1}^{C_{out}} \sum_{p=1}^{k_H} \sum_{q=1}^{k_W} \frac{\partial L}{\partial y_{n, c_{out}, i-p, j-q}}\cdot w_{c_{out}, c_{in}, p, q}
  $$

  其中，$L$ 为损失函数，$\frac{\partial L}{\partial y}$ 为输出张量 $y$ 对 $L$ 的梯度。

  对于 $w$ 的梯度 $\frac{\partial L}{\partial w}$：

  $$
  \frac{\partial L}{\partial w_{c_{out}, c_{in}, p, q}} = \sum_{n=1}^{N} \sum_{i=1}^{H_{out}} \sum_{j=1}^{W_{out}} x_{n, c_{in}, i \cdot s_H + p, j \cdot s_W + q} \cdot \frac{\partial L}{\partial y_{n, c_{out}, i, j}}
  $$

  对于 $b$ 的梯度 $\frac{\partial L}{\partial b}$：

  $$
  \frac{\partial L}{\partial b_{c_{out}}} = \sum_{n=1}^{N}       \sum_{i=1}^{H_{out}} \sum_{j=1}^{W_{out}} \frac{\partial L}{\partial y_{n, c_{out}, i, j}}
  $$



## aclnnConvolutionBackwardGetWorkspaceSize

- **参数说明：**

  - gradOutput: 卷积输出梯度在Device侧的aclTensor。 数据类型支持BFLOAT16(仅Atlas A2训练系列产品支持)、FLOAT、FLOAT16、INT8、BOOL, 且数据类型需要与input、weight保持一致。支持非连续的Tensor。[数据格式](common/数据格式.md)支持NCL、NCHW、NCDHW, 且[数据格式](common/数据格式.md)需要与input、weight一致。
  - input: 卷积输入在Device侧的aclTensor。 数据类型支持BFLOAT16(仅Atlas A2训练系列产品支持)、FLOAT、FLOAT16、INT8、BOOL, 且数据类型需要与grad_output、weight保持一致。支持非连续的Tensor。[数据格式](common/数据格式.md)支持NCL、NCHW、NCDHW, 且[数据格式](common/数据格式.md)需要与grad_output、weight一致。
  - weight: 卷积权重在Device侧的aclTensor。 数据类型支持BFLOAT16(仅Atlas A2训练系列产品支持)、FLOAT、FLOAT16、INT8、BOOL, 且数据类型需要与grad_output、input保持一致。支持非连续的Tensor。[数据格式](common/数据格式.md)支持NCL、NCHW、NCDHW, 且[数据格式](common/数据格式.md)需要与grad_output、input一致。
  - biasSizes: 卷积正向过程中偏差的shape, Device侧的aclIntArray。数据类型为int64, 数组长度是1。 其在普通卷积中等于[weight.shape[0]],在转置卷积中等于[weight.shape[1] * groups]
  - stride: 反向传播过程中卷积核在输入上移动的步长, Device侧的aclIntArray。数据类型为int64, 数组长度可以为1或weight.ndimension() - 2, 当为1时卷积核在输入上移动的步长在D、H和W方向上一样。 数值必须大于0。
  - padding: 反向传播过程中对于输入填充, Device侧的aclIntArray。数据类型为int64, 数组长度可以为1或weight.ndimension() - 2, 当为1时对于输入填充在D、H和W方向上一样。 数值必须大于0。
  - dilation: 反向传播过程中的膨胀参数, Device侧的aclIntArray。数据类型为int64, 数组长度可以为1或weight.ndimension() - 2, 当为1时膨胀参数在D、H和W方向上一样。 数值必须大于0。
  - transposed: 转置卷积使能标志位, Host侧的布尔类型, 当其值为True时使能转置卷积
  - outputPadding: 反向传播过程中对于输出填充, 仅在transposed为True时生效, Host侧的整形数组, 数据类型支持int32、int64, 数组长度可以为1或weight.ndimension() - 2, 当为1时对于输出填充在D、H和W方向上一样。 数值必须大于0。
  - groups: 反向传播过程中输入通道的分组数。 数据类型为int64, 数值必须大于0。
  - outputMask: 输出掩码参数, 指定输出中是否包含输入、权重、偏差的梯度。Host侧的布尔类型Array。反向传播过程输出掩码参数为True对应位置的梯度。
  - cubeMathType: 用于判断Cube单元应该使用哪种计算逻辑进行运算，INT8类型的枚举值，枚举如下：
    * 0:KEEP_DTYE，保持输入的数据类型进行计算。当输入是FLOAT，Atlas 训练产品和Atlas 推理系列产品（Ascend 310P处理器）暂不支持，取0时会报错。
    * 1:ALLOW_FP32_DOWN_PRECISON，允许将输入数据降精度计算。当输入是FLOAT，Atlas 训练产品和Atlas 推理系列产品（Ascend 310P处理器）允许转换为FLOAT16计算，Atlas A2训练系列产品允许转换为HFLOAT32计算。
    * 2:USE_FP16，允许转换为数据类型FLOAT16进行计算。当输入数据类型是FLOAT，转换为FLOAT16计算。
    * 3:USE_HF32，允许转换为数据类型HFLOAT32计算。当输入是FLOAT，Atlas 训练产品和Atlas 推理系列产品（Ascend 310P处理器）暂不支持，取3时会报错，Atlas A2训练系列产品转换为HFLOAT32计算。
  - gradInput: 卷积输入梯度在Device侧的aclTensor。
  - gradWeight: 卷积输入梯度在Device侧的aclTensor。
  - gradBias: 卷积输入梯度在Device侧的aclTensor。
  - workspaceSize: 返回用户需要在Device侧申请的workspace大小。
  - executor: 返回op执行器, 包含了算子计算流程。

- **返回值：**

  aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

## aclnnConvolutionBackward

- **参数说明：**

  - workspace：在Device侧申请的workspace内存起址。
  - workspace_size：在Device侧申请的workspace大小, 由第一段接口aclnnConvolutionBackwardGetWorkspaceSize获取。
  - stream：指定执行任务的 AscendCL Stream流。
  - executor：op执行器, 包含了算子计算流程。

- **返回值：**

  aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

## 约束与限制
无

## 调用示例

示例代码如下，仅供参考，具体编译和执行过程请参考[编译与运行样例](common/编译与运行样例.md)。

```c++
#include <iostream>
#include <vector>
#include "acl/acl.h"
#include "aclnnop/aclnn_convolution_backward.h"

#define CHECK_RET(cond, return_expr) \
  do {                               \
    if (!(cond)) {                   \
      return_expr;                   \
    }                                \
  } while (0)

#define LOG_PRINT(message, ...)     \
  do {                              \
    printf(message, ##__VA_ARGS__); \
  } while (0)

int64_t GetShapeSize(const std::vector<int64_t>& shape) {
  int64_t shapeSize = 1;
  for (auto i : shape) {
    shapeSize *= i;
  }
  return shapeSize;
}

int Init(int32_t deviceId, aclrtContext* context, aclrtStream* stream) {
  // 固定写法，AscendCL初始化
  auto ret = aclInit(nullptr);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclInit failed. ERROR: %d\n", ret); return ret);
  ret = aclrtSetDevice(deviceId);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSetDevice failed. ERROR: %d\n", ret); return ret);
  ret = aclrtCreateContext(context, deviceId);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtCreateContext failed. ERROR: %d\n", ret); return ret);
  ret = aclrtSetCurrentContext(*context);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSetCurrentContext failed. ERROR: %d\n", ret); return ret);
  ret = aclrtCreateStream(stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtCreateStream failed. ERROR: %d\n", ret); return ret);
  return 0;
}

template <typename T>
int CreateAclTensor(const std::vector<T>& hostData, const std::vector<int64_t>& shape, void** deviceAddr,
                    aclDataType dataType, aclTensor** tensor) {
  auto size = GetShapeSize(shape) * sizeof(T);
  // 调用aclrtMalloc申请device侧内存
  auto ret = aclrtMalloc(deviceAddr, size, ACL_MEM_MALLOC_HUGE_FIRST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtMalloc failed. ERROR: %d\n", ret); return ret);
  // 调用aclrtMemcpy将host侧数据拷贝到device侧内存上
  ret = aclrtMemcpy(*deviceAddr, size, hostData.data(), size, ACL_MEMCPY_HOST_TO_DEVICE);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtMemcpy failed. ERROR: %d\n", ret); return ret);

  // 计算连续tensor的strides
  std::vector<int64_t> strides(shape.size(), 1);
  for (int64_t i = shape.size() - 2; i >= 0; i--) {
    strides[i] = shape[i + 1] * strides[i + 1];
  }

  // 调用aclCreateTensor接口创建aclTensor
  if (shape.size() == 4) {
    *tensor = aclCreateTensor(shape.data(), shape.size(), dataType, strides.data(), 0, aclFormat::ACL_FORMAT_NCHW,
                              shape.data(), shape.size(), *deviceAddr);
  } else {
    *tensor = aclCreateTensor(shape.data(), shape.size(), dataType, strides.data(), 0, aclFormat::ACL_FORMAT_ND,
                              shape.data(), shape.size(), *deviceAddr);
  }

  return 0;
}

int main() {
  // 1. （固定写法）device/context/stream初始化
  // 根据自己的实际device填写deviceId
  int32_t deviceId = 0;
  aclrtContext context;
  aclrtStream stream;
  auto ret = Init(deviceId, &context, &stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("Init acl failed. ERROR: %d\n", ret); return ret);
  // 2. 构造输入与输出，需要根据API的接口自定义构造
  std::vector<int64_t> gradOutputShape = {2, 2, 7, 7};
  std::vector<int64_t> inputShape = {2, 2, 7, 7};
  std::vector<int64_t> weightShape = {2, 2, 1, 1};
  std::vector<int64_t> bias = {2};
  std::vector<int64_t> stride = {1, 1};
  std::vector<int64_t> padding = {0, 0};
  std::vector<int64_t> dilation = {1, 1};
  bool transposed = false;
  std::vector<int64_t> outputPadding = {0, 0};
  int groups = 1;
  bool outputMask[3] = {true, true, true};
  int8_t cubeMathType = 0;

  std::vector<int64_t> gradInputShape = {2, 2, 7, 7};
  std::vector<int64_t> gradWeightShape = {2, 2, 1, 1};
  std::vector<int64_t> gradBiasShape = {2};

  // 创建gradOut aclTensor
  std::vector<float> gradOutputData(GetShapeSize(gradOutputShape) * 2, 1);
  aclTensor* gradOutput = nullptr;
  void *gradOutputdeviceAddr = nullptr;
  ret = CreateAclTensor(gradOutputData, gradOutputShape, &gradOutputdeviceAddr, aclDataType::ACL_FLOAT16, &gradOutput);
  CHECK_RET(ret == ACL_SUCCESS, return ret);

  // 创建input aclTensor
  std::vector<float> inputData(GetShapeSize(inputShape) * 2, 1);
  aclTensor* input = nullptr;
  void *inputdeviceAddr = nullptr;
  ret = CreateAclTensor(inputData, inputShape, &inputdeviceAddr, aclDataType::ACL_FLOAT16, &input);
  CHECK_RET(ret == ACL_SUCCESS, return ret);

  // 创建weight aclTensor
  std::vector<float> weightData(GetShapeSize(weightShape) * 2, 1);
  aclTensor* weight = nullptr;
  void *weightDeviceAddr = nullptr;
  ret = CreateAclTensor(weightData, weightShape, &weightDeviceAddr, aclDataType::ACL_FLOAT16, &weight);
  CHECK_RET(ret == ACL_SUCCESS, return ret);

  // 创建gradInput aclTensor
  std::vector<float> gradInputData(GetShapeSize(gradInputShape) * 2, 1);
  aclTensor* gradInput = nullptr;
  void *gradInputDeviceAddr = nullptr;
  ret = CreateAclTensor(gradInputData, gradInputShape, &gradInputDeviceAddr, aclDataType::ACL_FLOAT16, &gradInput);
  CHECK_RET(ret == ACL_SUCCESS, return ret);

  // 创建gradWeight aclTensor
  std::vector<float> gradWeightData(GetShapeSize(gradWeightShape) * 2, 1);
  aclTensor* gradWeight = nullptr;
  void *gradWeightDeviceAddr = nullptr;
  ret = CreateAclTensor(gradWeightData, gradWeightShape, &gradWeightDeviceAddr, aclDataType::ACL_FLOAT16, &gradWeight);
  CHECK_RET(ret == ACL_SUCCESS, return ret);

  // 创建gradBias aclTensor
  std::vector<float> gradBiasData(GetShapeSize(gradBiasShape) * 2, 1);
  aclTensor* gradBias = nullptr;
  void *gradBiasDeviceAddr = nullptr;
  ret = CreateAclTensor(gradBiasData, gradBiasShape, &gradBiasDeviceAddr, aclDataType::ACL_FLOAT16, &gradBias);
  CHECK_RET(ret == ACL_SUCCESS, return ret);

  aclIntArray *biasSizes = aclCreateIntArray(bias.data(), 1);
  aclIntArray *strides = aclCreateIntArray(stride.data(), 2);
  aclIntArray *pads = aclCreateIntArray(padding.data(), 2);
  aclIntArray *dilations = aclCreateIntArray(dilation.data(), 2);
  aclIntArray *outputPads = aclCreateIntArray(outputPadding.data(), 2);
  aclBoolArray *outMask = aclCreateBoolArray(outputMask, 3);

  // 3. 调用CANN算子库API，需要修改为具体的Api名称
  uint64_t workspaceSize = 0;
  aclOpExecutor* executor;
  // 调用aclnnConvolutionBackwardGetWorkspaceSize第一段接口
  ret = aclnnConvolutionBackwardGetWorkspaceSize(gradOutput, input, weight, biasSizes, strides, pads, dilations, transposed,
                                                 outputPads, groups, outMask, cubeMathType, gradInput, gradWeight, gradBias,
                                                 &workspaceSize, &executor);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclnnConvolutionBackwardGetWorkspaceSize failed. ERROR: %d\n", ret); return ret);
  // 根据第一段接口计算出的workspaceSize申请device内存
  void* workspaceAddr = nullptr;
  if (workspaceSize > 0) {
    ret = aclrtMalloc(&workspaceAddr, workspaceSize, ACL_MEM_MALLOC_HUGE_FIRST);
    CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("allocate workspace failed. ERROR: %d\n", ret); return ret);
  }
  // 调用aclnnConvolutionBackward第二段接口
  ret = aclnnConvolutionBackward(workspaceAddr, workspaceSize, executor, stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclnnConvolutionBackward failed. ERROR: %d\n", ret); return ret);
  // 4. （固定写法）同步等待任务执行结束
  ret = aclrtSynchronizeStream(stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSynchronizeStream failed. ERROR: %d\n", ret); return ret);
  // 5. 获取输出的值，将device侧内存上的结果拷贝至host侧，需要根据具体API的接口定义修改
  auto size = GetShapeSize(gradInputShape);
  std::vector<float> gradInputResult(size, 0);
  ret = aclrtMemcpy(gradInputResult.data(), gradInputResult.size() * sizeof(gradInputResult[0]), gradInputDeviceAddr,
                    size * sizeof(gradInputResult[0]), ACL_MEMCPY_DEVICE_TO_HOST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("copy result from device to host failed. ERROR: %d\n", ret); return ret);
  for (int64_t i = 0; i < size; i++) {
    LOG_PRINT("gradInputResult[%ld] is: %f\n", i, gradInputResult[i]);
  }

  size = GetShapeSize(gradWeightShape);
  std::vector<float> gradWeightResult(size, 0);
  ret = aclrtMemcpy(gradWeightResult.data(), gradWeightResult.size() * sizeof(gradWeightResult[0]), gradWeightDeviceAddr,
                    size * sizeof(gradWeightResult[0]), ACL_MEMCPY_DEVICE_TO_HOST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("copy result from device to host failed. ERROR: %d\n", ret); return ret);
  for (int64_t i = 0; i < size; i++) {
    LOG_PRINT("gradWeightResult[%ld] is: %f\n", i, gradWeightResult[i]);
  }

  size = GetShapeSize(gradBiasShape);
  std::vector<float> gradBiasResult(size, 0);
  ret = aclrtMemcpy(gradBiasResult.data(), gradBiasResult.size() * sizeof(gradBiasResult[0]), gradInputDeviceAddr,
                    size * sizeof(gradBiasResult[0]), ACL_MEMCPY_DEVICE_TO_HOST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("copy result from device to host failed. ERROR: %d\n", ret); return ret);
  for (int64_t i = 0; i < size; i++) {
    LOG_PRINT("gradBiasResult[%ld] is: %f\n", i, gradBiasResult[i]);
  }
  // 6. 释放aclTensor和aclIntArray 需要根据具体API的接口定义修改
  aclDestroyTensor(gradOutput);
  aclDestroyTensor(input);
  aclDestroyTensor(weight);
  aclDestroyTensor(gradInput);
  aclDestroyTensor(gradWeight);
  aclDestroyTensor(gradBias);
  aclDestroyIntArray(biasSizes);
  aclDestroyIntArray(strides);
  aclDestroyIntArray(pads);
  aclDestroyIntArray(dilations);
  aclDestroyIntArray(outputPads);
  aclDestroyBoolArray(outMask);

  // 7.释放device资源，需要根据具体API的接口定义修改
  aclrtFree(gradOutputdeviceAddr);
  aclrtFree(inputdeviceAddr);
  aclrtFree(weightDeviceAddr);
  aclrtFree(gradInputDeviceAddr);
  aclrtFree(gradWeightDeviceAddr);
  aclrtFree(gradBiasDeviceAddr );
  if (workspaceSize > 0) {
    aclrtFree(workspaceAddr);
  }
  aclrtDestroyStream(stream);
  aclrtDestroyContext(context);
  aclrtResetDevice(deviceId);
  aclFinalize();
  return 0;
}
```

