# aclnnDeepNorm

## 支持的产品型号
Atlas A2训练系列产品。


## 接口原型

每个算子分为两段式接口，必须先调用“aclnnDeepNormGetWorkspaceSize”接口获取计算所需workspace大小以及包含了算子计算流程的执行器，再调用“aclnnDeepNorm”接口执行计算。

-   **aclnnStatus aclnnDeepNormGetWorkspaceSize\(const aclTensor \*x, const aclTensor \*gx, const aclTensor \*beta, const aclTensor \*gamma, double alphaOptional, double epsilonOptional, const aclTensor \*meanOut, const aclTensor \*rstdOut, const aclTensor \*yOut, uint64\_t \*workspaceSize, aclOpExecutor \*\*executor\)**
-   **aclnnStatus aclnnDeepNorm\(void \*workspace, uint64\_t workspaceSize, aclOpExecutor \*executor, const aclrtStream stream\)**

## 功能描述

-   算子功能：实现DeepNorm算子功能。
-   计算公式：

   $$
DeepNorm(x_i^`)=(\frac{x_i^`-\bar{x^`}}{STD})*gamma+beta,
   $$

   $$
 where STD=\sqrt{\frac{1}{n} \sum_{i=1}^n (x^`_i-\bar{x^`})^2+eps}, \quad \operatorname{x^`_i}= alpha * x_i + gx_i
   $$




## aclnnDeepNormGetWorkspaceSize

-   **参数说明**：
    -   x（aclTensor\*，计算输入）：Device侧的aclTensor， 公式中的输入x，数据类型支持FLOAT、FLOAT16、BFLOAT16、shape支持2-8维度，数据格式支持ND。
    -   gx （aclTensor\*，计算输入）：Device侧的aclTensor， 公式中的输入gx，数据类型支持FLOAT、FLOAT16、BFLOAT16、shape支持2-8维度，数据格式支持ND
    -   beta （aclTensor\*，计算输入）：Device侧的aclTensor， 公式中的输入beta，数据类型支持FLOAT、FLOAT16、BFLOAT16、shape支持2-8维度，数据格式支持ND。
    -   gamma（aclTensor\*，计算输入）：Device侧的aclTensor， 公式中的输入gamma，数据类型支持FLOAT、FLOAT16、BFLOAT16、shape支持2-8维度，数据格式支持ND。
    -   alphaOptional（double，计算输入）：host侧的double，公式中的输入alpha，是输入x的权重。
    -   epsilonOptional（double，计算输入）：host侧的double，公式中的输入eps，用于防止除0错误。
    -   meanOut（aclTensor\*，计算输出）：Device侧的aclTensor， 公式中的输出mean，数据类型支持FLOAT、shape支持2-8维度，数据格式支持ND。
    -   rstdOutr（aclTensor\*，计算输出）：Device侧的aclTensor， 公式中的输出rstd，数据类型支持FLOAT、shape支持2-8维度，数据格式支持ND。
    -   yOut（aclTensor\*，计算输出）：Device侧的aclTensor， 公式中的输出y，数据类型支持FLOAT、FLOAT16、BFLOAT16、shape支持2-8维度，数据格式支持ND。
    -   workspaceSize（uint64\_t\*，出参）：返回用户需要在npu device侧申请的workspace大小。
    -   executor（aclOpExecutor\*\*，出参）：返回op执行器，包含了算子计算流程。

-   **返回值**

    返回aclnnStatus状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

## aclnnDeepNorm

-   **参数说明：**
    -   workspace（void\*，入参）：在Device侧申请的workspace内存起址。
    -   workspaceSize（uint64\_t，入参）：在Device侧申请的workspace大小，由第一段接口aclnnDeepNormGetWorkspaceSize获取。
    -   executor（aclOpExecutor\*，入参）：op执行器，包含了算子计算流程。
    -   stream（aclrtStream，入参）：指定执行任务的AscendCL stream流。

-   **返回值：**

    返回aclnnStatus状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

## 约束与限制

-  功能维度：
    -  数据类型支持：
        1.  x、gx、beta、gamma、y：FLOAT32、FLOAT16、BFLOAT16。
        2.  rstd、mean：FLOAT32。

    -  数据格式支持：ND

-  未支持类型说明：
    -  DOUBLE：指令不支持DOUBLE。
    -  是否支持空tensor：不支持空进空出。
    -  是否非连续tensor：不支持输入非连续，不支持数据非连续。

-  边界值场景说明：
    -  当输入是inf时，输出为inf。
    -  当输入是nan时，输出为nan。

## 调用示例

```
#include <stdio.h>
#include <stdlib.h>
#include <filesystem>
#include <iostream>
#include <aclnn/acl_meta.h>
#include <chrono>
#include "acl/acl_rt.h"
#include "acl/acl.h"
#include "graph/operator_reg.h"
#include "../utils/data_utils.h"
#include "aclnn_deep_norm.h"

using namespace std::chrono;

#define FILESYS std::filesystem
const std::string OP_NAME = "DeepNorm";

void readFile2Tensor(const std::string &filePath, size_t dataSize, void *tensorData) {
  void *hostData;
  auto ret = aclrtMallocHost(&hostData, dataSize);
  if (ret != ACL_SUCCESS) {
    ERROR_LOG("aclrtMallocHost for %s failed", filePath.c_str());
  }
  size_t fileSize;
  auto success = ReadFile(filePath, fileSize, hostData, dataSize);
  if (!success) {
    ERROR_LOG("ReadFile from %s failed", filePath.c_str());
  }
  ret = aclrtMemcpy(tensorData, dataSize, hostData, dataSize, ACL_MEMCPY_HOST_TO_DEVICE);
  if (ret != ACL_SUCCESS) {
    ERROR_LOG("aclrtMemcpy for %s failed", filePath.c_str());
  }
  aclrtFreeHost(hostData);
}

void writeTensor2File(const std::string &filePath, size_t dataSize, void *tensorData) {
  void *hostData;
  auto ret = aclrtMallocHost(&hostData, dataSize);
  if (ret != ACL_SUCCESS) {
    ERROR_LOG("aclrtMallocHost for %s failed", filePath.c_str());
  }
  ret = aclrtMemcpy(hostData, dataSize, tensorData, dataSize, ACL_MEMCPY_DEVICE_TO_HOST);
  if (ret != ACL_SUCCESS) {
    ERROR_LOG("aclrtMemcpy for %s failed", filePath.c_str());
  }
  auto success = WriteFile(filePath, hostData, dataSize);
  if (!success) {
    ERROR_LOG("WriteFile to %s failed", filePath.c_str());
  }
  aclrtFreeHost(hostData);
}

bool runOp(std::string &workingDirPath, uint32_t B, uint32_t S, uint32_t H, std::string dataType) {
  auto type = ACL_FLOAT16;

  if (dataType == "fp32") {
    type = ACL_FLOAT;
  }
  if (dataType == "bf16") {
    type = ACL_BF16;
  }

  auto inputXPath = FILESYS::path(workingDirPath) / "x.bin";
  auto inputGXPath = FILESYS::path(workingDirPath) / "gx.bin";
  auto inputGammaPath = FILESYS::path(workingDirPath) / "gamma.bin";
  auto inputBetaPath = FILESYS::path(workingDirPath) / "beta.bin";
  auto outputYPath = FILESYS::path(workingDirPath) / "y.bin";
  auto outputMeanPath = FILESYS::path(workingDirPath) / "mean.bin";
  auto outputRstdPath = FILESYS::path(workingDirPath) / "rstd.bin";
  if (!FILESYS::exists(inputXPath)) {
    std::cout << "File " << inputXPath << " Not Exists" << std::endl;
    return -1;
  }
  if (!FILESYS::exists(inputGXPath)) {
    std::cout << "File " << inputGXPath << " Not Exists" << std::endl;
    return -1;
  }
  if (!FILESYS::exists(inputGammaPath)) {
    std::cout << "File " << inputGammaPath << " Not Exists" << std::endl;
    return -1;
  }
  if (!FILESYS::exists(inputBetaPath)) {
    std::cout << "File " << inputBetaPath << " Not Exists" << std::endl;
    return -1;
  }
  aclInit(nullptr);
  aclrtSetDevice(0);
  aclrtStream stream;
  aclrtCreateStream(&stream);
  // init tensor desc
  std::vector<int64_t> inputShape1{B, S, H};
  std::vector<int64_t> inputShape2{H};
  std::vector<int64_t> inputShape3{B, S, 1};
  aclTensorDesc *inputDescs[4];
  // x
  inputDescs[0] = aclCreateTensorDesc(type, inputShape1.size(), inputShape1.data(), ACL_FORMAT_ND);
  // gx
  inputDescs[1] = aclCreateTensorDesc(type, inputShape1.size(), inputShape1.data(), ACL_FORMAT_ND);
  // beta
  inputDescs[2] = aclCreateTensorDesc(type, inputShape2.size(), inputShape2.data(), ACL_FORMAT_ND);
  // gamma
  inputDescs[3] = aclCreateTensorDesc(type, inputShape2.size(), inputShape2.data(), ACL_FORMAT_ND);

  aclTensorDesc *outputDescs[3];

  // mean
  outputDescs[0] = aclCreateTensorDesc(ACL_FLOAT, inputShape3.size(), inputShape3.data(), ACL_FORMAT_ND);
  // rstd
  outputDescs[1] = aclCreateTensorDesc(ACL_FLOAT, inputShape3.size(), inputShape3.data(), ACL_FORMAT_ND);
  // y
  outputDescs[2] = aclCreateTensorDesc(type, inputShape1.size(), inputShape1.data(), ACL_FORMAT_ND);

  // input
  void *inputX, *inputGX, *inputBeta, *inputGamma;
  auto ret = aclrtMalloc(&inputX, B * S * H * aclDataTypeSize(aclGetTensorDescType(inputDescs[0])), ACL_MEM_MALLOC_HUGE_FIRST);
  if (ret != ACL_SUCCESS) {
    ERROR_LOG("aclrtMalloc X failed");
  }
  ret = aclrtMalloc(&inputGX, B * S * H * aclDataTypeSize(aclGetTensorDescType(inputDescs[1])), ACL_MEM_MALLOC_HUGE_FIRST);
  if (ret != ACL_SUCCESS) {
    ERROR_LOG("aclrtMalloc GX failed");
  }
  ret = aclrtMalloc(&inputBeta, H * aclDataTypeSize(aclGetTensorDescType(inputDescs[2])), ACL_MEM_MALLOC_HUGE_FIRST);
  if (ret != ACL_SUCCESS) {
    ERROR_LOG("aclrtMalloc Beta failed");
  }
  ret = aclrtMalloc(&inputGamma, H * aclDataTypeSize(aclGetTensorDescType(inputDescs[3])), ACL_MEM_MALLOC_HUGE_FIRST);
  if (ret != ACL_SUCCESS) {
    ERROR_LOG("aclrtMalloc Gamma failed");
  }
  // output
  void *outputY, *outputMean, *outputRstd;
  ret = aclrtMalloc(&outputMean, B * S * aclDataTypeSize(aclGetTensorDescType(outputDescs[0])), ACL_MEM_MALLOC_HUGE_FIRST);
  if (ret != ACL_SUCCESS) {
    ERROR_LOG("aclrtMalloc Mean failed");
  }
  ret = aclrtMalloc(&outputRstd, B * S * aclDataTypeSize(aclGetTensorDescType(outputDescs[1])), ACL_MEM_MALLOC_HUGE_FIRST);
  if (ret != ACL_SUCCESS) {
    ERROR_LOG("aclrtMalloc Rstd failed");
  }
  ret = aclrtMalloc(&outputY, B * S * H * aclDataTypeSize(aclGetTensorDescType(outputDescs[2])), ACL_MEM_MALLOC_HUGE_FIRST);
  if (ret != ACL_SUCCESS) {
    ERROR_LOG("aclrtMalloc Y failed");
  }

  aclTensor *inputXTensor = aclCreateTensor(inputShape1.data(), inputShape1.size(),
                                             aclGetTensorDescType(inputDescs[0]),
                                             nullptr, 0, aclGetTensorDescFormat(inputDescs[0]), inputShape1.data(),
                                             inputShape1.size(), inputX);
  aclTensor *inputGXTensor = aclCreateTensor(inputShape1.data(), inputShape1.size(),
                                             aclGetTensorDescType(inputDescs[1]),
                                             nullptr, 0, aclGetTensorDescFormat(inputDescs[1]), inputShape1.data(),
                                             inputShape1.size(), inputGX);
  aclTensor *inputBetaTensor = aclCreateTensor(inputShape2.data(), inputShape2.size(),
                                               aclGetTensorDescType(inputDescs[2]),
                                               nullptr, 0, aclGetTensorDescFormat(inputDescs[2]), inputShape2.data(),
                                               inputShape2.size(), inputBeta);
  aclTensor *inputGammaTensor = aclCreateTensor(inputShape2.data(), inputShape2.size(),
                                                aclGetTensorDescType(inputDescs[3]),
                                                nullptr, 0, aclGetTensorDescFormat(inputDescs[3]), inputShape2.data(),
                                                inputShape2.size(), inputGamma);


  aclTensor *outputMeanTensor = aclCreateTensor(inputShape3.data(), inputShape3.size(),
                                                aclGetTensorDescType(outputDescs[0]),
                                                nullptr, 0, aclGetTensorDescFormat(outputDescs[0]), inputShape3.data(),
                                                inputShape3.size(), outputMean);
  aclTensor *outputRstdTensor = aclCreateTensor(inputShape3.data(), inputShape3.size(),
                                                aclGetTensorDescType(outputDescs[1]),
                                                nullptr, 0, aclGetTensorDescFormat(outputDescs[1]), inputShape3.data(),
                                                inputShape3.size(), outputRstd);
  aclTensor *outputYTensor = aclCreateTensor(inputShape1.data(), inputShape1.size(),
                                             aclGetTensorDescType(outputDescs[2]),
                                             nullptr, 0, aclGetTensorDescFormat(outputDescs[2]), inputShape1.data(),
                                             inputShape1.size(), outputY); 

  readFile2Tensor(inputXPath, B * S * H * aclDataTypeSize(aclGetTensorDescType(inputDescs[0])), inputX);
  readFile2Tensor(inputGXPath, B * S * H * aclDataTypeSize(aclGetTensorDescType(inputDescs[1])), inputGX);
  readFile2Tensor(inputBetaPath, H * aclDataTypeSize(aclGetTensorDescType(inputDescs[2])), inputBeta);
  readFile2Tensor(inputGammaPath, H * aclDataTypeSize(aclGetTensorDescType(inputDescs[3])), inputGamma);

  size_t workspaceSize = 0;
  aclOpExecutor *handle = nullptr;
  float alpha = 0.3;
  float eps = 1e-6;
  ret = aclnnDeepNormGetWorkspaceSize(inputXTensor, inputGXTensor, inputBetaTensor, inputGammaTensor,
                                               alpha, eps,
                                               outputMeanTensor, outputRstdTensor, outputYTensor,
                                               &workspaceSize, &handle);
  if (ret != ACL_SUCCESS) {
    (void) aclrtDestroyStream(stream);
    ERROR_LOG("Get Operator Workspace failed. error code is %d", static_cast<int32_t>(ret));
    return false;
  }
  INFO_LOG("Execute aclnnDeepNormGetWorkspaceSize success, workspace size %lu", workspaceSize);

  void *workspace = nullptr;
  if (workspaceSize != 0) {
    if (aclrtMalloc(&workspace, workspaceSize, ACL_MEM_MALLOC_NORMAL_ONLY) != ACL_SUCCESS) {
      ERROR_LOG("Malloc device memory failed");
    }
  outputDescs[2] = aclCreateTensorDesc(type, inputShape1.size(), inputShape1.data(), ACL_FORMAT_ND);

  // input
  void *inputX, *inputGX, *inputBeta, *inputGamma;
  auto ret = aclrtMalloc(&inputX, B * S * H * aclDataTypeSize(aclGetTensorDescType(inputDescs[0])), ACL_MEM_MALLOC_HUGE_FIRST);
  if (ret != ACL_SUCCESS) {
    ERROR_LOG("aclrtMalloc X failed");
  }
  ret = aclrtMalloc(&inputGX, B * S * H * aclDataTypeSize(aclGetTensorDescType(inputDescs[1])), ACL_MEM_MALLOC_HUGE_FIRST);
  if (ret != ACL_SUCCESS) {
    ERROR_LOG("aclrtMalloc GX failed");
  }
  ret = aclrtMalloc(&inputBeta, H * aclDataTypeSize(aclGetTensorDescType(inputDescs[2])), ACL_MEM_MALLOC_HUGE_FIRST);
  if (ret != ACL_SUCCESS) {
    ERROR_LOG("aclrtMalloc Beta failed");
  }
  ret = aclrtMalloc(&inputGamma, H * aclDataTypeSize(aclGetTensorDescType(inputDescs[3])), ACL_MEM_MALLOC_HUGE_FIRST);
  if (ret != ACL_SUCCESS) {
    ERROR_LOG("aclrtMalloc Gamma failed");
  }
  // output
  void *outputY, *outputMean, *outputRstd;
  ret = aclrtMalloc(&outputMean, B * S * aclDataTypeSize(aclGetTensorDescType(outputDescs[0])), ACL_MEM_MALLOC_HUGE_FIRST);
  if (ret != ACL_SUCCESS) {
    ERROR_LOG("aclrtMalloc Mean failed");
  }
  ret = aclrtMalloc(&outputRstd, B * S * aclDataTypeSize(aclGetTensorDescType(outputDescs[1])), ACL_MEM_MALLOC_HUGE_FIRST);
  if (ret != ACL_SUCCESS) {
    ERROR_LOG("aclrtMalloc Rstd failed");
  }
  ret = aclrtMalloc(&outputY, B * S * H * aclDataTypeSize(aclGetTensorDescType(outputDescs[2])), ACL_MEM_MALLOC_HUGE_FIRST);
  if (ret != ACL_SUCCESS) {
    ERROR_LOG("aclrtMalloc Y failed");
  }

  aclTensor *inputXTensor = aclCreateTensor(inputShape1.data(), inputShape1.size(),
                                             aclGetTensorDescType(inputDescs[0]),
                                             nullptr, 0, aclGetTensorDescFormat(inputDescs[0]), inputShape1.data(),
                                             inputShape1.size(), inputX);
  aclTensor *inputGXTensor = aclCreateTensor(inputShape1.data(), inputShape1.size(),
                                             aclGetTensorDescType(inputDescs[1]),
                                             nullptr, 0, aclGetTensorDescFormat(inputDescs[1]), inputShape1.data(),
                                             inputShape1.size(), inputGX);
  aclTensor *inputBetaTensor = aclCreateTensor(inputShape2.data(), inputShape2.size(),
                                               aclGetTensorDescType(inputDescs[2]),
                                               nullptr, 0, aclGetTensorDescFormat(inputDescs[2]), inputShape2.data(),
                                               inputShape2.size(), inputBeta);
  aclTensor *inputGammaTensor = aclCreateTensor(inputShape2.data(), inputShape2.size(),
                                                aclGetTensorDescType(inputDescs[3]),
                                                nullptr, 0, aclGetTensorDescFormat(inputDescs[3]), inputShape2.data(),
                                                inputShape2.size(), inputGamma);


  aclTensor *outputMeanTensor = aclCreateTensor(inputShape3.data(), inputShape3.size(),
                                                aclGetTensorDescType(outputDescs[0]),
                                                nullptr, 0, aclGetTensorDescFormat(outputDescs[0]), inputShape3.data(),
                                                inputShape3.size(), outputMean);
  aclTensor *outputRstdTensor = aclCreateTensor(inputShape3.data(), inputShape3.size(),
                                                aclGetTensorDescType(outputDescs[1]),
                                                nullptr, 0, aclGetTensorDescFormat(outputDescs[1]), inputShape3.data(),
                                                inputShape3.size(), outputRstd);
  aclTensor *outputYTensor = aclCreateTensor(inputShape1.data(), inputShape1.size(),
                                             aclGetTensorDescType(outputDescs[2]),
                                             nullptr, 0, aclGetTensorDescFormat(outputDescs[2]), inputShape1.data(),
                                             inputShape1.size(), outputY); 

  readFile2Tensor(inputXPath, B * S * H * aclDataTypeSize(aclGetTensorDescType(inputDescs[0])), inputX);
  readFile2Tensor(inputGXPath, B * S * H * aclDataTypeSize(aclGetTensorDescType(inputDescs[1])), inputGX);
  readFile2Tensor(inputBetaPath, H * aclDataTypeSize(aclGetTensorDescType(inputDescs[2])), inputBeta);
  readFile2Tensor(inputGammaPath, H * aclDataTypeSize(aclGetTensorDescType(inputDescs[3])), inputGamma);

  size_t workspaceSize = 0;
  aclOpExecutor *handle = nullptr;
  float alpha = 0.3;
  float eps = 1e-6;
  ret = aclnnDeepNormGetWorkspaceSize(inputXTensor, inputGXTensor, inputBetaTensor, inputGammaTensor,
                                               alpha, eps,
                                               outputMeanTensor, outputRstdTensor, outputYTensor,
                                               &workspaceSize, &handle);
  if (ret != ACL_SUCCESS) {
    (void) aclrtDestroyStream(stream);
    ERROR_LOG("Get Operator Workspace failed. error code is %d", static_cast<int32_t>(ret));
    return false;
  }
  INFO_LOG("Execute aclnnDeepNormGetWorkspaceSize success, workspace size %lu", workspaceSize);

  void *workspace = nullptr;
  if (workspaceSize != 0) {
    if (aclrtMalloc(&workspace, workspaceSize, ACL_MEM_MALLOC_NORMAL_ONLY) != ACL_SUCCESS) {
      ERROR_LOG("Malloc device memory failed");
    }
  }

  auto start = high_resolution_clock::now();
  if (ret = aclnnDeepNorm(workspace, workspaceSize, handle, stream); ret != ACL_SUCCESS) {
    (void) aclrtDestroyStream(stream);
    ERROR_LOG("Execute Operator failed. error code is %d", static_cast<int32_t>(ret));
    return false;
  }
  auto stop = high_resolution_clock::now();
  auto duration = duration_cast<microseconds>(stop - start);
  INFO_LOG("Execute aclnnDeepNorm success");
  std::cout << "aclnnDeepNorm took: "
            << duration.count() << " microseconds" << std::endl;

  ret = aclrtSynchronizeStream(stream);

  if (ret != ACL_SUCCESS) {
    ERROR_LOG("aclrtSynchronizeStream failed");
  } else {
    INFO_LOG("aclrtSynchronizeStream success");
  }

  writeTensor2File(outputMeanPath, B * S * aclDataTypeSize(aclGetTensorDescType(outputDescs[0])), outputMean);
  writeTensor2File(outputRstdPath, B * S * aclDataTypeSize(aclGetTensorDescType(outputDescs[1])), outputRstd);
  writeTensor2File(outputYPath, B * S * H * aclDataTypeSize(aclGetTensorDescType(outputDescs[2])), outputY);

  aclrtFree(inputX);
  aclrtFree(inputGX);
  aclrtFree(inputBeta);
  aclrtFree(inputGamma);
  aclrtFree(outputY);
  aclrtFree(outputMean);
  aclrtFree(outputRstd);

  aclrtDestroyStream(&stream);
  aclrtResetDevice(0);
  aclFinalize();
  INFO_LOG("end.");
  return 0;
}

int main(int argc, char **argv) {
  if (argc != 6) {
    ERROR_LOG("Invalid Arguments");
    return -1;
  }
  auto workingDirPath = std::string(argv[1]);
  uint32_t B = atoi(argv[2]);
  uint32_t S = atoi(argv[3]);
  uint32_t H = atoi(argv[4]);
  auto inputType = std::string(argv[5]);
  std::filesystem::path workingDir{workingDirPath};
  if (!std::filesystem::exists(workingDir)) {
    std::cout << "Directory " << workingDirPath << " Not Exists" << std::endl;
    return -1;
  }
  std::cout << "B:" << B << "  S:" << S <<"  H:" << H << "  Input Type:" << inputType << std::endl;
  runOp(workingDirPath, B, S, H, inputType);
  return 0;
}

```

## 实现说明<a name="section19406171715553"></a>

计算图:

![](figures/zh-cn_image_0000001711977804.png)

