# aclnnEmbeddingBag

## 支持的产品型号

- Atlas 训练产品。
- Atlas A2训练系列产品。

## 接口原型
每个算子分为[两段式接口](common/两段式接口.md)，必须先调用“aclnnEmbeddingBagGetWorkspaceSize”接口获取计算所需workspace大小以及包含了算子计算流程的执行器，再调用“aclnnEmbeddingBag”接口执行计算。
* `aclnnStatus aclnnEmbeddingBagGetWorkspaceSize(const aclTensor *weight, const aclTensor *indices,const aclTensor *offsets, bool scaleGradByFreq,int64_t mode, bool sparse, const aclTensor* perSampleWeights, bool includeLastOffset,int64_t paddingIdx, aclTensor* output, aclTensor* offset2bag, aclTensor* bagSize, aclTensor* maxIndices, uint64_t* workspaceSize, aclOpExecutor** executor)`
* `aclnnStatus aclnnEmbeddingBag(void *workspace, uint64_t workspaceSize, aclOpExecutor *executor, aclrtStream stream)`

## 功能描述

根据indices从weight中获得一组被聚合的数，然后根据offsets的偏移和mode指定的聚合模式对获取的数进行max、sum、mean等聚合。其余参数则更细化了计算过程的控制。

## aclnnEmbeddingBagGetWorkspaceSize

- **参数说明**：

  * weight(const aclTensor*, 计算输入)：Device侧的aclTensor，输入数据类型支持FLOAT、BFLOAT16，必须是2D Tensor，支持[非连续的Tensor](common/非连续的Tensor.md)，[数据格式](common/数据格式.md)支持ND。
  * indices（const aclTensor*, 计算输入）：npu device侧的aclTensor，数据类型支持UINT8、INT8、INT16、INT32、INT64。必须是1D或2D tensor。支持[非连续的Tensor](common/非连续的Tensor.md)，[数据格式](common/数据格式.md)支持ND。
  * offsets(const aclTensor*, 计算输入): npu device侧的aclTensor，数据类型支持UINT8、INT8、INT16、INT32、INT64，但和indices必须有一个是INT32或INT64,在indices是1D时，offsets必须是1D tensor。
  * scaleGradByFreq(bool，计算输入): 用于控制缩放梯度。
  * mode(int64, 计算输入)：用于控制聚合模式。0表示sum聚合模式，1表示mean聚合模式，其他表示max聚合模式。
  * sparse(bool, 计算输入)：用于控制稀疏模式。
  * perSampleWeights(const aclTensor*, 计算输入): npu device侧的aclTensor，指定样本权重。必须是1D tensor，数据类型与weight一致，仅在sum可以不是null。
  * includeLastOffset(bool, 计算输入)：控制是否包含最后的偏移。
  * paddingIdx(int64_t,计算输入): 取值范围是[-n,n-1]，其中n是weigit第一维元素个数。
  * output(aclTensor*, 计算输出)：形状由offsets及includeLastOffset控制，数据类型与weight一致，支持非连续tensor，数据格式支持ND。
  * offset2bag(aclTensor*, 计算输出)：数据类型和indices与offsets中精度高的一致,支持非连续Tensor, 数据格式支持ND.
  * bagSize(aclTensor*, 计算输出): 数据类型和indices与offsets中精度高的一致,支持非连续Tensor, 数据格式支持ND.
  * maxIndices(aclTensor*, 计算输出): 数据类型和indices与offsets中精度高的一致,支持非连续Tensor, 数据格式支持ND.
  * workspaceSize(uint64_t*, 出参)：返回用户需要在Device侧申请的workspace大小。
  * executor(aclOpExecutor**, 出参)：返回op执行器，包含了算子计算流程。

- **返回值**：

  aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

  ```
  第一段接口完成入参校验，出现以下场景时报错：
  (161001)ACLNN_ERR_PARAM_NULLPTR: 1. 传入的 weight,indices,offsets,output,offset2bag,bagSize,maxIndices是空指针。

  (161002)ACLNN_ERR_PARAM_INVALID: 1. weight数据类型不在支持范围内,weight不是2D Tensor。
                                   2. indices数据类型不在支持范围内,indices不是1D Tensor或2D Tensor。
                                   3. offsets数据类型不在支持范围内, 在indices是1D Tensor时offsets不是1D Tensor。
                                   4. perSampleWeights数据类型与weight不一致, perSampleWeights不是1D Tensor, perSampleWeights元素数量与indices不相等, 在非sum模式perSampleWeights不是nullptr。
                                   5. paddingIdx超出范围.
                                   6. output数据类型与weight不一致,形状与定义不 符.
                                   7. offset2bag,bagSize,maxIndices数据类型与形状和定义不符.
  ```

## aclnnEmbeddingBag

- **参数说明**：

  * workspace(void*, 入参)：在Device侧申请的workspace内存起址。
  * workspaceSize(uint64_t, 入参)：在Device侧申请的workspace大小，由第一段接口aclnnEmbeddingBagGetWorkspaceSize获取。
  * executor(aclOpExecutor*, 入参)：op执行器，包含了算子计算流程。
  * stream(aclrtStream, 入参)：指定执行任务的 AscendCL Stream流。

- **返回值**：

  aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

## 约束与限制

无

## 调用示例
示例代码如下，仅供参考，具体编译和执行过程请参考[编译与运行样例](common/编译与运行样例.md)。
```c++
#include <iostream>
#include <vector>
#include "acl/acl.h"
#include "aclnnop/aclnn_embedding_bag.h"

#define CHECK_RET(cond, return_expr) \
  do {                               \
    if (!(cond)) {                   \
      return_expr;                   \
    }                                \
  } while (0)

#define LOG_PRINT(message, ...)     \
  do {                              \
    printf(message, ##__VA_ARGS__); \
  } while (0)

int64_t GetShapeSize(const std::vector<int64_t>& shape) {
  int64_t shapeSize = 1;
  for (auto i : shape) {
    shapeSize *= i;
  }
  return shapeSize;
}

int Init(int32_t deviceId, aclrtContext* context, aclrtStream* stream) {
  // 固定写法，AscendCL初始化
  auto ret = aclInit(nullptr);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclInit failed. ERROR: %d\n", ret); return ret);
  ret = aclrtSetDevice(deviceId);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSetDevice failed. ERROR: %d\n", ret); return ret);
  ret = aclrtCreateContext(context, deviceId);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtCreateContext failed. ERROR: %d\n", ret); return ret);
  ret = aclrtSetCurrentContext(*context);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSetCurrentContext failed. ERROR: %d\n", ret); return ret);
  ret = aclrtCreateStream(stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtCreateStream failed. ERROR: %d\n", ret); return ret);
  return 0;
}

template <typename T>
int CreateAclTensor(const std::vector<T>& hostData, const std::vector<int64_t>& shape, void** deviceAddr,
                    aclDataType dataType, aclTensor** tensor) {
  auto size = GetShapeSize(shape) * sizeof(T);
  // 调用aclrtMalloc申请device侧内存
  auto ret = aclrtMalloc(deviceAddr, size, ACL_MEM_MALLOC_HUGE_FIRST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtMalloc failed. ERROR: %d\n", ret); return ret);
  // 调用aclrtMemcpy将host侧数据拷贝到device侧内存上
  ret = aclrtMemcpy(*deviceAddr, size, hostData.data(), size, ACL_MEMCPY_HOST_TO_DEVICE);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtMemcpy failed. ERROR: %d\n", ret); return ret);

  // 计算连续tensor的strides
  std::vector<int64_t> strides(shape.size(), 1);
  for (int64_t i = shape.size() - 2; i >= 0; i--) {
    strides[i] = shape[i + 1] * strides[i + 1];
  }

  // 调用aclCreateTensor接口创建aclTensor
  *tensor = aclCreateTensor(shape.data(), shape.size(), dataType, strides.data(), 0, aclFormat::ACL_FORMAT_ND,
                            shape.data(), shape.size(), *deviceAddr);
  return 0;
}

int main() {
  // 1. （固定写法）device/context/stream初始化，参考AscendCL对外接口列表
  // 根据自己的实际device填写deviceId
  int32_t deviceId = 0;
  aclrtContext context;
  aclrtStream stream;
  auto ret = Init(deviceId, &context, &stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("Init AscendCL failed. ERROR: %d\n", ret); return ret);

  // 2. 构造输入与输出，需要根据API的接口自定义构造
  std::vector<int64_t> weightShape = {3, 3};
  std::vector<int64_t> indicesShape = {6};
  std::vector<int64_t> offsetsShape = {4};
  std::vector<int64_t> perSampleWeightsShape = {6};
  std::vector<int64_t> outputShape = {4, 3};
  std::vector<int64_t> offset2bagShape = {6};
  std::vector<int64_t> bagSizeShape = {4};
  std::vector<int64_t> maxIndicesShape = {4};

  void* weightDeviceAddr = nullptr;
  void* indicesDeviceAddr = nullptr;
  void* offsetsDeviceAddr = nullptr;
  void* perSampleWeightsDeviceAddr = nullptr;
  void* outputDeviceAddr = nullptr;
  void* offset2bagDeviceAddr = nullptr;
  void* bagSizeDeviceAddr = nullptr;
  void* maxIndicesDeviceAddr = nullptr;

  aclTensor* weight = nullptr;
  aclTensor* indices = nullptr;
  aclTensor* offsets = nullptr;
  aclTensor* perSampleWeights = nullptr;
  aclTensor* output = nullptr;
  aclTensor* offset2bag = nullptr;
  aclTensor* bagSize = nullptr;
  aclTensor* maxIndices = nullptr;

  std::vector<float> weightHostData = {1, 1, 1, 1, 1, 1, 1, 1, 1};
  std::vector<int64_t> indicesHostData = {1, 2, 0, 2, 2, 1};
  std::vector<int64_t> offsetsHostData = {0, 2, 4, 5};
  std::vector<float> perSampleWeightsHostData = {1, 1, 1, 1, 1, 1};
  std::vector<float> outputHostData = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
  std::vector<int64_t> offset2bagHostData = {0, 0, 0, 0, 0, 0};
  std::vector<int64_t> bagSizeHostData = {0, 0, 0, 0};
  std::vector<int64_t> maxIndicesHostData = {0, 0, 0, 0};

  //创建weight aclTensor
  ret = CreateAclTensor(weightHostData, weightShape, &weightDeviceAddr, aclDataType::ACL_FLOAT, &weight);
  CHECK_RET(ret == ACL_SUCCESS, return ret);

  //创建indices aclTensor
  ret = CreateAclTensor(indicesHostData, indicesShape, &indicesDeviceAddr, aclDataType::ACL_INT64, &indices);
  CHECK_RET(ret == ACL_SUCCESS, return ret);

  //创建offsets aclTensor
  ret = CreateAclTensor(offsetsHostData, offsetsShape, &offsetsDeviceAddr, aclDataType::ACL_INT64, &offsets);
  CHECK_RET(ret == ACL_SUCCESS, return ret);

  //创建perSampleWeights aclTensor
  ret = CreateAclTensor(perSampleWeightsHostData, perSampleWeightsShape, &perSampleWeightsDeviceAddr, aclDataType::ACL_FLOAT, &perSampleWeights);
  CHECK_RET(ret == ACL_SUCCESS, return ret);

  //创建output aclTensor
  ret = CreateAclTensor(outputHostData, outputShape, &outputDeviceAddr, aclDataType::ACL_FLOAT, &output);
  CHECK_RET(ret == ACL_SUCCESS, return ret);

  //创建offset2bag aclTensor
  ret = CreateAclTensor(offset2bagHostData, offset2bagShape, &offset2bagDeviceAddr, aclDataType::ACL_INT64, &offset2bag);
  CHECK_RET(ret == ACL_SUCCESS, return ret);

  //创建bagSize aclTensor
  ret = CreateAclTensor(bagSizeHostData, bagSizeShape, &bagSizeDeviceAddr, aclDataType::ACL_INT64, &bagSize);
  CHECK_RET(ret == ACL_SUCCESS, return ret);

  //创建maxIndices aclTensor
  ret = CreateAclTensor(maxIndicesHostData, maxIndicesShape, &maxIndicesDeviceAddr, aclDataType::ACL_INT64, &maxIndices);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  
  //非tensor参数
  bool scaleGradByFreq = false;
  int64_t mode = 0;
  bool sparse = false;
  bool includeLastOffset = false;
  int64_t paddingIdx = 1;

  // 3. 调用CANN算子库API，需要修改为具体的Api名称
  uint64_t workspaceSize = 0;
  aclOpExecutor* executor;
  // 调用aclnnEmbeddingBag第一段接口
  ret = aclnnEmbeddingBagGetWorkspaceSize(weight, indices, offsets, scaleGradByFreq, mode, sparse, perSampleWeights,
            includeLastOffset, paddingIdx, output, offset2bag, bagSize, maxIndices, &workspaceSize, &executor);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclnnEmbeddingBagGetWorkspaceSize failed. ERROR: %d\n", ret); return ret);
  // 根据第一段接口计算出的workspaceSize申请device内存
  void* workspaceAddr = nullptr;
  if (workspaceSize > 0) {
    ret = aclrtMalloc(&workspaceAddr, workspaceSize, ACL_MEM_MALLOC_HUGE_FIRST);
    CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("allocate workspace failed. ERROR: %d\n", ret); return ret);
  }
  // 调用aclnnEmbeddingBag第二段接口
  ret = aclnnEmbeddingBag(workspaceAddr, workspaceSize, executor, stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclnnEmbeddingBag failed. ERROR: %d\n", ret); return ret);

  // 4. （固定写法）同步等待任务执行结束
  ret = aclrtSynchronizeStream(stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSynchronizeStream failed. ERROR: %d\n", ret); return ret);

  // 5. 获取输出的值，将device侧内存上的结果拷贝至host侧，需要根据具体API的接口定义修改
  auto outputSize = GetShapeSize(outputShape);
  std::vector<float> outputResultData(outputSize, 0);
  ret = aclrtMemcpy(outputResultData.data(), outputResultData.size() * sizeof(outputResultData[0]), outputDeviceAddr,
                    outputSize * sizeof(float), ACL_MEMCPY_DEVICE_TO_HOST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("copy result from device to host failed. ERROR: %d\n", ret); return ret);
  for (int64_t i = 0; i < outputSize; i++) {
    LOG_PRINT("outputResult[%ld] is: %f\n", i, outputResultData[i]);
  }

  auto offset2bagSize = GetShapeSize(offset2bagShape);
  std::vector<int64_t> offset2bagResultData(offset2bagSize, 0);
  ret = aclrtMemcpy(offset2bagResultData.data(), offset2bagResultData.size() * sizeof(offset2bagResultData[0]), offset2bagDeviceAddr,
                    offset2bagSize * sizeof(int64_t), ACL_MEMCPY_DEVICE_TO_HOST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("copy result from device to host failed. ERROR: %d\n", ret); return ret);
  for (int64_t i = 0; i < offset2bagSize; i++) {
    LOG_PRINT("offset2bagResult[%ld] is: %ld\n", i, offset2bagResultData[i]);
  }

  auto bagSizeSize = GetShapeSize(bagSizeShape);
  std::vector<int64_t> bagSizeResultData(bagSizeSize, 0);
  ret = aclrtMemcpy(bagSizeResultData.data(), bagSizeResultData.size() * sizeof(bagSizeResultData[0]), bagSizeDeviceAddr,
                    bagSizeSize * sizeof(int64_t), ACL_MEMCPY_DEVICE_TO_HOST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("copy result from device to host failed. ERROR: %d\n", ret); return ret);
  for (int64_t i = 0; i < bagSizeSize; i++) {
    LOG_PRINT("bagSizeResult[%ld] is: %ld\n", i, bagSizeResultData[i]);
  }

  auto maxIndicesSize = GetShapeSize(maxIndicesShape);
  std::vector<int64_t> maxIndicesResultData(maxIndicesSize, 0);
  ret = aclrtMemcpy(maxIndicesResultData.data(), maxIndicesResultData.size() * sizeof(maxIndicesResultData[0]), maxIndicesDeviceAddr,
                    maxIndicesSize * sizeof(int64_t), ACL_MEMCPY_DEVICE_TO_HOST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("copy result from device to host failed. ERROR: %d\n", ret); return ret);
  for (int64_t i = 0; i < maxIndicesSize; i++) {
    LOG_PRINT("maxIndicesResult[%ld] is: %ld\n", i, maxIndicesResultData[i]);
  }

  // 6. 释放aclTensor和aclScalar，需要根据具体API的接口定义修改
  aclDestroyTensor(weight);
  aclDestroyTensor(indices);
  aclDestroyTensor(offsets);
  aclDestroyTensor(perSampleWeights);
  aclDestroyTensor(output);
  aclDestroyTensor(offset2bag);
  aclDestroyTensor(bagSize);
  aclDestroyTensor(maxIndices);
  
  // 7. 释放device资源, 需要根据具体API的接口定义修改
  aclrtFree(weightDeviceAddr);
  aclrtFree(indicesDeviceAddr);
  aclrtFree(offsetsDeviceAddr);
  aclrtFree(perSampleWeightsDeviceAddr);
  aclrtFree(outputDeviceAddr);
  aclrtFree(offset2bagDeviceAddr);
  aclrtFree(bagSizeDeviceAddr);
  aclrtFree(maxIndicesDeviceAddr);

  if (workspaceSize > 0) {
    aclrtFree(workspaceAddr);
  }
  aclrtDestroyStream(stream);
  aclrtDestroyContext(context);
  aclrtResetDevice(deviceId);
  aclFinalize();
  return 0;
}
```

