# aclnnFlashAttentionVarLenScore

## 支持的产品型号

Atlas A2训练系列产品。

## 接口原型

每个算子分为两段式接口，必须先调用“aclnnFlashAttentionVarLenScoreGetWorkspaceSize”接口获取计算所需workspace大小以及包含了算子计算流程的执行器，再调用“aclnnFlashAttentionVarLenScore”接口执行计算。

* `aclnnStatus aclnnFlashAttentionVarLenScoreGetWorkspaceSize(const aclTensor *query, const aclTensor *key, const aclTensor *value, const aclTensor *realShiftOptional, const aclTensor *dropMaskOptional, const aclTensor *paddingMaskOptional,const aclTensor *attenMaskOptional, const aclIntArray *prefixOptional, const aclIntArray *actualSeqQLenOptional, const aclIntArray *actualSeqKvLenOptional, double scaleValueOptional, double keepProbOptional, int64_t preTockensOptional, int64_t nextTockensOptional, int64_t headNum, char *inputLayout, int64_t innerPreciseOptional, int64_t sparseModeOptional, const aclTensor *softmaxMaxOut, const aclTensor *softmaxSumOut, const aclTensor *softmaxOutOut, const aclTensor *attentionOutOut, uint64_t *workspaceSize, aclOpExecutor **executor);)`
* `aclnnStatus aclnnFlashAttentionVarLenScore(void *workspace, uint64_t workspaceSize, aclOpExecutor *executor, const aclrtStream stream)`

## 功能描述

-   算子功能：训练场景下，使用FlashAttention算法实现self-attention（自注意力）的计算。跟aclnnFlashAttentionScoreGetWorkspaceSize的区别是该接口支持可变长S的计算。
- 计算公式：

    注意力的正向计算公式如下：

    $$
    Attention(Q,K,V)=Dropout(Softmax(Mask(\frac{QK^T+pse}{\sqrt{d}})))V
    $$


## aclnnFlashAttentionVarLenScoreGetWorkspaceSize

- **参数说明：**
    - query（aclTensor\*，计算输入）：Device侧的aclTensor，公式中的输入Q。数据类型支持：FLOAT16、BFLOAT16。数据类型与key/value的数据类型一致。数据格式支持ND。

    - key（aclTensor\*，计算输入）：Device侧的aclTensor，公式中的输入K。数据类型支持：FLOAT16、BFLOAT16。数据类型与query/value的数据类型一致。数据格式支持ND。

    - value（aclTensor\*，计算输入）：Device侧的aclTensor，公式中的输入V。数据类型支持：FLOAT16、BFLOAT16。数据类型与query/key的数据类型一致。数据格式支持ND。

    - realShiftOptional（aclTensor\*，计算输入）：Device侧的aclTensor，公式中的输入pse，可选参数。数据类型支持：FLOAT16、BFLOAT16。数据类型与query的数据类型一致。数据格式支持ND，输入shape类型需为\[B,N,S,S\]、\[B,N,1,S\]、\[1,N,S,S\]。

        alibi位置编码场景下，非压缩场景支持\[B,N,S,S\]、\[1,N,S,S\]，压缩场景支持\[B,N,H,S\]、\[1,N,H,S\]，H固定为1024。

    - dropMaskOptional（aclTensor\*，计算输入）：Device侧的aclTensor，公式中的dropout，可选参数。数据类型支持：UINT8（标识8个1bit BOOL）。数据格式支持ND。

    - paddingMaskOptional（aclTensor\*，计算输入）：预留参数，暂未使用。

    - attenMaskOptional（aclTensor\*，计算输入）：Device侧的aclTensor，公式中的mask，可选参数。数据类型支持：BOOL。数据格式支持ND，输入shape类型需为\[B,N,S,S\]、\[B,1,S,S\]、\[1,1,S,S\]、[S,S]。

    - prefixOptional（aclIntArray\*，计算输入）：Device侧的aclTensor。可选参数。数据类型支持：INT64。代表prefix稀疏计算场景每个Batch的N值。数据格式支持ND。

    - actualSeqQLenOptional（aclIntArray\*，计算输入）：Device侧的aclTensor。数据类型支持：INT64。数据格式支持ND。描述了每个Batch对应的query S大小。

    - actualSeqKvLenOptional（aclIntArray\*，计算输入）：Device侧的aclTensor。数据类型支持：INT64。数据格式支持ND。描述了每个Batch对应的key/value S大小。

    - scaleValueOptional（double，计算输入）：Host侧的double，公式中d开根号的倒数。数据类型支持：DOUBLE，代表缩放系数，作为计算流中Muls的scalar值。

    - keepProbOptional（double，计算输入）：Host侧的double。数据类型支持：DOUBLE，代表dropMaskOptional中1的比例。

    - preTockensOptional（int64\_t，计算输入）：Host侧的int64_t。数据类型支持：INT64，用于稀疏计算 ，表示slides window的左边界。

    - nextTockensOptional（int64\_t，计算输入）：Host侧的int64_t。数据类型支持：INT64，用于稀疏计算，表示slides window的右边界。

    - headNum（int64\_t，计算输入）：Host侧的int64_t。数据类型支持：INT64，代表单卡的head个数。

    - inputLayout（string\*，计算输入）：Host侧的string。数据类型支持：String，代表输入query、key、value的数据排布格式，支持TND。

        **说明：** 
        query、key、value数据排布格式支持从多种维度解读，其中B（Batch）表示输入样本批量大小、S（Seq-Length）表示输入样本序列长度、H（Head-Size）表示隐藏层的大小、N（Head-Num）表示多头数、D（Head-Dim）表示隐藏层最小的单元尺寸，且满足D=H/N。

    - innerPreciseOptional（int64_t，计算输入）：Host侧的int64_t。数据类型支持：INT64，用于提升精度，默认配置为0即可。

        **说明：** 
        当前0、1为保留配置值，当计算过程中存在整行mask进而导致精度有损失时，可以尝试将该参数配置为2以提升精度，但是该配置可能会导致性能下降。

    - sparseModeOptional（int64\_t，计算输入）：Host侧的int64_t。数据类型支持：INT64。用户不特意指定时可传入默认值：0。

        -   sparseModeOptional为0时，代表defaultMask模式，如果attenMaskOptional未传入则不做mask操作，忽略preTockensOptional和nextTockensOptional\(内部赋值为INT\_MAX\)；如果传入，则需要传入完整的attenMaskOptional矩阵（S1max \* S2max2），表示preTockensOptional和nextTockensOptional之间的部分需要计算。
        -   sparseModeOptional为1时，代表allMask，即传入完整的attenMaskOptional矩阵。
        -   sparseModeOptional为2时，代表leftUpCausal模式的mask，对应以左上顶点为划分的下三角场景，需要传入shape为\[2048, 2048\]的下三角attenMaskOptional矩阵。
        -   sparseModeOptional为3时，代表rightDownCausal模式的mask，对应以右下顶点为划分的下三角场景，需要传入shape为\[2048, 2048\]的下三角attenMaskOptional矩阵。
        -   sparseModeOptional为4时，代表band场景，即计算preTockensOptional和nextTockensOptional之间的部分，需要传入shape为\[2048, 2048\]的下三角attenMaskOptional矩阵。
        -   sparseModeOptional为5、6、7、8时，分别代表代表prefix、global、dilated、block\_local，**均暂不支持**。
        -   sparseModeOptional为2、3、4时，需要S等长，不等长暂不支持。

        **说明：**
        当所有的attenmask的shape小于2048且相同的时候，建议使用default模式，来减少内存使用量；sparse\_mode配置为1、2、3、5时，用户配置的preTokens、nextTokens不会生效；sparse\_mode配置为0、4时，须保证attenmask与preTokens、nextTokens的范围一致。

    - softmaxMaxOut（aclTensor\*，计算输出）：Device侧的aclTensor。数据类型支持：FLOAT。数据格式支持ND。

    - softmaxSumOut（aclTensor\*，计算输出）：Device侧的aclTensor。数据类型支持：FLOAT。数据格式支持ND。

    - softmaxOutOut（aclTensor\*，计算输出）：预留参数，暂未使用。

    - attentionOutOut（aclTensor\*，计算输出）：Device侧的aclTensor。数据类型支持：FLOAT16、BFLOAT16。数据类型query的数据类型一致。数据格式支持ND。

    - workspaceSize（uint64\_t\*，出参）：返回需要在npu device侧申请的workspace大小。

    - executor（aclOpExecutor\*\*，出参）：返回op执行器，包含了算子计算流程。

- **返回值：**

    返回aclnnStatus状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

    **说明：** 
    第一段接口完成入参校验，若出现以下错误码，则对应原因为：
    -   返回161001（ACLNN_ERR_PARAM_NULLPTR）：如果传入参数是必选输入，输出或者必选属性，且是空指针，则返回161001。
    -   返回161002（ACLNN_ERR_PARAM_INVALID）：query、key、value、realShiftOptional、dropMaskOptional、paddingMaskOptional、attenMaskOptional、softmaxMaxOut、softmaxSumOut、softmaxOutOut、attentionOutOut的数据类型和数据格式不在支持的范围内。

## aclnnFlashAttentionVarLenScore

-   **参数说明：**
    -   workspace（void\*，入参）：在Device侧申请的workspace内存起址。
    -   workspaceSize（uint64\_t，入参）：在Device侧申请的workspace大小，由第一段接口aclnnFlashAttentionVarLenScoreGetWorkspaceSize获取。
    -   executor（aclOpExecutor\*，入参）：op执行器，包含了算子计算流程。
    -   stream（aclrtStream，入参）：指定执行任务的AscendCL stream流。

-   **返回值：**

    返回aclnnStatus状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

## 约束与限制

- 该接口与pytorch配合使用时，需要保证CANN相关包与PyTorch相关包的版本匹配

- 关于数据shape的约束，以inputLayout的TND为例，其中：

    -   B：取值范围为1\~2K。
    -   N：取值范围为1\~256。支持N不等长
    -   S：取值范围为1\~64K。支持S不等长
    -   D：取值范围为1\~512。

- prefix稀疏计算场景暂不支持。

## 调用示例

```C++
#include <iostream>
#include <vector>
#include "acl/acl.h"
#include "aclnnop/aclnn_flash_attention_score.h"

#define CHECK_RET(cond, return_expr) \
  do {                               \
    if (!(cond)) {                   \
      return_expr;                   \
    }                                \
  } while (0)

#define LOG_PRINT(message, ...)     \
  do {                              \
    printf(message, ##__VA_ARGS__); \
  } while (0)

int64_t GetShapeSize(const std::vector<int64_t>& shape) {
  int64_t shapeSize = 1;
  for (auto i : shape) {
    shapeSize *= i;
  }
  return shapeSize;
}

void PrintOutResult(std::vector<int64_t> &shape, void** deviceAddr) {
  auto size = GetShapeSize(shape);
  std::vector<float> resultData(size, 0);
  auto ret = aclrtMemcpy(resultData.data(), resultData.size() * sizeof(resultData[0]),
                         *deviceAddr, size * sizeof(resultData[0]), ACL_MEMCPY_DEVICE_TO_HOST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("copy result from device to host failed. ERROR: %d\n", ret); return);
  for (int64_t i = 0; i < size; i++) {
    LOG_PRINT("mean result[%ld] is: %f\n", i, resultData[i]);
  }
}

int Init(int32_t deviceId, aclrtContext* context, aclrtStream* stream) {
  // 固定写法，AscendCL初始化
  auto ret = aclInit(nullptr);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclInit failed. ERROR: %d\n", ret); return ret);
  ret = aclrtSetDevice(deviceId);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSetDevice failed. ERROR: %d\n", ret); return ret);
  ret = aclrtCreateContext(context, deviceId);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtCreateContext failed. ERROR: %d\n", ret); return ret);
  ret = aclrtSetCurrentContext(*context);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSetCurrentContext failed. ERROR: %d\n", ret); return ret);
  ret = aclrtCreateStream(stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtCreateStream failed. ERROR: %d\n", ret); return ret);
  return 0;
}

template <typename T>
int CreateAclTensor(const std::vector<T>& hostData, const std::vector<int64_t>& shape, void** deviceAddr,
                    aclDataType dataType, aclTensor** tensor) {
  auto size = GetShapeSize(shape) * sizeof(T);
  // 调用aclrtMalloc申请device侧内存
  auto ret = aclrtMalloc(deviceAddr, size, ACL_MEM_MALLOC_HUGE_FIRST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtMalloc failed. ERROR: %d\n", ret); return ret);
  // 调用aclrtMemcpy将host侧数据拷贝到device侧内存上
  ret = aclrtMemcpy(*deviceAddr, size, hostData.data(), size, ACL_MEMCPY_HOST_TO_DEVICE);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtMemcpy failed. ERROR: %d\n", ret); return ret);

  // 计算连续tensor的strides
  std::vector<int64_t> strides(shape.size(), 1);
  for (int64_t i = shape.size() - 2; i >= 0; i--) {
    strides[i] = shape[i + 1] * strides[i + 1];
  }

  // 调用aclCreateTensor接口创建aclTensor
  *tensor = aclCreateTensor(shape.data(), shape.size(), dataType, strides.data(), 0, aclFormat::ACL_FORMAT_ND,
                            shape.data(), shape.size(), *deviceAddr);
  return 0;
}

int main() {
  // 1. （固定写法）device/context/stream初始化，参考AscendCL对外接口列表
  // 根据自己的实际device填写deviceId
  int32_t deviceId = 0;
  aclrtContext context;
  aclrtStream stream;
  auto ret = Init(deviceId, &context, &stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("Init acl failed. ERROR: %d\n", ret); return ret);

  // 2. 构造输入与输出，需要根据API的接口自定义构造
  std::vector<int64_t> qShape = {256, 1, 128};
  std::vector<int64_t> kShape = {256, 1, 128};
  std::vector<int64_t> vShape = {256, 1, 128};
  std::vector<int64_t> attenmaskShape = {256, 256};

  std::vector<int64_t> attentionOutShape = {256, 1, 128};
  std::vector<int64_t> softmaxMaxShape = {256, 1, 8};
  std::vector<int64_t> softmaxSumShape = {256, 1, 8};

  void* qDeviceAddr = nullptr;
  void* kDeviceAddr = nullptr;
  void* vDeviceAddr = nullptr;
  void* attenmaskDeviceAddr = nullptr;
  void* attentionOutDeviceAddr = nullptr;
  void* softmaxMaxDeviceAddr = nullptr;
  void* softmaxSumDeviceAddr = nullptr;

  aclTensor* q = nullptr;
  aclTensor* k = nullptr;
  aclTensor* v = nullptr;
  aclTensor* pse = nullptr;
  aclTensor* dropMask = nullptr;
  aclTensor* padding = nullptr;
  aclTensor* attenmask = nullptr;
  aclTensor* attentionOut = nullptr;
  aclTensor* softmaxMax = nullptr;
  aclTensor* softmaxSum = nullptr;
  aclTensor* softmaxOut = nullptr;

  std::vector<short> qHostData(32768, 1);
  std::vector<short> kHostData(32768, 1);
  std::vector<short> vHostData(32768, 1);
  std::vector<uint8_t> attenmaskHostData(65536, 0);
  std::vector<short> attentionOutHostData(32768, 0);
  std::vector<float> softmaxMaxHostData(2048, 3.0);
  std::vector<float> softmaxSumHostData(2048, 3.0);


  ret = CreateAclTensor(qHostData, qShape, &qDeviceAddr, aclDataType::ACL_FLOAT16, &q);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  ret = CreateAclTensor(kHostData, kShape, &kDeviceAddr, aclDataType::ACL_FLOAT16, &k);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  ret = CreateAclTensor(vHostData, vShape, &vDeviceAddr, aclDataType::ACL_FLOAT16, &v);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  ret = CreateAclTensor(attenmaskHostData, attenmaskShape, &attenmaskDeviceAddr, aclDataType::ACL_UINT8, &attenmask);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  ret = CreateAclTensor(attentionOutHostData, attentionOutShape, &attentionOutDeviceAddr, aclDataType::ACL_FLOAT16, &attentionOut);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  ret = CreateAclTensor(softmaxMaxHostData, softmaxMaxShape, &softmaxMaxDeviceAddr, aclDataType::ACL_FLOAT, &softmaxMax);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  ret = CreateAclTensor(softmaxSumHostData, softmaxSumShape, &softmaxSumDeviceAddr, aclDataType::ACL_FLOAT, &softmaxSum);
  CHECK_RET(ret == ACL_SUCCESS, return ret);

  std::vector<int64_t> prefixOp = {0};
  aclIntArray *prefix = aclCreateIntArray(prefixOp.data(), 1);
  std::vector<int64_t>  acSeqQLenOp = {256};
  std::vector<int64_t>  acSeqKvLenOp = {256};
  aclIntArray* acSeqQLen = aclCreateIntArray(acSeqQLenOp.data(), acSeqQLenOp.size());
  aclIntArray* acSeqKvLen = aclCreateIntArray(acSeqKvLenOp.data(), acSeqKvLenOp.size());
  double scaleValue = 0.088388;
  double keepProb = 1;
  int64_t preTockens = 65536;
  int64_t nextTockens = 65536;
  int64_t headNum = 1;
  int64_t innerPrecise = 0;
  int64_t sparseMod = 0;

  char layOut[5] = {'T', 'N', 'D', 0};

  // 3. 调用CANN算子库API，需要修改为具体的Api名称
  uint64_t workspaceSize = 0;
  aclOpExecutor* executor;

  // 调用aclnnFlashAttentionVarLenScore第一段接口
  ret = aclnnFlashAttentionVarLenScoreGetWorkspaceSize(
            q, k, v, pse, dropMask, padding, attenmask, prefix, acSeqQLen, acSeqKvLen,
            scaleValue, keepProb, preTockens, nextTockens, headNum, layOut, innerPrecise,
            sparseMod, softmaxMax, softmaxSum, softmaxOut, attentionOut, &workspaceSize, &executor);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclnnFlashAttentionVarLenScoreGetWorkspaceSize failed. ERROR: %d\n", ret);
            return ret);

  // 根据第一段接口计算出的workspaceSize申请device内存
  void* workspaceAddr = nullptr;
  if (workspaceSize > 0) {
    ret = aclrtMalloc(&workspaceAddr, workspaceSize, ACL_MEM_MALLOC_HUGE_FIRST);
    CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("allocate workspace failed. ERROR: %d\n", ret); return ret);
  }

  // 调用aclnnFlashAttentionVarLenScore第二段接口
  ret = aclnnFlashAttentionVarLenScore(workspaceAddr, workspaceSize, executor, stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclnnFlashAttentionVarLenScore failed. ERROR: %d\n", ret); return ret);

  // 4. （固定写法）同步等待任务执行结束
  ret = aclrtSynchronizeStream(stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSynchronizeStream failed. ERROR: %d\n", ret); return ret);

  // 5. 获取输出的值，将device侧内存上的结果拷贝至host侧，需要根据具体API的接口定义修改
  PrintOutResult(attentionOutShape, &attentionOutDeviceAddr);
  PrintOutResult(softmaxMaxShape, &softmaxMaxDeviceAddr);
  PrintOutResult(softmaxSumShape, &softmaxSumDeviceAddr);

  // 6. 释放aclTensor和aclScalar，需要根据具体API的接口定义修改
  aclDestroyTensor(q);
  aclDestroyTensor(k);
  aclDestroyTensor(v);
  aclDestroyTensor(attenmask);
  aclDestroyTensor(attentionOut);
  aclDestroyTensor(softmaxMax);
  aclDestroyTensor(softmaxSum);

  // 7. 释放device资源
  aclrtFree(qDeviceAddr);
  aclrtFree(kDeviceAddr);
  aclrtFree(vDeviceAddr);
  aclrtFree(attenmaskDeviceAddr);
  aclrtFree(attentionOutDeviceAddr);
  aclrtFree(softmaxMaxDeviceAddr);
  aclrtFree(softmaxSumDeviceAddr);
  if (workspaceSize > 0) {
    aclrtFree(workspaceAddr);
  }
  aclrtDestroyStream(stream);
  aclrtDestroyContext(context);
  aclrtResetDevice(deviceId);
  aclFinalize();

  return 0;
}
```