## 支持的产品型号

- Atlas 训练产品。
- Atlas A2训练系列产品。
- Atlas 推理系列产品（Ascend 310P处理器）。

## 接口原型

每个算子分为[两段式接口](./common/两段式接口.md)，必须先调用“aclnnGemmGetWorkspaceSize”接口获取入参并根据计算流程计算所需workspace大小，再调用“aclnnGemm”接口执行计算。

- `aclnnStatus aclnnGemmGetWorkspaceSize(const aclTensor *A, const aclTensor *B, const aclTensor *C, float alpha, float beta, int64_t transA, int64_t transB, aclTensor *out, int8_t cubeMathType, uint64_t *workspaceSize, aclOpExecutor **executor)`
- `aclnnStatus aclnnGemm(void *workspace, uint64_t workspaceSize, aclOpExecutor *executor, aclrtStream stream)`

## 功能描述

- 算子功能：计算α 乘以A与B的乘积，再与β 和input的乘积求和。若transA非零，计算前会将A进行转置，同样的，若transB非零，则会将B进行转置。
  计算公式：

$$
out = α  (A @ B) + β  C
$$

- 算子功能：若transA与transB都为非零，则计算公式为：

$$
out = α  (A^T @ B^T) + βC
$$

## aclnnGemmGetWorkspaceSize

- **参数说明**：

  - self（aclTensor\*, 计算输入）：公式中的`self`，Device侧的aclTensor，数据类型支持FLOAT，FLOAT16，BFLOAT16 ，且数据类型与other的数据类型需满足数据类型推导规则（参见[互推导关系](./common/互推导关系.md)），shape需要与other满足[broadcast关系](./common/broadcast关系.md)。支持[非连续的Tensor](./common/非连续的Tensor.md)，[数据格式](./common/数据格式.md)支持ND。**其中BFLOAT16仅Atlas A2训练系列产品支持**。
  - A(aclTensor*, 计算输入)：公式中的输入`A`，Device侧的aclTensor，数据类型支持FLOAT16, FLOAT，BFLOAT16，且数据类型需要与C，B构成互相推导关系，维度支持两维，shape（或者转置后shape）需要满足与B相乘条件。支持[非连续的Tensor](./common/非连续的Tensor.md)，[数据格式](./common/数据格式.md)支持ND。**其中BFLOAT16仅Atlas A2训练系列产品支持**。
  - B(aclTensor*, 计算输入)：公式中的输入`B`，Device侧的aclTensor，数据类型支持FLOAT16, FLOAT，BFLOAT16，且数据类型需要与C，A构成互相推导关系，维度支持两维，shape（或者转置后shape）需要满足与A相乘条件。支持[非连续的Tensor](./common/非连续的Tensor.md)，[数据格式](./common/数据格式.md)支持ND。**其中BFLOAT16仅Atlas A2训练系列产品支持**。
  - C(aclTensor*, 计算输入)：公式中的输入`C`，Device侧的aclTensor，数据类型支持FLOAT16, FLOAT，BFLOAT16，且数据类型需要与AB计算后的结果构成互相推导关系，维度支持两维，shape需要与AB计算后的结果 一致。支持[非连续的Tensor](./common/非连续的Tensor.md)，[数据格式](./common/数据格式.md)支持ND。**其中BFLOAT16仅Atlas A2训练系列产品支持**。
  - alpha(float*, 计算输入)：公式中的输入`α`，Host侧的浮点型，表示A和B乘积的系数。
  - beta(float*, 计算输入)：公式中的输入`β`，Host侧的浮点型，表示C的系数。
  - transA(int64_t*, 计算输入)：公式中的输入`transA`，Host侧的整型，表示矩阵A是否需要转置，非零表示转置，零表示不需要转置。
  - transB(int64_t*, 计算输入)：公式中的输入`transB`，Host侧的整型，表示矩阵B是否需要转置，非零表示转置，零表示不需要转置。
  - out(aclTensor*, 计算输出)：公式中的`out`，Device侧的aclTensor，数据类型支持FLOAT16，FLOAT，BFLOAT16，且数据类型需要与C构成互相推导关系，维度支持两维，shape需要与C以及AB计算后的结果一致。支持[非连续的Tensor](./common/非连续的Tensor.md)，[数据格式](./common/数据格式.md)支持ND。**其中BFLOAT16仅Atlas A2训练系列产品支持**。
  - cubeMathType（int8_t, 计算输入）：Host侧的整型，判断Cube单元应使用哪种计算逻辑进行运算。数据类型支持INT8，支持的枚举值如下：
    - 0：KEPP_DTYPE，保持输入数据类型进行计算，若输入是FLOAT，Atlas 训练产品和Atlas 推理系列产品（Ascend 310P处理器）暂不支持，该值为0时会报错。
    - 1：ALLOW_FP32_DOWN_PRECISION，允许转换输入数据类型降低精度计算。当输入是FLOAT，Atlas 训练产品和Atlas 推理系列产品（Ascend 310P处理器）允许转换为FLOAT16计算，Atlas A2训练系列产品允许转换为HFLOAT32计算。
    - 2：USE_FP16，允许转换输入数据类型至FLOAT16计算。当输入是FLOAT，允许转换为FLOAT16计算。
    - 3：USE_HF32，允许转换输入数据类型至HFLOAT32计算。当输入是FLOAT，Atlas 训练产品和Atlas 推理系列产品（Ascend 310P处理器）暂不支持，取3时会报错。
  - workspaceSize(uint64_t*, 出参)：返回用户需要在npu device侧申请的workspace大小。
  - executor(aclOpExecutor**, 出参)：返回op执行器，包含了算子计算流程。
- **返回值**：

  aclnnStatus：返回状态码，具体参见[aclnn返回码](./common/aclnn返回码.md)。

```
161001 (ACLNN_ERR_PARAM_NULLPTR)：1. 传入的A, B，C或out是空指针。
  161002 (ACLNN_ERR_PARAM_INVALID)：1. 数据类型和数据格式不在支持的范围之内。
                                    2. A或B不是2维，或者进行计算时，shape不满足[m, k]和[k, n]的k维度相等关系。
                                    3. C和AB计算后的结果不满足broadcast关系。
```

## aclnnGemm

- **参数说明**：

  - workspace（void\*, 入参）：在Device侧申请的workspace内存起址。
  - workspaceSize（uint64_t, 入参）：在Device侧申请的workspace大小，由第一段接口aclnnXxxGetWorkspaceSize获取。
  - executor（aclOpExecutor\*, 入参）：op执行器，包含了算子计算流程。
  - stream（aclrtStream, 入参）：指定执行任务的 AscendCL Stream流。
- **返回值**：

  aclnnStatus：返回状态码，具体参见[aclnn返回码](./common/aclnn返回码.md)。

## 约束与限制

- 对于Atlas 训练产品和Atlas 推理系列产品（Ascend 310P处理器），Cube单元不支持FLOAT32计算。当输入为FLOAT32，可通过设置cubeMathType=1（ALLOW_FP32_DOWN_PRECISION）来允许接口内部cast到FLOAT16进行计算.

## 调用示例

示例代码如下，仅供参考，具体编译和执行过程请参考[编译与运行样例](./common/编译与运行样例.md)。

```
#include <iostream>
#include <vector>
#include "acl/acl.h"
#include "aclnnop/level2/aclnn_gemm.h"

#define CHECK_RET(cond, return_expr) \
  do {                               \
    if (!(cond)) {                   \
      return_expr;                   \
    }                                \
  } while (0)

#define LOG_PRINT(message, ...)     \
  do {                              \
    printf(message, ##__VA_ARGS__); \
  } while (0)

int64_t GetShapeSize(const std::vector<int64_t>& shape) {
  int64_t shape_size = 1;
  for (auto i : shape) {
    shape_size *= i;
  }
  return shape_size;
}

int Init(int32_t deviceId, aclrtContext* context, aclrtStream* stream) {
  // 固定写法，acl初始化
  auto ret = aclInit(nullptr);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclInit failed. ERROR: %d\n", ret); return ret);
  ret = aclrtSetDevice(deviceId);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSetDevice failed. ERROR: %d\n", ret); return ret);
  ret = aclrtCreateContext(context, deviceId);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtCreateContext failed. ERROR: %d\n", ret); return ret);
  ret = aclrtSetCurrentContext(*context);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSetCurrentContext failed. ERROR: %d\n", ret); return ret);
  ret = aclrtCreateStream(stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtCreateStream failed. ERROR: %d\n", ret); return ret);
  return 0;
}

template <typename T>
int CreateAclTensor(const std::vector<T>& hostData, const std::vector<int64_t>& shape, void** deviceAddr,
                    aclDataType dataType, aclTensor** tensor) {
  auto size = GetShapeSize(shape) * sizeof(T);
  // 调用aclrtMalloc申请device侧内存
  auto ret = aclrtMalloc(deviceAddr, size, ACL_MEM_MALLOC_HUGE_FIRST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtMalloc failed. ERROR: %d\n", ret); return ret);

  // 调用aclrtMemcpy将host侧数据拷贝到device侧内存上
  ret = aclrtMemcpy(*deviceAddr, size, hostData.data(), size, ACL_MEMCPY_HOST_TO_DEVICE);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtMemcpy failed. ERROR: %d\n", ret); return ret);

  // 计算连续tensor的strides
  std::vector<int64_t> strides(shape.size(), 1);
  for (int64_t i = shape.size() - 2; i >= 0; i--) {
    strides[i] = shape[i + 1] * strides[i + 1];
  }

  // 调用aclCreateTensor接口创建aclTensor
  *tensor = aclCreateTensor(shape.data(), shape.size(), dataType, strides.data(), 0, aclFormat::ACL_FORMAT_ND,
                            shape.data(), shape.size(), *deviceAddr);
  return 0;
}

int main() {
  // 1. （固定写法）device/context/stream初始化, 参考acl对外接口列表
  // 根据自己的实际device填写deviceId
  int32_t deviceId = 0;
  aclrtContext context;
  aclrtStream stream;
  auto ret = Init(deviceId, &context, &stream);
  // check根据自己的需要处理
  CHECK_RET(ret == 0, LOG_PRINT("Init acl failed. ERROR: %d\n", ret); return ret);
  // 2. 构造输入与输出，需要根据API的接口自定义构造
  std::vector<int64_t> AShape = {2, 2};
  std::vector<int64_t> BShape = {2, 2};
  std::vector<int64_t> CShape = {2, 2};
  std::vector<int64_t> outShape = {2, 2};
  void* ADeviceAddr = nullptr;
  void* BDeviceAddr = nullptr;
  void* CDeviceAddr = nullptr;
  void* outDeviceAddr = nullptr;
  aclTensor* A = nullptr;
  aclTensor* B = nullptr;
  aclTensor* C = nullptr;
  aclTensor* out = nullptr;
  std::vector<float> AHostData = {1, 2, 1, 2};
  std::vector<float> BHostData = {1, 2, 1, 2};
  std::vector<float> CHostData = {1, 1, 1, 1};
  std::vector<float> outHostData = {0, 0, 0, 0};
  float alpha = 1.0f;
  float beta = 2.0f;
  int64_t transA = 0;
  int64_t transB = 0;
  int8_t cubeMathType = 1;

  // 创建A aclTensor
  ret = CreateAclTensor(AHostData, AShape, &ADeviceAddr, aclDataType::ACL_FLOAT, &A);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  // 创建B aclTensor
  ret = CreateAclTensor(BHostData, BShape, &BDeviceAddr, aclDataType::ACL_FLOAT, &B);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  // 创建C aclTensor
  ret = CreateAclTensor(CHostData, CShape, &CDeviceAddr, aclDataType::ACL_FLOAT, &C);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  // 创建out aclTensor
  ret = CreateAclTensor(outHostData, outShape, &outDeviceAddr, aclDataType::ACL_FLOAT, &out);
  CHECK_RET(ret == ACL_SUCCESS, return ret);

  // 3. 调用CANN算子库API
  uint64_t workspaceSize = 0;
  aclOpExecutor* executor;
  // 调用aclnnAdd第一段接口
  ret = aclnnGemmGetWorkspaceSize(A, B, C, alpha, beta, transA, transB, out, cubeMathType, &workspaceSize, &executor);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclnnGemmGetWorkspaceSize failed. ERROR: %d\n", ret); return ret);
  // 根据第一段接口计算出的workspaceSize申请device内存
  void* workspaceAddr = nullptr;
  if (workspaceSize > 0) {
    ret = aclrtMalloc(&workspaceAddr, workspaceSize, ACL_MEM_MALLOC_HUGE_FIRST);
    CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("allocate workspace failed. ERROR: %d\n", ret); return ret;);
  }
  // 调用aclnnAdd第二段接口
  ret = aclnnGemm(workspaceAddr, workspaceSize, executor, stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclnnGemm failed. ERROR: %d\n", ret); return ret);
  // 4. （固定写法）同步等待任务执行结束
  ret = aclrtSynchronizeStream(stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSynchronizeStream failed. ERROR: %d\n", ret); return ret);
  // 5. 获取输出的值，将device侧内存上的结果拷贝至host侧，需要根据具体API的接口定义修改
  auto size = GetShapeSize(outShape);
  std::vector<float> resultData(size, 0);
  ret = aclrtMemcpy(resultData.data(), resultData.size() * sizeof(resultData[0]), outDeviceAddr, size * sizeof(float),
                    ACL_MEMCPY_DEVICE_TO_HOST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("copy result from device to host failed. ERROR: %d\n", ret); return ret);
  for (int64_t i = 0; i < size; i++) {
    LOG_PRINT("result[%ld] is: %f\n", i, resultData[i]);
  }

  // 6. 释放aclTensor和aclScalar，需要根据具体API的接口定义修改
  aclDestroyTensor(A);
  aclDestroyTensor(B);
  aclDestroyTensor(C);
  aclDestroyTensor(out);

  // 7. 释放device资源，需要根据具体API的接口定义修改
  aclrtFree(ADeviceAddr);
  aclrtFree(BDeviceAddr);
  aclrtFree(CDeviceAddr);
  aclrtFree(outDeviceAddr);
  if (workspaceSize > 0) {
    aclrtFree(workspaceAddr);
  }
  aclrtDestroyStream(stream);
  aclrtDestroyContext(context);
  aclrtResetDevice(deviceId);
  aclFinalize();
  return 0;
}
```
