# aclnnGridSampler3DBackwardBackward

## 支持的产品型号

- Atlas 训练产品。
- Atlas A2训练系列产品。

## 接口原型

- 每个算子分为[两段式接口](common/两段式接口.md)，必须先调用“aclnnGridSampler3DBackwardGetWorkspaceSize”接口获取计算所需workspace大小以及包含了算子计算流程的执行器，再调用“aclnnGridSampler3DBackward”接口执行计算。
  * `aclnnStatus aclnnGridSampler3DBackwardGetWorkspaceSize(const aclTensor *gradOutput, const aclTensor *input, const aclTensor *grid, int64_t interpolationMode, int64_t paddingMode, bool alignCorners, const aclBoolArray* outputMask, aclTensor *inputGrad, aclTensor *gridGrad, uint64_t *workspaceSize, aclOpExecutor **executor)`
  * `aclnnStatus aclnnGridSampler3DBackward(void *workspace, uint64_t workspaceSize, aclOpExecutor *executor, aclrtStream stream)`

## 功能描述

- aclnnGridSampler3D算子的反向计算。


## aclnnGridSampler3DBackwardGetWorkspaceSize

- **参数说明**：

  * gradOutput(const aclTensor*, 计算输入)：Device侧的aclTensor，数据类型支持FLOAT16、FLOAT、DOUBLE，且数据类型与input的数据类型一致。支持[非连续的Tensor](common/非连续的Tensor.md)，[数据格式](common/数据格式.md)支持NCDHW、NDHWC。
  * input(const aclTensor*, 计算输入)：Device侧的aclTensor，数据类型支持FLOAT16、FLOAT、DOUBLE。支持[非连续的Tensor](common/非连续的Tensor.md)，[数据格式](common/数据格式.md)支持NCDHW、NDHWC。
  * grid(const aclTensor*, 计算输入)：Device侧的aclTensor，指定输入归一化的采用像素位置，数据类型支持FLOAT16、FLOAT、DOUBLE，且数据类型与input数据类型一致。支持[非连续的Tensor](common/非连续的Tensor.md)，[数据格式](common/数据格式.md)支持NCDHW、NDHWC。
  * interpolationMode(int64_t, 计算输入)：int64_t类型，表示计算输出的插值模式，支持0('bilinear') | 1('nearest')。
  * paddingMode(int64_t, 计算输入)：int64_t类型，表示grid范围外填充模式。即当grid有超过[-1, 1]范围的值，则按照paddingMode定义的方式处理相应的输出，支持0('zeros') | 1('border') | 2('reflection')。
  * alignCorners(bool, 计算输入)：bool类型，表示对齐角像素点的方式。如果为True，则将极值-1和1视为参考输入的角像素点的中心点。如果为False，则视为参考输入的角像素点的角点。
  * outputMask(const aclBoolArray*, 计算输入)：aclBoolArray类型，表示输出的掩码。
  * inputGrad(aclTensor*, 计算输出)：Device侧的aclTensor，数据类型支持FLOAT16、FLOAT、DOUBLE。支持[非连续的Tensor](common/非连续的Tensor.md)，[数据格式](common/数据格式.md)支持NCDHW、NDHWC。
  * gridGrad(aclTensor*, 计算输出)：Device侧的aclTensor，数据类型支持FLOAT16、FLOAT、DOUBLE。支持[非连续的Tensor](common/非连续的Tensor.md)，[数据格式](common/数据格式.md)支持NCDHW、NDHWC。
  * workspaceSize(uint64_t*, 出参)：返回用户需要在Device侧申请的workspace大小。
  * executor(aclOpExecutor**, 出参)：返回op执行器，包含了算子计算流程。

- **返回值**：

  aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

```
第一段接口完成入参校验，出现以下场景时报错：
返回161001 (ACLNN_ERR_PARAM_NULLPTR): 1. 传入的gradOutput、input、grid、inputGrad或gridGrad是空指针。
返回161002 (ACLNN_ERR_PARAM_INVALID): 1. gradOutput、input、grid、inputGrad或gridGrad的数据类型不在支持的范围之内或数据类型不一致。
                                     2. interpolationMode和paddingMode的值不在支持范围内。
                                     3. gradOutput、input、grid、inputGrad、gridGrad的维度关系不匹配。
```

## aclnnGridSampler3DBackward

- **参数说明**：

  * workspace(void*, 入参)：在Device侧申请的workspace内存起址。
  * workspaceSize(uint64_t, 入参)：在Device侧申请的workspace大小，由第一段接口aclnnGridSampler3DBackwardGetWorkspaceSize获取。
  * executor(aclOpExecutor*, 入参)：op执行器，包含了算子计算流程。
  * stream(aclrtStream, 入参)：指定执行任务的 AscendCL Stream流。

- **返回值**：

  **aclnnStatus**：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

## 约束与限制

无

## 调用示例

示例代码如下，仅供参考，具体编译和执行过程请参考[编译与运行样例](common/编译与运行样例.md)。

```
#include <iostream>
#include <vector>
#include "acl/acl.h"
#include "aclnnop/aclnn_grid_sampler3d_backward.h"

#define CHECK_RET(cond, return_expr) \
  do {                               \
    if (!(cond)) {                   \
      return_expr;                   \
    }                                \
  } while (0)

#define LOG_PRINT(message, ...)     \
  do {                              \
    printf(message, ##__VA_ARGS__); \
  } while (0)

int64_t GetShapeSize(const std::vector<int64_t>& shape) {
  int64_t shapeSize = 1;
  for (auto i : shape) {
    shapeSize *= i;
  }
  return shapeSize;
}

int Init(int32_t deviceId, aclrtContext* context, aclrtStream* stream) {
  // 固定写法，AscendCL初始化
  auto ret = aclInit(nullptr);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclInit failed. ERROR: %d\n", ret); return ret);
  ret = aclrtSetDevice(deviceId);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSetDevice failed. ERROR: %d\n", ret); return ret);
  ret = aclrtCreateContext(context, deviceId);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtCreateContext failed. ERROR: %d\n", ret); return ret);
  ret = aclrtSetCurrentContext(*context);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSetCurrentContext failed. ERROR: %d\n", ret); return ret);
  ret = aclrtCreateStream(stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtCreateStream failed. ERROR: %d\n", ret); return ret);
  return 0;
}

template <typename T>
int CreateAclTensor(const std::vector<T>& hostData, const std::vector<int64_t>& shape, void** deviceAddr,
                    aclDataType dataType, aclTensor** tensor) {
  auto size = GetShapeSize(shape) * sizeof(T);
  // 调用aclrtMalloc申请device侧内存
  auto ret = aclrtMalloc(deviceAddr, size, ACL_MEM_MALLOC_HUGE_FIRST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtMalloc failed. ERROR: %d\n", ret); return ret);
  // 调用aclrtMemcpy将host侧数据拷贝到device侧内存上
  ret = aclrtMemcpy(*deviceAddr, size, hostData.data(), size, ACL_MEMCPY_HOST_TO_DEVICE);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtMemcpy failed. ERROR: %d\n", ret); return ret);

  // 计算连续tensor的strides
  std::vector<int64_t> strides(shape.size(), 1);
  for (int64_t i = shape.size() - 2; i >= 0; i--) {
    strides[i] = shape[i + 1] * strides[i + 1];
  }

  // 调用aclCreateTensor接口创建aclTensor
  *tensor = aclCreateTensor(shape.data(), shape.size(), dataType, strides.data(), 0, aclFormat::ACL_FORMAT_ND,
                            shape.data(), shape.size(), *deviceAddr);
  return 0;
}

int main() {
  // 1. （固定写法）device/context/stream初始化，参考AscendCL对外接口列表
  // 根据自己的实际device填写deviceId
  int32_t deviceId = 0;
  aclrtContext context;
  aclrtStream stream;
  auto ret = Init(deviceId, &context, &stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("Init acl failed. ERROR: %d\n", ret); return ret);

  // 2. 构造输入与输出，需要根据API的接口自定义构造
  int64_t interpolationMode = 0;
  int64_t paddingMode = 0;
  bool alignCorners = false;
  aclBoolArray* outputMask = nullptr;
  std::vector<int64_t> gradOutputShape = {1, 1, 1, 2, 2};
  std::vector<int64_t> inputShape = {1, 1, 1, 3, 3};
  std::vector<int64_t> gridShape = {1, 1, 2, 2, 3};
  std::vector<int64_t> inputGradShape = {1, 1, 1, 3, 3};
  std::vector<int64_t> gridGradShape = {1, 1, 2, 2, 3};
  void* gradOutputDeviceAddr = nullptr;
  void* inputDeviceAddr = nullptr;
  void* gridDeviceAddr = nullptr;
  void* inputGradDeviceAddr = nullptr;
  void* gridGradDeviceAddr = nullptr;
  aclTensor* gradOutput = nullptr;
  aclTensor* input = nullptr;
  aclTensor* grid = nullptr;
  aclTensor* inputGrad = nullptr;
  aclTensor* gridGrad = nullptr;

  std::vector<float> gradOutputHostData = {1, 1, 1, 1};
  std::vector<float> inputHostData = {1, 2, 3, 4, 5, 6, 7, 8, 9,};
  std::vector<float> gridHostData = {-1, -1, 0, -1, 1, -1, -1, 0, 0, 0, 1, 0};
  std::vector<float> inputGradHostData = {0, 0, 0, 0, 0, 0, 0, 0, 0};
  std::vector<float> gridGradHostData = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

  bool maskValue[2] = {true, true};
  outputMask = aclCreateBoolArray(&(maskValue[0]), 2);

  // 创建gradOutput aclTensor
  ret = CreateAclTensor(gradOutputHostData, gradOutputShape, &gradOutputDeviceAddr, aclDataType::ACL_FLOAT, &gradOutput);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  // 创建input aclTensor
  ret = CreateAclTensor(inputHostData, inputShape, &inputDeviceAddr, aclDataType::ACL_FLOAT, &input);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  // 创建grid aclTensor
  ret = CreateAclTensor(gridHostData, gridShape, &gridDeviceAddr, aclDataType::ACL_FLOAT, &grid);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  // 创建inputGrad aclTensor
  ret = CreateAclTensor(inputGradHostData, inputGradShape, &inputGradDeviceAddr, aclDataType::ACL_FLOAT, &inputGrad);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  // 创建gridGrad aclTensor
  ret = CreateAclTensor(gridGradHostData, gridGradShape, &gridGradDeviceAddr, aclDataType::ACL_FLOAT, &gridGrad);
  CHECK_RET(ret == ACL_SUCCESS, return ret);

  // 3. 调用CANN算子库API，需要修改为具体的Api名称
  uint64_t workspaceSize = 0;
  aclOpExecutor* executor;
  // 调用aclnnGridSampler3DBackward第一段接口
  ret = aclnnGridSampler3DBackwardGetWorkspaceSize(gradOutput, input, grid, interpolationMode, paddingMode,
                                                   alignCorners, outputMask, inputGrad, gridGrad,
                                                   &workspaceSize, &executor);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclnnGridSampler3DBackwardGetWorkspaceSize failed. ERROR: %d\n", ret);
            return ret);
  // 根据第一段接口计算出的workspaceSize申请device内存
  void* workspaceAddr = nullptr;
  if (workspaceSize > 0) {
    ret = aclrtMalloc(&workspaceAddr, workspaceSize, ACL_MEM_MALLOC_HUGE_FIRST);
    CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("allocate workspace failed. ERROR: %d\n", ret); return ret);
  }
  // 调用aclnnGridSampler3DBackward第二段接口
  ret = aclnnGridSampler3DBackward(workspaceAddr, workspaceSize, executor, stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclnnGridSampler3DBackward failed. ERROR: %d\n", ret); return ret);

  // 4. （固定写法）同步等待任务执行结束
  ret = aclrtSynchronizeStream(stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSynchronizeStream failed. ERROR: %d\n", ret); return ret);

  // 5. 获取输出的值，将device侧内存上的结果拷贝至host侧，需要根据具体API的接口定义修改
  auto inputGradSize = GetShapeSize(inputGradShape);
  std::vector<float> inputGradResultData(inputGradSize, 0);
  ret = aclrtMemcpy(inputGradResultData.data(), inputGradResultData.size() * sizeof(inputGradResultData[0]),
                    inputGradDeviceAddr, inputGradSize * sizeof(inputGradResultData[0]), ACL_MEMCPY_DEVICE_TO_HOST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("copy inputGradResultData from device to host failed. ERROR: %d\n", ret);
            return ret);
  for (int64_t i = 0; i < inputGradSize; i++) {
    LOG_PRINT("inputGradResultData[%ld] is: %f\n", i, inputGradResultData[i]);
  }

  auto gridGradSize = GetShapeSize(gridGradShape);
  std::vector<float> gridGradResultData(gridGradSize, 0);
  ret = aclrtMemcpy(gridGradResultData.data(), gridGradResultData.size() * sizeof(gridGradResultData[0]),
                    gridGradDeviceAddr, gridGradSize * sizeof(gridGradResultData[0]), ACL_MEMCPY_DEVICE_TO_HOST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("copy gridGradResultData from device to host failed. ERROR: %d\n", ret);
            return ret);
  for (int64_t i = 0; i < gridGradSize; i++) {
    LOG_PRINT("gridGradResultData[%ld] is: %f\n", i, gridGradResultData[i]);
  }

  // 6. 释放aclTensor和aclBoolArray，需要根据具体API的接口定义修改
  aclDestroyTensor(gradOutput);
  aclDestroyTensor(input);
  aclDestroyTensor(grid);
  aclDestroyTensor(inputGrad);
  aclDestroyTensor(gridGrad);
  aclDestroyBoolArray(outputMask);
  
  // 7. 释放device资源，需要根据具体API的接口定义修改
  aclrtFree(gradOutputDeviceAddr);
  aclrtFree(inputDeviceAddr);
  aclrtFree(gridDeviceAddr);
  aclrtFree(inputGradDeviceAddr);
  aclrtFree(gridGradDeviceAddr);
  if (workspaceSize > 0) {
    aclrtFree(workspaceAddr);
  }
  aclrtDestroyStream(stream);
  aclrtDestroyContext(context);
  aclrtResetDevice(deviceId);
  aclFinalize();
  return 0;
}
```

