# aclnnMatmulAllReduceV2

## 支持的产品型号

-   Atlas 推理系列产品（Ascend 310P3处理器）。
-   Atlas A2训练系列产品。

## 接口原型

每个算子分为两段式接口，必须先调用“aclnnMatmulAllReduceV2GetWorkspaceSize”接口获取计算所需workspace大小以及包含了算子计算流程的执行器，再调用“aclnnMatmulAllReduceV2”接口执行计算。

* `aclnnStatus aclnnMatmulAllReduceV2GetWorkspaceSize(const aclTensor *x1, const aclTensor *x2, const aclTensor *bias, const aclTensor *x3, const char* group, const char *reduceOp, int64_t commTurn, int64_t streamMode, const aclTensor *output, uint64_t *workspaceSize, aclOpExecutor **executor);`
* `aclnnStatus aclnnMatmulAllReduceV2(void *workspace, uint64_t workspaceSize, aclOpExecutor *executor, const aclrtStream stream);`

## 功能描述

-   **算子功能**：完成mm + all\_reduce\_base计算。
-   **计算公式**：

    $$
    output=x1*x2
    $$

## aclnnMatmulAllReduceV2GetWorkspaceSize

-   **参数说明：**
    -   **x1(const aclTensor *, 计算输入)**：npu device侧的2维aclTensor，mm左矩阵。数据类型支持：FLOAT16、BFLOAT16。数据格式支持：ND。不支持非连续输入。
    -   **x2(const aclTensor *, 计算输入)**：npu device侧的2维aclTensor，mm右矩阵。数据类型支持：FLOAT16、BFLOAT16。支持[最后两轴转置情况下的非连续tensor传入]，数据格式支持：ND。x1的第二维和x2的第一维维度相等，或x2最后两轴转置后，对应维度相等。非连续的Tensor仅支持transpose场景。
    -   **bias(const aclTensor *, 计算输入)**：bias。数据类型支持：FLOAT16、BFLOAT16。数据格式支持：ND，可选，可为空。非空时shape最后一维和x2最后一维相等
    -   **x3(const aclTensor *, 计算输入)**：x3，可选，matmul计算后的偏移。数据类型支持：FLOAT16、BFLOAT16。数据格式支持：ND，必选，shape与matmul计算后的shape相同。
    -   **group(const char *, 计算输入)**：通信域名称。数据类型支持：String。通过Hccl提供的接口获取：extern HcclResult HcclGetCommName(HcclComm comm, char* commName); commName即为group。
    -   **reduceOp(const char *, 计算输入)**：reduce操作类型。数据类型支持：String。目前仅支持"sum"。
    -   **commTurn(int64_t, 计算输入)**：通信数据切分数，即总数据量/单次通信量。数据类型支持：int64_t。**当前版本仅支持输入0。**
    -   **streamMode(int64_t，计算输入)**：Host侧的整型，acl流模式的枚举，当前只支持枚举值1，类型支持：int64\_t。
    -   **output(aclTensor *, 输出)**：计算+通信的结果：。数据类型支持：同x1、x2输入。shape第一维与x1第一维相等，最后一维和x2最后一维相等
    -   **workspaceSize(uint64_t *, 出参)**：返回需要在npu device侧申请的workspace大小。
    -   **executor(aclOpExecutor \*\*, 出参)**：返回op执行器，包含了算子计算流程。

-   **返回值：**

    返回aclnnStatus状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

## aclnnMatmulAllReduceV2

-   **参数说明：**
    -   workspace（void\*，入参）：在Device侧申请的workspace内存起址。
    -   workspaceSize（uint64\_t,入参）：在Device侧申请的workspace大小，由第一段接口aclnnMatmulAllReduceV2GetWorkspaceSize。
    -   executor（aclOpExecutor\*，入参）：op执行器，包含了算子计算流程。
    -   stream（aclrtStream，入参）：指定执行任务的AscendCL stream流。

-   **返回值：**

    返回aclnnStatus状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

## 约束与限制

1.  输入x1可为2维或者3维，且不为空Tensor，其维度为(b, m, k)或者(m, k)。x2必须是2维，且不为空Tensor。其维度为(k, n)，轴满足mm算子入参要求，k轴相等。bias若非空，bias为1维。
2.  x3以及输出output除最后一维皆与输入x1除最后一维相等，x3以及输出output的最后1维与输入x2的最后1维相等。bias若非空，shape大小与output最后一维相等。
3.  传入的x1、x2、x3或者output不为空指针。
4.  x1和x2，x3、output、bias（非空场景）的数据类型和数据格式需要在支持的范围之内。
5.  只支持x2矩阵转置/不转置，x1矩阵支持不转置场景。
6.  Atlas 推理系列产品（Ascend 310P3处理器）只支持2卡。
7.  Atlas A2训练系列产品支持2、4、8卡。

## 调用示例

```
#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include <chrono>
#include <vector>
#include <string>
#include <cmath>
#include <cstdint>
#include <memory>
#include <thread>
#include "hccl/hccl.h"
#include <hccl/hccl_types.h>
#include <mutex>
#include <unistd.h>
#include <vector>
#include <sys/prctl.h>
#include <map>

#include <iostream>
#include <fstream>
#include <cstdio>
#include <string>
#include <vector>
#include <iomanip>
#include <cassert>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <algorithm>
#include <iostream>
#include "acl/acl.h"
#include "acl/acl_base.h"
#include "aclnnop/level2/aclnn_matmul.h"
#include "aclnnop/aclnn_matmul_all_reduce.h"
#include "aclnnop/aclnn_matmul_all_reduce_v2.h"
#include <climits>

#define epsilon 1e-4

using namespace std;
// using namespace hccl;
std::mutex mt1;
//=====================================================Start Base function======================================
#define ALLOW_FP32_DOWN_PRECISION 1
#define N 128
float diffThd = 0.001;
float pctThd = 0.001;

thread_local int g_dev_id;
void set_device_id(int id)
{
    g_dev_id = id;
}
int get_device_id(void)
{
    return g_dev_id;
}

typedef signed char s8;
typedef signed short s16;
typedef signed int s32;
typedef signed long long s64;
typedef unsigned char u8;
typedef unsigned short u16;
typedef unsigned int u32;
typedef unsigned long long u64;

#define ACL_CHECK(ret)                                                                                     \
    do {                                                                                                   \
        auto retcode = ret;                                                                                \
        if (retcode != ACL_SUCCESS) {                                                                      \
            printf("[ERROR] acl interface return err %s:%d, retcode: %d \n", __FILE__, __LINE__, retcode); \
            return retcode;                                                                                \
        }                                                                                                  \
    } while (0)

// using namespace std;
#define CHECK_RET(cond, return_expr) \
    do {                             \
        if (!(cond)) {               \
            return_expr;             \
        }                            \
    } while (0)

struct DataSize {
    u64 min_bytes;
    u64 max_bytes;
    u64 step_bytes = 0;
    double step_factor;
    u64 count;
    u64 data_size;
    u64 type_size;
    int op_flag;
    void Print()
    {
        printf("acl min_bytes : %lld, max_bytes: %lld, step_bytes: %lld, step_factor: %f, count: %lld, data_size: "
               "%lld,  type_size: %lld\n",
            min_bytes,
            max_bytes,
            step_bytes,
            step_factor,
            count,
            data_size,
            type_size);
    }
};

struct Resource {
    aclrtStream rtStream;
    aclrtStream rtStream_aicpu;
    aclrtStream rtStream_record;
    aclrtEvent startEvent;
    aclrtEvent endEvent;
    aclrtContext context;
};

struct Args {
    int ndev;
    int rankId;
    uint32_t logicDeviceId;
    uint32_t rootRank;
    HcclComm hcclComm;
    Resource resources;
    DataSize dataPara;
    uint32_t m;
    uint32_t k;
    uint32_t n;
    std::string dtype;
    std::string bin_path;
    std::string run_type;
    uint32_t loop_cnt;
    int bias_flag;
};

int dtype_to_hccltype(std::string dtype)
{
    int ret = 0;
    if (dtype == "float32") {
        ret = HCCL_DATA_TYPE_FP32;
    } else if (dtype == "float16") {
        ret = HCCL_DATA_TYPE_FP16;
    } else if (dtype == "bfloat16") {
        ret = HCCL_DATA_TYPE_BFP16;
    } else if (dtype == "int8") {
        ret = HCCL_DATA_TYPE_INT8;
    } else if (dtype == "int32") {
        ret = HCCL_DATA_TYPE_INT32;
    } else if (dtype == "int16") {
        ret = HCCL_DATA_TYPE_INT16;
    } else {
        printf("[ERROR] no match hccltype\n");
        return -1;
    }
    return ret;
}

std::map<std::string, HcclDataType> dtype_to_HcclDataType = {{"float32", HCCL_DATA_TYPE_FP32},
    {"float16", HCCL_DATA_TYPE_FP16},
    {"bfloat16", HCCL_DATA_TYPE_BFP16},
    {"int8", HCCL_DATA_TYPE_INT8},
    {"int16", HCCL_DATA_TYPE_INT16},
    {"int32", HCCL_DATA_TYPE_INT32}};

std::map<std::string, aclDataType> g_data_type_map = {{"float32", ACL_FLOAT},
    {"float16", ACL_FLOAT16},
    {"bfloat16", ACL_BF16},
    {"int8", ACL_INT8},
    {"int16", ACL_INT16},
    {"int32", ACL_INT32}};

size_t dtype_to_sizeof(std::string dtype)
{
    size_t ret = 0;
    if (dtype == "float32") {
        ret = sizeof(int32_t);
    } else if (dtype == "float16") {
        ret = sizeof(int16_t);
    } else if (dtype == "bfloat16") {
        ret = sizeof(int16_t);
    } else if (dtype == "int8") {
        ret = sizeof(int8_t);
    } else if (dtype == "int32") {
        ret = sizeof(int32_t);
    } else if (dtype == "int16") {
        ret = sizeof(int16_t);
    } else {
        printf("[ERROR] not invalid dtype\n");
        return -1;
    }
    return ret;
}

aclTensor *CreateTensor(const int64_t *dims, uint64_t dimNum, aclDataType dataType, aclFormat format, void *dataAddr)
{
    std::vector<int64_t> strides(dimNum, 1);
    for (int64_t i = dimNum - 2; i >= 0; i--) {
        strides[i] = dims[i + 1] * strides[i + 1];
    }
    auto ret = aclCreateTensor(dims, dimNum, dataType, strides.data(), 0, format, dims, dimNum, dataAddr);
    // std::cout << ret << std::endl;
    return ret;
}

int64_t GetShapeSize(const std::vector<int64_t> &shape)
{
    int64_t shape_size = 1;
    for (auto i : shape) {
        shape_size *= i;
    }
    return shape_size;
}

std::string GetShapeString(std::string tensor_name, const std::vector<int64_t> &shape, std::string dtype)
{
    std::string res = tensor_name;
    for (auto i : shape) {
        res += "_" + to_string(i);
    }
    res += "_" + dtype;
    return res;
}

int Init(int32_t deviceId, aclrtContext *context, aclrtStream *stream)
{
    auto ret = aclrtSetDevice(deviceId);
    CHECK_RET(ret == ACL_SUCCESS, std::cout << "aclrtSetDevice failed. ERROR" << ret << std::endl; return -1);
    ret = aclrtCreateContext(context, deviceId);
    CHECK_RET(ret == ACL_SUCCESS, std::cout << "aclrtCreateContext failed. ERROR" << ret << std::endl; return -1);
    ret = aclrtSetCurrentContext(*context);
    CHECK_RET(ret == ACL_SUCCESS, std::cout << "aclrtSetCurrentContext failed. ERROR" << ret << std::endl; return -1);
    ret = aclrtCreateStream(stream);
    CHECK_RET(ret == ACL_SUCCESS, std::cout << "aclrtCreateStream failed. ERROR" << ret << std::endl; return -1);

    ret = aclInit(nullptr);
    CHECK_RET(ret == ACL_SUCCESS, std::cout << "aclInit failed. ERROR" << ret << std::endl; return -1);
    return 0;
}

//=====================================================End Base function======================================

//=====================================================Start Result Compare Function===============================

// 浮点数计算精度，当前算误差百分比
#define HCCL_EPSION_FLOAT 0.000001
#define HCCL_EPSION_FP16 0.0001

bool ReadFile(const std::string &filePath, size_t &fileSize, void *buffer, size_t bufferSize)
{
    struct stat sBuf;
    int fileStatus = stat(filePath.data(), &sBuf);
    if (fileStatus == -1) {
        printf("[ERROR]failed to get file = %s ", filePath.c_str());
        return false;
    }
    if (S_ISREG(sBuf.st_mode) == 0) {
        printf("[ERROR]%s is not a file, please enter a file", filePath.c_str());
        return false;
    }

    std::ifstream file;
    file.open(filePath, std::ios::binary);
    if (!file.is_open()) {
        printf("[ERROR]Open file failed. path = %s", filePath.c_str());
        return false;
    }

    std::filebuf *buf = file.rdbuf();
    size_t size = buf->pubseekoff(0, std::ios::end, std::ios::in);
    if (size == 0) {
        printf("[ERROR]file size is 0");
        file.close();
        return false;
    }
    if (size > bufferSize) {
        printf("[ERROR]file size is larger than buffer size");
        file.close();
        return false;
    }
    buf->pubseekpos(0, std::ios::in);
    buf->sgetn(static_cast<char *>(buffer), size);
    fileSize = size;
    file.close();
    return true;
}

// 保存bin文件
void writeBin(std::string path, void *buf, size_t size)
{
    std::ofstream outfile(path, std::ifstream::binary);
    outfile.write((char *)buf, size);
    outfile.close();
}

int check_buf_result_float(const void *result_buf, const void *check_buf, u64 count)
{
    u64 i = 0;  // j = 0;
    // int n = 0;
    int err = 0;
    float *c_buf = (float *)check_buf;
    float *result = (float *)result_buf;
    u64 first_err_pos = ULLONG_MAX;
    for (i = 0; i < count; ++i) {
        if (fabsf(c_buf[i] - result[i]) > 0.000001) {
            if (fabsf(c_buf[i] - result[i]) / (max(fabsf(c_buf[i]), fabsf(result[i])) + 10e-10) > diffThd) {
                err++;
                if (err < 20) {
                    printf("check index:%llu error, exp:%f, act:%f \n", i, c_buf, result);
                }
            }
        }
    }

    if (err / count > pctThd) {
        printf("[ERROR] error rate is %f\n", err / count);
        return 1;
    }
    return 0;
}

int check_buf_result_half(const void *result_buf, const void *check_buf, u64 count)
{
    u64 i = 0;  // j = 0;
    // int n = 0;
    int err = 0;
    float c_buf;
    float result;
    for (i = 0; i < count; ++i) {
        c_buf = aclFloat16ToFloat(reinterpret_cast<const aclFloat16 *>(check_buf)[i]);
        result = aclFloat16ToFloat(reinterpret_cast<const aclFloat16 *>(result_buf)[i]);
        if (fabsf(c_buf - result) > 0.0001) {
            if (fabsf(c_buf - result) / (max(fabsf(c_buf), fabsf(result)) + 10e-10) > diffThd) {
                err++;
                if (err < 20) {
                    printf("check index:%llu error, exp:%f, act:%f \n", i, c_buf, result);
                }
            }
        }
    }
    if (err / count > pctThd) {
        printf("[ERROR] error rate is %f\n", err / count);
        return 1;
    }
    return 0;
}

int check_buf_result_int8(const void *result_buf, const void *check_buf, u64 count)
{
    u64 i = 0;
    s8 *c_buf = (s8 *)check_buf;
    s8 *result = (s8 *)result_buf;
    int err = 0;
    u64 first_err_pos = ULLONG_MAX;
    for (i = 0; i < count; ++i) {
        if (c_buf[i] != result[i]) {
            if (first_err_pos == ULLONG_MAX) {
                first_err_pos = i;
            }
            err++;
        }
    }

    if (err > 0) {
        printf("result buf[%llu] is not right,exp: %d, act:%d \n",
            first_err_pos,
            c_buf[first_err_pos],
            result[first_err_pos]);
    }
    if (err > 0) {
        printf("total err is %d\n", err);
    }
    return err;
}

int check_buf_result_int16(const void *result_buf, const void *check_buf, u64 count)
{
    u64 i = 0;
    u16 *result = (u16 *)result_buf;
    u16 *s = (u16 *)check_buf;
    int err = 0;
    u64 first_err_pos = ULLONG_MAX;

    for (i = 0; i < count; ++i) {
        if (i == 0) {
            printf("------------aicpu is %u and hccl is %u\n", s[i], result[i]);
        }
        if (s[i] != result[i]) {
            if (first_err_pos == ULLONG_MAX) {
                first_err_pos = i;
            }
            err++;
        }
    }

    if (err > 0) {
        printf("[ERROR] result buf[%llu] is not right,exp: %u, act:%u \n",
            first_err_pos,
            s[first_err_pos],
            result[first_err_pos]);
    }
    if (err > 0) {
        printf("[ERROR] total err is %d\n", err);
    }
    return err;
}

int check_buf_result_int32(const void *result_buf, const void *check_buf, u64 count)
{
    u64 i = 0;
    int *c_buf = (int *)check_buf;
    int *result = (int *)result_buf;
    int err = 0;
    u64 first_err_pos = ULLONG_MAX;
    for (i = 0; i < count; ++i) {
        if (c_buf[i] != result[i]) {
            if (first_err_pos == ULLONG_MAX) {
                first_err_pos = i;
            }
            err++;
        }
    }

    if (err > 0) {
        printf("result buf[%llu] is not right,exp: %d, act:%d \n",
            first_err_pos,
            c_buf[first_err_pos],
            result[first_err_pos]);
    }
    if (err > 0) {
        printf("total err is %d\n", err);
    }
    return err;
}

int check_buf_result_int64(const void *result_buf, const void *check_buf, u64 count)
{
    u64 i = 0;
    s64 *c_buf = (s64 *)check_buf;
    s64 *result = (s64 *)result_buf;
    int err = 0;
    u64 first_err_pos = ULLONG_MAX;
    for (i = 0; i < count; ++i) {
        if (c_buf[i] != result[i]) {
            if (first_err_pos == ULLONG_MAX) {
                first_err_pos = i;
            }
            err++;
        }
    }

    if (err > 0) {
        printf("result buf[%llu] is not right,exp: %lld, act:%lld \n",
            first_err_pos,
            c_buf[first_err_pos],
            result[first_err_pos]);
    }
    if (err > 0) {
        printf("total err is %d\n", err);
    }

    return err;
}

int check_buf_result_u64(const void *result_buf, const void *check_buf, u64 count)
{
    u64 i = 0;
    u64 *c_buf = (u64 *)check_buf;
    u64 *result = (u64 *)result_buf;
    int err = 0;
    u64 first_err_pos = ULLONG_MAX;
    for (i = 0; i < count; ++i) {
        if (c_buf[i] != result[i]) {
            if (first_err_pos == ULLONG_MAX) {
                first_err_pos = i;
            }
            err++;
        }
    }

    if (err > 0) {
        printf("result buf[%llu] is not right,exp: %llu, act:%llu \n",
            first_err_pos,
            c_buf[first_err_pos],
            result[first_err_pos]);
    }

    if (err > 0) {
        printf("total err is %d\n", err);
    }

    return err;
}

static inline float fp32_from_bits(uint32_t w)
{
#if defined(__OPENCL_VERSION__)
    return as_float(w);
#elif defined(__CUDA_ARCH__)
    return __uint_as_float((unsigned int)w);
#elif defined(__INTEL_COMPILER)
    return _castu32_f32(w);
#else
    union {
        uint32_t as_bits;
        float as_value;
    } fp32 = {w};
    return fp32.as_value;
#endif
}

static inline uint32_t fp32_to_bits(float f)
{
#if defined(__OPENCL_VERSION__)
    return as_uint(f);
#elif defined(__CUDA_ARCH__)
    return (uint32_t)__float_as_uint(f);
#elif defined(__INTEL_COMPILER)
    return _castf32_u32(f);
#else
    union {
        float as_value;
        uint32_t as_bits;
    } fp32 = {f};
    return fp32.as_bits;
#endif
}

static inline uint16_t fp16_ieee_from_fp32_value(float f)
{
#if defined(__STDC_VERSION__) && (__STDC_VERSION__ >= 199901L) || defined(__GNUC__) && !defined(__STRICT_ANSI__)
    const float scale_to_inf = 0x1.0p+112f;
    const float scale_to_zero = 0x1.0p-110f;
#else
    const float scale_to_inf = fp32_from_bits(UINT32_C(0x77800000));
    const float scale_to_zero = fp32_from_bits(UINT32_C(0x08800000));
#endif
    float base = (fabsf(f) * scale_to_inf) * scale_to_zero;

    const uint32_t w = fp32_to_bits(f);
    const uint32_t shl1_w = w + w;
    const uint32_t sign = w & UINT32_C(0x80000000);
    uint32_t bias = shl1_w & UINT32_C(0xFF000000);
    if (bias < UINT32_C(0x71000000)) {
        bias = UINT32_C(0x71000000);
    }

    base = fp32_from_bits((bias >> 1) + UINT32_C(0x07800000)) + base;
    const uint32_t bits = fp32_to_bits(base);
    const uint32_t exp_bits = (bits >> 13) & UINT32_C(0x00007C00);
    const uint32_t mantissa_bits = bits & UINT32_C(0x00000FFF);
    const uint32_t nonsign = exp_bits + mantissa_bits;
    return (sign >> 16) | (shl1_w > UINT32_C(0xFF000000) ? UINT16_C(0x7E00) : nonsign);
}


//=================================================End Result Compare Function======================================

//=====================================================Start Matmul function======================================

int PrepareDeviceMemAndDataMatMul(const std::vector<int64_t> &selfShape, void **selfDeviceAddr,
    const std::vector<int64_t> &otherShape, void **otherDeviceAddr, const std::vector<int64_t> &biasShape,
    void **biasDeviceAddr, std::string dtype, std::string bin_path, int rankId)
{
    auto selfSize = GetShapeSize(selfShape) * dtype_to_sizeof(dtype);
    auto ret = aclrtMalloc(selfDeviceAddr, selfSize, ACL_MEM_MALLOC_HUGE_FIRST);
    CHECK_RET(ret == ACL_SUCCESS, std::cout << "aclrtMalloc failed. ERROR" << ret << std::endl; return -1);
    void *selfHostData = nullptr;
    ACL_CHECK(aclrtMallocHost((void **)&selfHostData, selfSize));
    std::string shape_str = GetShapeString("x1", selfShape, dtype) + "_input_" + to_string(rankId) + ".bin";
    ReadFile(bin_path + shape_str, selfSize, selfHostData, selfSize);
    std::cout << "qmy x1[0]" << ((int16_t *)selfHostData)[0] << std::endl;
    ret = aclrtMemcpy(*selfDeviceAddr, selfSize, selfHostData, selfSize, ACL_MEMCPY_HOST_TO_DEVICE);
    CHECK_RET(ret == ACL_SUCCESS, std::cout << "aclrtMalloc failed. ERROR" << ret << std::endl; return -1);

    auto otherSize = GetShapeSize(otherShape) * dtype_to_sizeof(dtype);
    ret = aclrtMalloc(otherDeviceAddr, otherSize, ACL_MEM_MALLOC_HUGE_FIRST);
    CHECK_RET(ret == ACL_SUCCESS, std::cout << "aclrtMalloc failed. ERROR" << ret << std::endl; return -1);
    void *otherHostData = nullptr;
    ACL_CHECK(aclrtMallocHost((void **)&otherHostData, otherSize));
    shape_str = GetShapeString("x2", otherShape, dtype) + "_input_0.bin";
    ReadFile(bin_path + shape_str, otherSize, otherHostData, otherSize);
    // std::cout<<"qmy x2_gm[0]" << ((int16_t*)otherHostData)[0]<<std::endl;
    ret = aclrtMemcpy(*otherDeviceAddr, otherSize, otherHostData, otherSize, ACL_MEMCPY_HOST_TO_DEVICE);
    CHECK_RET(ret == ACL_SUCCESS, std::cout << "aclrtMalloc failed. ERROR" << ret << std::endl; return -1);

    auto biasSize = GetShapeSize(biasShape) * dtype_to_sizeof(dtype);
    ret = aclrtMalloc(biasDeviceAddr, biasSize, ACL_MEM_MALLOC_HUGE_FIRST);
    CHECK_RET(ret == ACL_SUCCESS, std::cout << "aclrtMalloc failed. ERROR" << ret << std::endl; return -1);
    void *biasHostData = nullptr;
    ACL_CHECK(aclrtMallocHost((void **)&biasHostData, biasSize));
    shape_str = GetShapeString("bias", biasShape, dtype) + "_input_0.bin";
    ReadFile(bin_path + shape_str, biasSize, biasHostData, biasSize);
    // std::cout<<"qmy x2_gm[0]" << ((int16_t*)otherHostData)[0]<<std::endl;
    ret = aclrtMemcpy(*biasDeviceAddr, biasSize, biasHostData, biasSize, ACL_MEMCPY_HOST_TO_DEVICE);
    CHECK_RET(ret == ACL_SUCCESS, std::cout << "aclrtMalloc failed. ERROR" << ret << std::endl; return -1);

    ACL_CHECK(aclrtFreeHost(selfHostData));
    ACL_CHECK(aclrtFreeHost(otherHostData));
    ACL_CHECK(aclrtFreeHost(biasHostData));
    return 0;
}

int PrepareTestParamsMatMul(const std::vector<int64_t> &selfShape, void *selfDeviceAddr,
    const std::vector<int64_t> &otherShape, void *otherDeviceAddr, const std::vector<int64_t> &outShape,
    void *outDeviceAddr, aclDataType dataType, aclTensor **self, aclTensor **other, aclTensor **out)
{
    *self = CreateTensor(selfShape.data(), selfShape.size(), dataType, aclFormat::ACL_FORMAT_ND, selfDeviceAddr);
    *other = CreateTensor(otherShape.data(), otherShape.size(), dataType, aclFormat::ACL_FORMAT_ND, otherDeviceAddr);
    *out = CreateTensor(outShape.data(), outShape.size(), dataType, aclFormat::ACL_FORMAT_ND, outDeviceAddr);
    return 0;
}

//==================================================End Matmul function==========================================

//==================================================Start Test  function==========================================
typedef int32_t rtError_t;
typedef void *rtNotify_t;
typedef void *rtStream_t;

// init hccl args end
extern "C" rtError_t rtNotifyWait(rtNotify_t notify, rtStream_t stm);
extern "C" rtError_t rtNotifyRecord(rtNotify_t notify, rtStream_t stm);
extern "C" rtError_t rtNotifyCreate(int32_t deviceId, rtNotify_t *notify);
extern "C" HcclResult HcclCreateComResource(char *commName, uint32_t streamMode, void **commContext);
extern "C" HcclResult HcclGetAicpuOpStreamNotify(char *commName, rtStream_t *Opstream, void **aicpuNotify);

int callMatmulAndAicpu(HcclComm hcclComm, void *&gatherOut, void *&matmul2Out, void *&scatterOut, void *&reduceOut,
    DataSize &dataPara, aclrtStream rtStream, aclrtStream rtStream_aicpu, aclrtStream rtStream_record,
    aclrtEvent startEvent, aclrtEvent endEvent, int checkErr, int ndev, int rankId, int rootRank, uint32_t m,
    uint32_t k, uint32_t n, std::string dtype, std::string bin_path, std::string run_type, int bias_flag)  // 主函数
{
    struct timespec tp1, tp2;
    long cost;
    clock_gettime(CLOCK_MONOTONIC, &tp1);
    set_device_id(rankId);

    // **********Add Hcommname start********
    char hcom_name[N];
    auto hccl_name_ret = HcclGetCommName(hcclComm, hcom_name);
    CHECK_RET(
        hccl_name_ret == ACL_SUCCESS, printf("[ERROR] HcclGetCommName failed. hccl_name_ret = %d \n", hccl_name_ret);
        return -1);
    printf("rank %d hcom name is %s\n", rankId, hcom_name);
    // **********Add Hcommname end*********

    std::vector<int64_t> selfShape = {m, k};
    void *selfDeviceAddr = nullptr;
    std::vector<int64_t> otherShape = {k, n};
    void *otherDeviceAddr = nullptr;
    std::vector<int64_t> biasShape = {n};
    void *biasDeviceAddr = nullptr;
    std::vector<int64_t> scatterOutShape = {m / ndev, n};
    std::vector<int64_t> gatherOutShape = {m * ndev, k};
    std::vector<int64_t> matmul2OutShape = {m * ndev, n};
    std::vector<int64_t> reduceOutShape = {m, n};

    void *workspaceAddr = nullptr;
    auto ret = PrepareDeviceMemAndDataMatMul(selfShape,
        &selfDeviceAddr,
        otherShape,
        &otherDeviceAddr,
        biasShape,
        &biasDeviceAddr,
        dtype,
        bin_path,
        rankId);
    CHECK_RET(ret == 0, return ret);
    uint64_t workspaceSize = 0;
    aclOpExecutor *executor;
    int64_t comm_turn = 0;
    int64_t gatherIndex = 0;
    aclTensor *self = nullptr;
    aclTensor *other = nullptr;
    aclTensor *gm_bias = nullptr;
    aclTensor *scOut = nullptr;
    aclTensor *gaOut = nullptr;
    aclTensor *out2 = nullptr;
    aclTensor *reOut = nullptr;

    printf("start run Aclnn_MatmulRcsTp \n");
    if (run_type == "reduce") {
        auto reduceOutSize = GetShapeSize(reduceOutShape) * dtype_to_sizeof(dtype);
        ret = aclrtMalloc((void **)&reduceOut, reduceOutSize, ACL_MEM_MALLOC_HUGE_FIRST);
        CHECK_RET(ret == ACL_SUCCESS, std::cout << "aclMalloc reduceOut failed. ERROR" << ret << std::endl; return -1);

        self = CreateTensor(
            selfShape.data(), selfShape.size(), g_data_type_map[dtype], aclFormat::ACL_FORMAT_ND, selfDeviceAddr);
        other = CreateTensor(
            otherShape.data(), otherShape.size(), g_data_type_map[dtype], aclFormat::ACL_FORMAT_ND, otherDeviceAddr);
        reOut = CreateTensor(
            reduceOutShape.data(), reduceOutShape.size(), g_data_type_map[dtype], aclFormat::ACL_FORMAT_ND, reduceOut);
        if (bias_flag != 0){
            gm_bias = CreateTensor(biasShape.data(), biasShape.size(), g_data_type_map[dtype], aclFormat::ACL_FORMAT_ND, biasDeviceAddr);
        }

        auto aclnnRet = aclnnMatmulAllReduceV2GetWorkspaceSize(
            self, other, gm_bias, hcom_name, "sum", comm_turn, 1, reOut, &workspaceSize, &executor);
        CHECK_RET(aclnnRet == ACL_SUCCESS,
                  printf("[ERROR] aclnnMatmulAllReduceV2GetWorkspaceSize failed. aclnnRet = %d \n", aclnnRet);
                  return -1);
        printf("[INFO] mm_allreduce workspaceSize = %lu\n", workspaceSize);
        if (workspaceSize > 0) {
            ACL_CHECK(aclrtMalloc(&workspaceAddr, workspaceSize, ACL_MEM_MALLOC_HUGE_FIRST));
        }
        aclnnRet = aclnnMatmulAllReduceV2(workspaceAddr, workspaceSize, executor, rtStream);
        CHECK_RET(aclnnRet == ACL_SUCCESS, printf("[ERROR] aclnnMatmulAllReduceV2 failed. aclnnRet = %d \n", aclnnRet);
                  return -1);
        aclDestroyTensor(reOut);
        if (bias_flag != 0){
            aclDestroyTensor(gm_bias);
        }
    }

    aclDestroyTensor(self);
    aclDestroyTensor(other);
    ACL_CHECK(aclrtSynchronizeStreamWithTimeout(rtStream, 10000));

    ret = aclrtFree(workspaceAddr);
    CHECK_RET(ret == ACL_SUCCESS, printf("[ERROR] aclnnFree scatter workspaceAddr failed. ret = %d \n", ret);
              return -1);
    aclrtFree(selfDeviceAddr);
    aclrtFree(otherDeviceAddr);
    aclrtFree(biasDeviceAddr);

    clock_gettime(CLOCK_MONOTONIC, &tp2);
    if (tp2.tv_sec != tp1.tv_sec) {
        cost = tp2.tv_nsec - tp1.tv_nsec + 1000000000;
    } else {
        cost = tp2.tv_nsec - tp1.tv_nsec;
    }
    printf("[INFO] mc2 costtime = %lu, deviceid = %d\n", cost, rankId);

    return 0;
}

int launchOneThread(Args &args)
{
    std::string thread_name = "test" + std::to_string(args.logicDeviceId);
    prctl(PR_SET_NAME, thread_name.c_str());
    cpu_set_t cpu_set;
    CPU_ZERO(&cpu_set);

    int numa_node = args.logicDeviceId;
    // int cpu_start = 24 * numa_node;
    int cpu_start = 0;
    int cpu_num = 24;
    for (int i = cpu_start; i < cpu_start + cpu_num; i++) {
        CPU_SET(i, &cpu_set);
    }
    int ret = sched_setaffinity(gettid(), sizeof(cpu_set), &cpu_set);
    if (ret) {
        printf("failed to set cpu affinity, errno = %d\n", errno);
        return -1;
    }
    printf("0 rankId: %d, stream: %p, context : %p\n",
        args.logicDeviceId,
        args.resources.rtStream,
        args.resources.context);

    DataSize dataPara = args.dataPara;
    ret = 0;

    int checkErr = 0;
    ACL_CHECK(aclrtSetDevice(args.logicDeviceId));
    ACL_CHECK(aclrtSetCurrentContext(args.resources.context));
    // ACL_CHECK(aclrtCreateStream(&args.resources.rtStream));

    void *gatherOutAicpu = nullptr;
    void *scatterOutAicpu = nullptr;
    void *reduceOutAicpu = nullptr;
    void *matmul2OutAicpu = nullptr;

    for (int i = 0; i < args.loop_cnt; i++) {
        std::cout << "======Startloop: " << i << "============" << std::endl;

        ret = callMatmulAndAicpu(args.hcclComm,
            gatherOutAicpu,
            matmul2OutAicpu,
            scatterOutAicpu,
            reduceOutAicpu,
            dataPara,
            args.resources.rtStream,
            args.resources.rtStream_aicpu,
            args.resources.rtStream_record,
            args.resources.startEvent,
            args.resources.endEvent,
            checkErr,
            args.ndev,
            args.logicDeviceId,
            args.rootRank,
            args.m,
            args.k,
            args.n,
            args.dtype,
            args.bin_path,
            args.run_type,
            args.bias_flag);
        if (ret != 0) {
            printf("TestCall execute AICPU failed, Detailed logs are stored in path: /root/ascend/log/");
            return ret;
        }
        if (args.run_type == "reduce") {
            u64 compareSize = args.m * args.n * dtype_to_sizeof(args.dtype);

            // 保存结果
            void *hostAicpuBuff = nullptr;
            ACL_CHECK(aclrtMallocHost((void **)&hostAicpuBuff, compareSize));
            ACL_CHECK(aclrtMemcpy(
                (void *)hostAicpuBuff, compareSize, (void *)reduceOutAicpu, compareSize, ACL_MEMCPY_DEVICE_TO_HOST));
            writeBin(args.bin_path + "reduceOutAicpu" + "_" + to_string(args.logicDeviceId) + ".bin",
                hostAicpuBuff,
                compareSize);

            ACL_CHECK(aclrtFreeHost(hostAicpuBuff));
            ACL_CHECK(aclrtFree(reduceOutAicpu));
        }
    }

    ACL_CHECK(aclrtSynchronizeStreamWithTimeout(args.resources.rtStream, 10000));
    ACL_CHECK(aclrtDestroyStreamForce(args.resources.rtStream));
    ACL_CHECK(aclrtSynchronizeStreamWithTimeout(args.resources.rtStream_aicpu, 10000));
    ACL_CHECK(aclrtDestroyStreamForce(args.resources.rtStream_aicpu));
    ACL_CHECK(aclrtSynchronizeStreamWithTimeout(args.resources.rtStream_record, 10));
    ACL_CHECK(aclrtDestroyStreamForce(args.resources.rtStream_record));
    // 销毁集合通信域
    //  HCCLCHECK(HcclCommDestroy(args.hcclComm));
    HcclCommDestroy(args.hcclComm);
    // 重置设备
    ACL_CHECK(aclrtResetDevice(args.logicDeviceId));
    return 0;
}

extern "C" int rtStreamCreateWithFlags(void **stm, int32_t priority, uint32_t flags);
int launchMultiThread(Args &input_args, int32_t *devices, HcclComm *comms, Resource *resources)
{
    DataSize dataPara;
    dataPara.data_size = 10485760;
    dataPara.type_size = sizeof(uint16_t);
    dataPara.count = (dataPara.data_size + sizeof(uint16_t) - 1) / sizeof(uint16_t);  // data->count;
    dataPara.step_factor = 1;                                                         // data->step_factor;
    dataPara.step_bytes = 1;                                                          // data->step_bytes;
    dataPara.max_bytes = 10485760;                                                    // data->max_bytes;
    dataPara.min_bytes = 10485760;                                                    // data->min_bytes;
    dataPara.op_flag = 0;                                                             // op_flag;
    uint32_t ndev = input_args.ndev;
    Args args[ndev];
    aclrtStream rtStream[ndev * 3];
    aclrtContext context[ndev];
    for (uint32_t rankId = 0; rankId < ndev; rankId++) {
        ACL_CHECK(aclrtSetDevice(rankId));
        // ACL_CHECK(aclrtCreateStream(&rtStream[rankId]));
        // ACL_CHECK(aclrtCreateStream(&rtStream[ndev + rankId]));
        ACL_CHECK(rtStreamCreateWithFlags(&rtStream[rankId], 0, 0x600));
        ACL_CHECK(rtStreamCreateWithFlags(&rtStream[ndev + rankId], 0, 0x600));
        ACL_CHECK(rtStreamCreateWithFlags(&rtStream[ndev * 2 + rankId], 0, 0x600));
        ACL_CHECK(aclrtGetCurrentContext(&context[rankId]));
        // printf("1 rankId: %d, stream: %p\n", rankId,  rtStream[rankId]);
    }

    std::vector<std::unique_ptr<std::thread>> threads(ndev);
    for (uint32_t rankId = 0; rankId < ndev; rankId++) {
        args[rankId].rankId = rankId;
        args[rankId].rootRank = devices[0];
        args[rankId].ndev = ndev;
        args[rankId].logicDeviceId = devices[rankId];
        args[rankId].hcclComm = comms[rankId];
        args[rankId].resources = resources[rankId];
        args[rankId].dataPara = dataPara;
        args[rankId].resources.rtStream = rtStream[rankId];
        args[rankId].resources.rtStream_aicpu = rtStream[ndev + rankId];
        args[rankId].resources.rtStream_record = rtStream[ndev * 2 + rankId];
        args[rankId].resources.context = context[rankId];
        args[rankId].m = input_args.m;
        args[rankId].k = input_args.k;
        args[rankId].n = input_args.n;
        args[rankId].dtype = input_args.dtype;
        args[rankId].bin_path = input_args.bin_path;
        args[rankId].run_type = input_args.run_type;
        args[rankId].loop_cnt = input_args.loop_cnt;
        args[rankId].bias_flag = input_args.bias_flag;
        // printf("2 rankId: %d, stream: %p\n", rankId,  args[rankId].resources.rtStream);
        threads[rankId].reset(new (std::nothrow) std::thread(&launchOneThread, std::ref(args[rankId])));
    }
    for (uint32_t rankId = 0; rankId < ndev; rankId++) {
        threads[rankId]->join();
    }
    return 0;
}

//==================================================End Test function==========================================

int main(int argc, char *argv[])
{
    printf("Usage: xxx device_num m k n dtype bin_path run_type loop_cnt diffThd pctThd \n");
    int ret = 0;
    Args input_args;
    input_args.ndev = atoi(argv[1]);
    input_args.m = atoi(argv[2]);
    input_args.k = atoi(argv[3]);
    input_args.n = atoi(argv[4]);
    input_args.dtype = string(argv[5]);
    input_args.bin_path = string(argv[6]);
    input_args.run_type = string(argv[7]);  // gather, scatter, reduce
    input_args.loop_cnt = atoi(argv[8]);
    input_args.bias_flag = atof(argv[9]);

    if (input_args.ndev != 1 && input_args.ndev != 2 && input_args.ndev != 4 && input_args.ndev != 8) {
        printf("device_num input error, only support 8,4,2,1.\n");
        return -1;
    }
    int32_t devices[input_args.ndev];
    for (int i = 0; i < input_args.ndev; i++) {
        devices[i] = i;
    }

    HcclComm comms[N];
    Resource resources[N];

    ACL_CHECK(aclInit(NULL));
    for (int i = 0; i < 1; i++) {
        // 初始化集合通信域
        for (int i = 0; i < input_args.ndev; i++) {
            ACL_CHECK(aclrtSetDevice(devices[i]));
        }
        ret = HcclCommInitAll(input_args.ndev, devices, comms);
        if (ret != 0) {
            printf("This is an error in init_hcclComm.\n");
            return -1;
        }

        // //启动测试
        ret = launchMultiThread(input_args, devices, comms, resources);
        if (ret != 0) {
            printf("This is an error in opbase_test_by_data_size.\n");
            // ACL_CHECK(aclFinalize());
            return -1;
        }
    }
    ACL_CHECK(aclFinalize());
    return ret;
}
```

